#ifndef IATK_ADAPTER_H
#define IATK_ADAPTER_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/ae-api/api/adapter.h, tk_aeapi, abe.l6
// VERSION: 1.23 97/06/24 11:07:04
//
// INCLUDE AS:  <iatk/adapter.h>
//
// PURPOSE:  Defines class IAAdapter, the toolkit Adapter interface.
//
// CLASSES DECLARED:
//    IAAdapter - Abstract adapter interface object, from which all
//                concrete (domain-specific) adapters should be derived.
//
// NOTES:
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   =====  === ====== ===========
// New = 960607 JMG f10    Initial implementation of adapter/engine intf
// --- = 960622 JMG f82    Cleanup; add stop() and shutdown() methods.
// --- = 960626 JMG d151   Improve comments and fix names.
// $E1 = 960908 JMG f307   Allow adapters to be built using MS Visual C++
// $E2 = 960910 JEG f313   Remove exception throws from adapter interface
//       to allow adapters to be written in languages that do not support
//       throw/catch.
//
// $B1 = 960930 JMG d358   Add #include for iatk/iaerrors.h.
// $B2 = 961009 JMG d408   Make MS Visual C++ API library a DLL
// $A1 = 961114 JEG f480   Add quiesce, reset() and restart() for conduct
//       set reloading
//
// $E3 = 961209 WES f512   Add throw to registerProcedure and notify
//
// $E4 = 970422 JEG f771   Add service method to report errors to agent
// $B3 = 970623 JMG d900   With changes to provide IAAgent::reset(), there
//       is no longer an adapter::restart() call after adapter::reset().
//       So default for adapter::reset() is now like adapter::restart().
//
//******************************************************************************

#include <iatk/types.h>
#include <iatk/uniqueid.h>
#include <iatk/trigevnt.h>
#include <iatk/proctokn.h>
#include <iatk/atom.h>
#include <iatk/iaerrors.h>                                                      //@B1

class IAError;     // Forward declare for parameter type                        //@E2


//#############################################################################
//
// Class IAAdapter
//
// The base class from which all toolkit-conforming adapters are derived.
// IAAdapter provides implementation support used by concrete adapters
// (eg. the notify() and registerProcedure() methods).  It also defines
// "callback" methods that are implemented by the concrete Adapter, for
// example to permit Engines to sense additional attributes for an event
// being processed, or to initiate actions.
//
//#############################################################################

class IAAdapterImp;

class IAEXPCLASS IAAdapter {                                                    //@B2

 public:
   virtual ~IAAdapter() IAPURE;

   //=========================================================================
   // Service methods provided by IAAdapter for use in implementing concrete
   // toolkit Adapters.  These methods are not usually overridden by concrete
   // adapters derived from IAAdapter.
   //==========================================================================

   void notify(IATriggerEvent &event); // throw(IAError);                       //@E3
      //
      // notify - Begin processing of a new event
      //
      // A concrete Adapter calls the notify() function to inform all of the
      // observing engines of an event that should be processed.  The engines
      // (always) process the event asynchronously, hence the call to notify()
      // may (and usually does) return before event processing has completed
      // or perhaps has even begun.
      //
      // Completion of all engine processing of the event is signaled by a
      // a call to the adapter's eventComplete() member function.
      //
      // The toolkit framework guarantees that all "callbacks" to the Adapter
      // during processing of the event occur on a different thread than the
      // Adapter thread that issued notify().
      //
      // Input:
      //    event =
      //       Trigger event to be forwarded to engines for processing.
      //       The event header within the trigger event is completed by
      //       supplying the adapter's domain and event time stamp as part
      //       of the notify() processing.
      //
      //       Each instance of an IATriggerEvent can be used in at most
      //       one notify() call.  This requirement is in place to guarantee
      //       uniqueness of event ids (they are imbedded within the event
      //       header inside of the trigger event).
      //
      // The caller is free to destroy <event> immediately upon return
      // from this call.
      //
      // Exceptions                                                             //@E3
      //     IAError object thrown for an error                                 //@E3
      //
      // Threading/Locking Requirements:
      //
      // - The call to notify() may be temporarily delayed in order to control
      //   the rate at which events are queued at an engine.  To allow such
      //   flow control to work without stalls, the adapter thread that calls
      //   notify() must not hold any locks that would cause the adapter's
      //   answerQuery(), testCondition(), performAction() or eventComplete()
      //   functions to block.


   void registerProcedure(const char *procName,  IAProcType type,               //@E2
                                const char *signature, IAProcToken token);      //@E2
                                // throw(IAError)                               //@E3
      //
      // registerProcedure - Register a sensor or effector procedure
      //                     with the agent.
      //
      // Input:
      //    *procName =
      //       String containing the domain-specific identifier for the
      //       sensor or effector procedure.
      //
      //       This identifier is part of the fully-qualified name of the
      //       sensor or effector procedure, which is formed by concatenating
      //       the adapters domain name, a colon, and this identifier.
      //       For example, NNTP:GetNewNews.
      //
      //    type =
      //       Type of named procedure: boolean sensor (IABooleanSensorProc),
      //       sensor (IASensorProc) or effector (IAEffectorProc).
      //
      //    signature =
      //       A string defining the number and types of the terms expected
      //       in the atom passed to the named procedure.
      //
      //       The string is a comma- or blank-separated list of the following
      //       signature elements:
      //
      //          Integer       String       Symbol
      //          IntegerLVar   StringLVar   SymbolLVar
      //
      //       The signature elements can be in upper, lower, or mixed case.
      //
      //    token =
      //       Unique token assigned by the Adapter to represent this named
      //       procedure.  This token is passed on calls to testCondition(),
      //       answerQuery(), and performAction() to identify the procedure
      //       to be run.  The token can be an integer procedure index/ordinal,
      //       or a function pointer.  (See IAProcToken class.)
      //
      // Exceptions                                                             //@E3
      //              IAError object thrown when an error occurs                //@E3

   void IAAdapter::generateAlert(const char *type,                              //@E4
                               const IAAlertLevel sev,                          //@E4
                               const IAFactSet *fset,                           //@E4
                               const IAEventHeader *evhdr = NULL) const;        //@E4
                                // throw(IAError)                               //@E4
      //                                                                        //@E4
      // generateAlert - Let agent know about a problem that has occurred.      //@E4
      //                                                                        //@E4
      // Input:                                                                 //@E4
      //    *type =                                                             //@E4
      //       String containing an identifier for the type of error that       //@E4
      //       occurred.                                                        //@E4
      //                                                                        //@E4
      //       This identifier is comparable to the EventType fact of an        //@E4
      //       event.                                                           //@E4
      //                                                                        //@E4
      //    sev =                                                               //@E4
      //       IAAlertLevel indication of the severity of the alert.            //@E4
      //                                                                        //@E4
      //       Valid values are:                                                //@E4
      //          IAWarning -  Indicates that there is potential for a problem, //@E4
      //                       and some action may be required to prevent it    //@E4
      //                       from happening.                                  //@E4
      //                                                                        //@E4
      //          IADegraded - Indicates that the component is working in a     //@E4
      //                       degraded mode.  Generally, the degradation would //@E4
      //                       have a limited scope (i.e. broken for only a     //@E4
      //                       particular selector, or only a particular        //@E4
      //                       function isn't working).  This is more severe    //@E4
      //                       than a warning, but the component is still able  //@E4
      //                       to 'limp along'.                                 //@E4
      //                                                                        //@E4
      //          IAFatal -    Indicates that the component is no longer        //@E4
      //                       functional.                                      //@E4
      //                                                                        //@E4
      //    *fset =                                                             //@E4
      //       Fact set containing facts that describe the error.               //@E4
      //                                                                        //@E4
      //    *evhdr =                                                            //@E4
      //       (Optional) Event header containing event that was being          //@E4
      //       processed when the problem occurred.                             //@E4
      //                                                                        //@E4
      // Exceptions                                                             //@E4
      //              IAError object thrown when an error occurs                //@E4


   //==========================================================================
   // "Callback" methods called by the toolkit framework.  These must be
   // implemented by concrete Adapters that derive from IAAdapter.
   //==========================================================================

   virtual IABool IACALLBK identify() = 0;                                      //@E1
      //
      // identify - Sollicits registration of sensors and effectors.
      //
      // The toolkit framework calls an adapter's identify() function during
      // agent initialization to request that the adapter register the sensor
      // and effector procedures that it provides.
      //
      // identify() is called once for each Adapter instance created by the
      // adapter module's newAdapter() function.
      //
      // The adapter's identify() function should call registerProcedure
      // (zero or more times) to register each of its named procedures.
      //
      // Return:
      //    If the adapter is able to register its sensor and effector
      //    procedures successfully, it should return a value of 1 (true)
      //    from this call.  Otherwise it should return a zero value (false).


   virtual IABool IACALLBK start() {return (1==1);}                             //@E1
      //
      // start - Starts active adapter processing.
      //
      // The toolkit framework calls an adapter's start() function during
      // the agent-startup sequence to request that the adapter begin
      // active processing.
      //
      // When start() is called, the agent's configuration of adapters and
      // engines is established,  the engines have been started, and processing
      // of the AGENT:STARTING event has completed.  As a result, all sensor
      // or effector calls (for rule-based configuration) which may result
      // during the startup sequence have been completed by the time start()
      // is called.
      //
      // Return:
      //    If the adapter is able to start active processing successfully, it
      //    should return a value of 1 (true) from this call.  Otherwise, it
      //    should return a zero value (false).


   virtual IABool IACALLBK stop() {return (1==1);}                              //@E1
      //
      // stop - Stops active adapter processing (event generation)
      //
      // The toolkit framework calls an adapter's stop() function during
      // agent shutdown to request that the adapter end active processing.
      // In particular, this call is a request to the adapter to stop
      // generating new events.
      //
      // Before returning from the stop() call, the adapter should make
      // arrangements to prevent new calls to notify() from occurring.
      // If an adapter calls notify after returning from the call to stop(),
      // an error or exception may result on notify and the new event is
      // ignored.
      //
      // Even though the adapter is to cease generating new events, it should
      // remain prepared for sensor and effector calls for in-progress events,
      // until such time that its shutdown() function is called.
      //
      // Return:
      //    If the adapter is able to stop event generation successfully, it
      //    should return a value of 1 (true) from this call.  Otherwise, it
      //    should return a zero value (false).


   virtual void IACALLBK shutdown() {return;}                                   //@E1
      //
      // shutdown - Terminates an adapter.
      //
      // The toolkit framework calls an adapter's shutdown() function during
      // the agent shutdown sequence to request that the adaper perform an
      // orderly termination.
      //
      // An adapter's shutdown() function is called after its stop() function
      // has been called and returns, and after all event processing within
      // the agent has been completed.  Once shutdown() is called, no more
      // calls to the adapter's event-related functions (answerQuery(),
      // testCondition(), performAction() or eventComplete()) will occur.


   virtual void IACALLBK eventComplete(const IAEventHeader &hdr) {return;}      //@E1
      //
      // eventComplete - Informs an adapter of event completion.
      //
      // The toolkit framework calls an adapter's eventComplete() function
      // at the completion of processing of an inferencing event.
      //
      // At the time of the call, all sensor and effector calls that were
      // triggered by processing of the event have been completed.  However,
      // asynchronous actions initiated by some Adapter's effector procedures
      // may still be outstanding.
      //
      // The eventComplete() function is called on the adapter that initiated
      // processing of the event via notify(), and also on all adapters on
      // which answerQuery(), testCondition() or performAction() calls were
      // made during processing of the event.  The eventComplete() function
      // is called only once for each such adapter.
      //
      // Inputs:
      //    *hdr =
      //       A copy of the event header identifying the event for which
      //       completion is being signaled.


   virtual IABool IACALLBK testCondition(IAProcToken token,                     //@E1
                                         const IAEventHeader &hdr,
                                         const char *atom) {return 0;}
      //
      // testCondition - Runs a boolean sensor procedure.
      //
      // The toolkit framework calls an adapter's testCondition() function
      // when an engine wishes to employ a boolean sensor to ascertain the
      // truth or falsity of a condition (predicate) relative to a set
      // of arguments (bound terms).  If true, the condition results in an
      // additional short-term fact that the engine uses during processing
      // of the event.
      //
      // Processing of the event identified by <hdr> is suspended
      // until this call returns.
      //
      // Inputs:
      //    token =
      //       Token associated with the boolean sensor procedure to be run.
      //       This token matches the token assigned to the procedure on a
      //       registerProcedure() call.
      //
      //    hdr =
      //       A copy of the event header identifying the event that is
      //       being processed.
      //
      //    *atom =
      //       An atom, in the form of a null-terminated KIF string, the
      //       terms of which represent the arguments for this procedure
      //       call.  Use the IAAtom class to parse/analyze these terms.
      //
      // Return:
      //    true  = The condition evaluated by the sensor is true.
      //
      //    false = The condition evaluated by the sensor is false.


   virtual IAFactSet IACALLBK answerQuery(IAProcToken token,                    //@E1
                                          const IAEventHeader &hdr,
                                          const char *atom)
      {return IAFactSet();}
      //
      // answerQuery - Runs a sensor procedure.
      //
      // The toolkit framework calls an adapter's answerQuery() function
      // when an engine wishes to employ a sensor to obtain additional
      // information for event processing.  The sensor call represents a
      // query relative to a particular set of arguments (term).  The
      // results of the query are expressed as a set of additional short-
      // term facts the engine uses during processing of the event.
      //
      // Processing of the event identified by <hdr> is suspended
      // until this call returns.
      //
      // Inputs:
      //    token =
      //       Token associated with the sensor procedure to be run.
      //       This token matches the token assigned to the procedure on
      //       a registerProcedure() call.
      //
      //    hdr =
      //       A copy of the event header identifying the event that is
      //       being processed.
      //
      //    *atom =
      //       An atom, in the form of a null-terminated KIF string, the
      //       terms of which represent the arguments for this procedure
      //       call.  Use the IAAtom class to parse/analyze these terms.
      //
      // Return:
      //    A fact set containing the answers to the query.  The fact
      //    set should be empty if there are no facts for which the query
      //    is true.


   virtual void IACALLBK performAction(IAProcToken token,                       //@E1
                                       const IAEventHeader &hdr,
                                       const char *bindingStr) {return;}
      //
      // performAction - Runs an effector procedure.
      //
      // The toolkit framework calls an adapter's performAction() function
      // when an engine wishes to perform an action that is deemed appropriate
      // as a result of event processing.
      //
      // Logically, the requested actions is performed asynchrously to the
      // Engine's continued processing of the event.  Because of this
      // asynchronous nature, this function has been defined to return void.
      //
      // However, to allow effficient processing of short-running actions
      // (by avoid unnecessary thread creations), the Engine may be suspended
      // until this call returns (because the call is made on the Engine's
      // processing thread. The Adapter should execute the action on the
      // invoking thread only if it can be completed quickly.  If not, the
      // Adapter should defer the execution of the action to another thread
      // so that a prompt return from this call can be made.
      //
      // Inputs:
      //    token =
      //       Token associated with the effector procedure to be run.
      //       This token matches the token assigned to the procedure on
      //       a registerProcedure() call.
      //
      //    hdr =
      //       A copy of the event header identifying the event that is
      //       being processed.
      //
      //    *atom =
      //       An atom, in the form of a null-terminated KIF string, the
      //       terms of which represent the arguments for this procedure
      //       call.  Use the IAAtom class to parse/analyze these terms.

   virtual IABool IACALLBK restart(const char* selector) {return start();};     //@A1
      //
      // restart - Re-starts active adapter processing for a given selector.
      //
      // The toolkit framework calls an adapter's restart() function during
      // the conduct set re-load sequence to request that the adapter begin
      // active processing for the given selector (whose conduct sets have
      // been successfully reloaded).
      //
      // When restart() is called, the agent's configuration of adapters and
      // engines is established,  the engines have been started, and processing
      // of the AGENT:RELOAD event has completed.  As a result, all sensor
      // or effector calls (for rule-based configuration) which may result
      // during the startup sequence have been completed by the time restart()
      // is called.
      //
      // The concrete adapter must override the restart() function if the
      // adapter generates trigger events.
      //
      //  Failure to override the restart() function will result in the default
      //  restart() action, which is to invoke the adapter's start() function.
      //
      // Return:
      //    If the adapter is able to start active processing successfully, it
      //    should return a value of 1 (true) from this call.  Otherwise, it
      //    should return a zero value (false).

   virtual IABool IACALLBK reset(const char *selector) {return start();}        //@B3
      //
      // Reset - Discard data associated with the selector.                     //@B3
      //
      // The toolkit framework calls an adapter's reset() function during
      // agent reset processing to request that the adapter discard state or    //@B3
      // configuration data that is associated with the specified selector.
      // Generally, the semantic effect of the reset() call should be to
      // discard selector-related information so that afterwards the
      // adapter is in the same state as if it had never seen or processed
      // any calls for the given selector.  As a result, after this call
      // the adapter should no longer be generating trigger events for
      // the specified selector (until subsequent sensor or effector calls
      // result in new selector-related work for the adapter).
      //
      // Since the adapter should no longer be generating events for the        //@B3
      // specified selector, a call to reset() is NOT followed by a call
      // to restart().
      //
      // Return:
      //    If the adapter is able to reset for this selector successfully,
      //    it should return a value of 1 (true) from this call.  Otherwise,
      //    it should return a zero value (false).
      //

   virtual IABool IACALLBK quiesce(const char *selector) {return stop();};      //@A1
      //
      // Quiesce - Stops generating events for the given selector.
      //
      // The toolkit framework calls an adapter's quiesce() function during
      // conduct set reload and agent reset to request that the adapter stop
      // generating new events for the selector.
      //
      // Before returning from the quiesce() call, the adapter should make
      // arrangements to prevent new calls to notify() from occurring.
      // If an adapter calls notify after returning from the call to reset(),
      // an error or exception may result on notify and the new event is
      // ignored.
      //
      // Even though the adapter is to cease generating new events, it should
      // remain prepared for sensor and effector calls for in-progress events.
      //
      // The concrete adapter should override the quiesce() function if the
      // adapter generates trigger events.
      //
      // Return:
      //    If the adapter is able to stop event generation successfully, it
      //    should return a value of 1 (true) from this call.  Otherwise, it
      //    should return a zero value (false).
      //

 protected:
   IAAdapter() {vers = 2; impl = NULL;}                                         //@E1

 private:
   friend class IAAdapterImp;

   // The following four member functions are intended for use by the toolkit
   // framework only.  In spite of the use of "virtual" in the declaration,
   // these functions should not be overridden by adapters that derive from
   // IAAdaper.

   virtual int  IACALLBK version() const;                                       //@E1
   virtual void IACALLBK associate(IAAdapterImp *pimp);                         //@E1
   virtual void IACALLBK disassociate();                                        //@E1
   virtual void IACALLBK delete_();                                             //@E1

   int            vers;     // Adapter interface version number
   IAAdapterImp   *impl;    // Impl-specific private object.
};


//#############################################################################
//
// Global function newAdapter
//
// Function newAdapter() is a "factory" function for creating new instances of
// toolkit Adapters.  Each Adapter module (DLL) must provide an exported
// implementation of this function.  The toolkit Framework calls an Adapter
// module's newAdapter function to instantiate a new instance of a domain-
// specific adapter implemented within that Adapter module.
//
// Input:
//    *domain  = Domain name that identifies the type of concrete
//               Adapter requested.
//
//    *parms   = Adapter start-up parameter.  Usually, this parameter
//               will be a parameter string (const char *) obtained from
//               an Agent configuration file.  But in special cases, some
//               other data may be passed provided that the agent
//               constructor and the Adapter agree on the format.
//
// The adapter instance created by a call to this function is owned
// by the toolkit framework and should NOT ever be deleted by code in the
// Adapter module.
//
//
// The following is an example implemetation of ::newAdapter for an
// Adapter module that implements a single type of concrete Adapter:
//
// IAAdapter * newAdapter(const char *domain, void *parm)
// {
//    if (strcmp(domain, "MYDOMAIN")) {
//       ...issue msg about being asked to create wrong kind of adapter...
//       return NULL;
//    }
//
//    IAAdapter *adapter = new MyDomainAdapter();
//    if (!adapter) {
//       ...issue msg about not being able to new-up an adapter...
//       return NULL;
//    }
//
//    ...use parm string for configuration, as appropriate
//
//    return adapter;
// }
//
//#############################################################################

extern "C"
 IAEXPORT1 IAAdapter * IAEXPORT2 IACLINK                                        //@E1
   newAdapter(const char *domain, void *parm);

#endif // IATK_ADAPTER_H
