#ifndef IATK_AGENT_H
#define IATK_AGENT_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Intelligent Agent Toolkit                                              */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/ae-api/api/agent.h, tk_aeapi, abe.l6
// VERSION: 1.15 97/06/19 15:30:58
//
// INCLUDE AS:  <iatk/agent.h>
//
// PURPOSE:  Defines the IAAgent object, the top-level entity in an
//           agent configuration.
//
// DESCRIPTION:
//
// CLASSES DECLARED:
//    IAAgent - Top level object in an Agent configuration.
//
// NOTES:
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   =====  === ====== ===========
// New = 960607 JMG f10    Initial implementation of adapter/engine intf
// --- = 960622 JMG f82    Added processingRate method.
//
// $E4 = 960926 AC  f336   Provide a loadConductSet member function that has a
//                         selector parameter.
//
// $E5 = 961125 JMG f480   Implement conduct-set loading while agent is running
//
// $E7 = 961202 JEG F445   Implement addLibrary() to react to library changes
//
// $E8 = 961209 JMG f517   Enable IAAgent for use with MS Visual C++
//
// $E9 = 961211 JMG f528   Add unloadConductSet to IAAgent interface.
//
// $EA = 961212 JMG f527   Interface housecleaning (not chg flagged)
//
// $EB = 970131 WES f589   Support java adapters with java agents
//
// $EC = 970416 JEG f771   Implement agent callbacks
//
// $ED = 970617 JMG f883   Move reset function out to agent level
//
//******************************************************************************

#include <iatk/types.h>                                                         //@EC
#include <iatk/handles.h>
#include <iatk/iaerrors.h>                                                      //@E4
#include <iatk/atom.h>                                                          //@EC

//##############################################################################
//
// Class IAAgent
//
// Class IAAgent represents the top-level object in an Agent configuration.
//
// Threading Notes:
// - The IAAgent class is thread safe.
//
//##############################################################################

class IAAgentImp;          // Other side of intf-impl relationship
class ostream;


class IAEXPCLASS IAAgent {                                                      //@E8
 public:
   IAAgent();

   IAAgent(const char * type);                                            //@EB

   virtual ~IAAgent() IAPURE;

   virtual void IACALLBK handleAlert(const char *id, const IAAlertLevel sev,    //@EC
                                     IAFactSet &fset);                          //@EC
      //                                                                        //@EC
      // handleAlert - This method gets control when a component of the agent   //@EC
      //               (engine, adapter, library, etc.) generates an alert.     //@EC
      //               The alert could be a warning of some potential problem,  //@EC
      //               a notification that the component is disabled in some    //@EC
      //               way, or an indication that a fatal condition has caused  //@EC
      //               the component to cease all operations.                   //@EC
      //                                                                        //@EC
      // Input:                                                                 //@EC
      //    *id =                                                               //@EC
      //       String containing an alert identifier, in the form of            //@EC
      //            component:domain:problemType                                //@EC
      //                                                                        //@EC
      //       This alert id is composed of the FailingComponentType,           //@EC
      //       FailingDomain, and FailingEvent facts that are contained         //@EC
      //       in the IAFactSet argument.                                       //@EC
      //                                                                        //@EC
      //    sev =                                                               //@EC
      //       IAAlertLevel indication of the severity of the alert.            //@EC
      //                                                                        //@EC
      //       Valid values are:                                                //@EC
      //          IAWarning -  Indicates that there is potential for a problem, //@EC
      //                       and some action may be required to prevent it    //@EC
      //                       from happening.                                  //@EC
      //                                                                        //@EC
      //          IADegraded - Indicates that the component is working in a     //@EC
      //                       degraded mode.  Generally, the degradation would //@EC
      //                       have a limited scope (i.e. broken for only a     //@EC
      //                       particular selector, or only a particular        //@EC
      //                       function isn't working).  This is more severe    //@EC
      //                       than a warning, but the component is still able  //@EC
      //                       to 'limp along'.                                 //@EC
      //                                                                        //@EC
      //          IAFatal -    Indicates that the component is no longer        //@EC
      //                       functional.                                      //@EC
      //                                                                        //@EC
      //    *fset =                                                             //@EC
      //       Fact set containing facts that describe the error.               //@EC
      //                                                                        //@EC
      //       Standard facts that are included with every alert                //@EC
      //       (not necessarily in this order) are:                             //@EC
      //          FailingComponentType(String)                                  //@EC
      //          FailingType(String)                                           //@EC
      //          FailingDomain(String)                                         //@EC
      //          FailingName(String)                                           //@EC
      //                                                                        //@EC
      //       The type, domain, and name are analogous to EventType, EventName //@EC
      //       and EventDomain.                                                 //@EC
      //                                                                        //@EC
      //       If the alert is related to the processing of a particular        //@EC
      //       event, then the following facts may also be included             //@EC
      //       (not necessarily in this order):                                 //@EC
      //          FailingEventId(String)                                        //@EC
      //          FailingEventType(String)                                      //@EC
      //          FailingEventName(String)                                      //@EC
      //          FailingEventDomain(String)                                    //@EC
      //          FailingEventSelector(String)                                  //@EC
      //          FailingEventTime(String String)                               //@EC
      //                                                                        //@EC
      //       Any additional facts that may be included in the fact set        //@EC
      //       are determined by (and set by) the failing component, and should //@EC
      //       be documented with the component.                                //@EC
      //                                                                        //@EC
      // Defaults                                                               //@EC
      //                                                                        //@EC
      //       If this method is not implemented in the concrete adapter,       //@EC
      //       the default implementation will display an alert message,        //@EC
      //       followed by the alert id, the severity, and the facts.           //@EC
      //       The default implementation takes action based on the severity    //@EC
      //       of the alert, as follows:                                        //@EC
      //          IAWarning -  Agent continues running.  No additional action   //@EC
      //                       is taken.                                        //@EC
      //                                                                        //@EC
      //          IADegraded - Agent continues running.  No additional action   //@EC
      //                       is taken.                                        //@EC
      //                                                                        //@EC
      //          IAFatal -    Agent shuts down.                                //@EC
      //                       The shutdown occurs on a separate thread to      //@EC
      //                       prevent deadlocks with the calling component.    //@EC
      //                       The shutdown is accomplished by calling the      //@EC
      //                       concrete agent stop() method, and then exiting   //@EC
      //                       the process.                                     //@EC
      //                                                                        //@EC
      // Serialization                                                          //@EC
      //       This method receives synchronous control directly from the       //@EC
      //       reporting component.  The component may be in the middle of      //@EC
      //       processing an event, and could be delaying the event from        //@EC
      //       completing.  Therefore, if you implement any actions that        //@EC
      //       involve invoking any ABE API's, you must implement those actions //@EC
      //       on a separate thread, and return immediately from this thread.   //@EC
      //       Otherwise, your agent may experience a deadlock condition.       //@EC
      //       You do not need to start a new thread to loop through the input  //@EC
      //       fact set analysing or printing the facts.                        //@EC
      //                                                                        //@EC
      // Exceptions                                                             //@EC
      //              IAError object thrown when an error occurs                //@EC

   // Composing the Agent:

   void addAdapter(const char *domain,  const char *module,
                   void *parm, IAAdapterHandle &ahdl);

   void addEngine(const char *kind, const char *module,
                  void *parm, IAEngineHandle &ehdl);

   void connect(const IAAdapterHandle &ahdl, const IAEngineHandle &ehdl);

   void addLibrary(const char *map,                    // Map to library        //@E7
                   const char *topCollector,           // Top Collector name    //@E7
                   const char *ImplDllName,            // Implementation DLL    //@E7
                   const char *ImplType,               // Implementation type   //@E7
                   int levels,                         // # collector levels    //@E7
                   IALibraryHandle &lhdl);             // Returned lib handle   //@E7


   // Managing rules used by Engines:

   void loadConductSet(const IALibraryHandle &lhdl,                             //@E9
                       const IAEngineHandle  &ehdl,
                       const char *csname, const char *selector);               //@E9

   void unloadConductSet(const IALibraryHandle &lhdl,                           //@E9
                         const IAEngineHandle  &ehdl,                           //@E9
                         const char *csname, const char *selector);             //@E9


   // Controlling the Agent:

   void start();
   void stop();
   void reset(const char *selector);                                            //@ED

   // Diagnostic/debugging info:

   void dbgInfo(ostream &o) const;


   // Statistics:

   long eventsCompleted() const;

  static const char * agentType();                                              //@EB

 private:
   IAAgent(const IAAgent &from);               // Not implemented
   IAAgent & operator=(const IAAgent &from);   // Not implmeneted


 private:
   IAAgentImp    *impl;    // Impl-specific private object.
};

#endif // IATK_AGENT_H
