#ifndef IATK_ATOM_H
#define IATK_ATOM_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/ae-api/api/atom.h, tk_aeapi, abe.l6
// VERSION: 1.15 97/04/02 12:31:35
//
// INCLUDE AS:  #include <iatk/atom.h>
//
// PURPOSE:  Defines the IAAtom class that appears within
//           the IATK Adapter API.
//
// DESCRIPTION:
//
// CLASSES DECLARED:
//    IAAtom
//    IAFactSet
//    IAAtomPvt
//    IAError
//
// NOTES:
// - Concurrency control is the caller's responsibility; the implementation
//   of these classes does not perform its own locking.  All of the classes
//   require exclusive access to the object being operated upon (the "this"
//   object) as well as any others identified in function parameter lists.
//
//******************************************************************************

//******************************************************************************
//
// How to use this class:
//
// IAAtom is used to construct and view atoms and extract terms from atoms.
// Facts are atoms that contain all bound terms, so this class supports facts.
// Sensors and effectors use atoms too.
//
// KIF facts:  ( Subject "My new PC" )   ( MotherOf "Bill" "Lorraine" )
// once these facts are constructed as IAAtoms, getKif would return the
//      syntax above
//
// constructing the IAAtom:
// for all methods(default constructor only): IAAtom thisIAAtom;
//
// Building the Atoms
// Define an IAError object
//   IAError & status;
//
// Method 1: (from a predicate and KIF terms)
// thisIAAtom.setPredicate(string1,status)
// thisIAAtom.setTerms(string2,status)
// string1 must be a valid predicate and string 2 valid KIF terms
// A valid predicate is a symbol.  A symbol must begin with an alphabetic
// character and contain only alphanumeric characters
//   string1=Subject
//   string2="My new PC" (The quotes must be included in the string itself)
//     or
//   string1=MotherOf
//   string2="Bill" "Lorraine"
//
// Method 2: (adding term by term)
// thisIAAtom.setPredicate(string1,status)
// string1 must be a valid predicate
// thisIAAtom.addTerm(string2)
//  string1=Subject
//  string2=My new PC  (The quotes are not include in the string itself)
//         or
//  thisIAAtom.setPredicate(string1,status)
//  thisIAAtom.addTerm(string2)
//  thisIAAtom.addTerm(string3)
//  string1=MotherOf
//  string2=Bill
//  string3=Lorraine
//
// Method 3:  (Construction from whole KIF string)
//
//  thisIAAtom.setKIF(string1,status)
// string1 must be a valid KIF fact
//     string1= ( Subject "My new PC" )
//            or
//     string2= ( MotherOf "Bill" "Lorraine" )
//
// TriggerEvents have sets of IAAtoms (they'd better be facts).
//
// Sensors and Effectors pass IAAtoms as their binding strings.
//  The predicate is immaterial, but use getStringTerm, getIntegerTerm, etc.
// methods to extract the terms.
//
// If the above facts were actually sensors (or effectors)
//
// getStringTerm(1,status) on the first one returns the string,
//    My new PC (no quotes contained in string)
//
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   =====  === ====== ===========
// New = 960607 WES f10    Initial implementation of adapter/engine intf
// --- = 960622 JMG f82    Minor cleanup
// $A1 = 960807 WES f256   IABool results
// $A2 = 960910 WES f310   Change interface to handle return codes
// $A3 = 961009 JMG d408   Make MS Visual C++ API library a DLL
// $A4 = 961205 WES f512   Change interface to throw errors
//
//******************************************************************************

#include <iostream.h>
#include <iatk/types.h>                                                   //@A1

//##############################################################################
//
// Class IAAtom
//
//##############################################################################

class IAAtomImpl;

class IAError;

class IAEXPCLASS IAAtom {                                                 //@A3
 public:

   //---------------------------------------------------------------------------
   // Constructors and Destructors
   //---------------------------------------------------------------------------

   // default constructor only
   IAAtom();                                                              //@A2
   ~IAAtom();

   IAAtom(const IAAtom &from);

   //---------------------------------------------------------------------------
   // Assignment operator
   //---------------------------------------------------------------------------

   IAAtom & operator=(const IAAtom &from);

   //---------------------------------------------------------------------------
   // Equality and ordering operators
   //---------------------------------------------------------------------------

   IABool operator ==(const IAAtom &rhs) const;                           //@A1
   IABool operator !=(const IAAtom &rhs) const;                           //@A1
   IABool operator <(const IAAtom &rhs) const;                            //@A1

   //---------------------------------------------------------------------------
   // Conversion operator
   //---------------------------------------------------------------------------

   operator const char *() const;

   //---------------------------------------------------------------------------
   // Setter and extractor
   //---------------------------------------------------------------------------

   // May use this to replace logical variables with a constant term
   // Returns false if not a logical variable, nothing done, true if replaced
   void replaceLogicalVariable(int position,const char * term);           //@A4
                                    // throw(IAError);                    //@A4
   void replaceSymbolLogicalVariable(int position,const char * term);     //@A4
                                    // throw(IAError);                    //@A4
   void replaceLogicalVariable(int position,int term);                    //@A4
                                    // throw(IAError);                    //@A4
   void replaceLogicalVariable(int position,double term);                 //@A4
                                    // throw(IAError);                    //@A4

   const char * getKif() const;
   void setPredicate(const char *predicate); // throw(IAError);           //@A4
   void setTerms(const char * terms); // throw(IAError);                  //@A4
   void setKIF(const char * KIF); // throw(IAError);                      //@A4

   //---------------------------------------------------------------------------
   // Add and extract terms
   // Currently, terms are only added on the end
   // Terms are extracted from specific positions
   // Do I really need return code for addTerm?
   //---------------------------------------------------------------------------
                                                                          //@A2
   void addTerm(int integerValue);                                        //@A2
   void addTerm(double realValue);                                        //@A2
   void addTerm(const char * stringValue);                                //@A2
   void addSymbolTerm(const char * symbolValue); // throw(IAError);       //@A4

   // provided for completeness
   // first parm is type, must be a valid symbol - has KIF function syntax  @A2
   void addLogicalVariable(const char *);   // throw(IAError);            //@A4
                                            //  parameter is type

   int getNumberOfTerms() const;

   const char * getTermType(int position) const; // throw(IAError)        //@A4

   // is this a fact? - no logical variables or predicate placeholders
   IABool isFact() const;                                                 //@A1
   // second parm of isLogicalVariable is type
   IABool isLogicalVariable(int position, const char *) const;         //@A1@A4
                               // throw(IAError);                         //@A4
   // in this overload we check for any type of logical variable
   IABool isLogicalVariable(int position) const;  // throw(IAError)      //@A1@A4

   const char * getSymbolTerm(int position) const; // throw(IAError)      //@A4
   const char * getStringTerm(int position) const; // throw(IAError)      //@A4
   int getIntegerTerm(int position) const; // throw(IAError)              //@A4
   double getRealTerm(int position) const; // throw(IAError)              //@A4
   const char * getPredicate() const;

   //---------------------------------------------------------------------------
   // Debugging help
   //---------------------------------------------------------------------------

   void dbgInfo(ostream &o) const;


 private:
   IAAtomImpl *impl;      // Impl-specific private object.
};


//##############################################################################
//
// Class IAFactSet
//
//##############################################################################

class IAFactSetPvt;

class IAEXPCLASS IAFactSet {                                              //@A3
 public:
   //---------------------------------------------------------------------------
   // Constructors and Destructor
   //---------------------------------------------------------------------------

   IAFactSet();
   ~IAFactSet();

   IAFactSet(const IAFactSet &from);

   //---------------------------------------------------------------------------
   // Operator Overloads
   //---------------------------------------------------------------------------

   IAFactSet & operator=(const IAFactSet &rhs);

   // (No comparison operator defined.)

   //---------------------------------------------------------------------------
   // List manipulation functions
   //
   // add()
   //    Adds a copy of <atom> to the set.  The set's internal
   //    iteration cursor is invalidated.
   //
   // clear()
   //    Removes all facts in the set.  The set's iteration cursor
   //    is invalidated.
   //
   // size()
   //    Returns the number of elements in the fact set.
   //---------------------------------------------------------------------------

   void add(const IAAtom &fact);  // throw(IAError)                       //@A4
   void add(const char *fact);    // throw(IAError)                       //@A4
   void clear();
   int size() const;

   //---------------------------------------------------------------------------
   // List iteration functions
   //
   // first()
   //    If the list is non-empty, first() sets the internal list cursor
   //    to the first element of the list and returns a pointer to that
   //    element.  If the list is empty, first() invalidates the internal
   //    cursor and returns NULL.
   //
   // next()
   //    If the internal cursor is valid and list items remain, next()
   //    advances the cursor to the next item in the list and returns a
   //    pointer to that element.  If either the internal cursor is not
   //    valid, or there are no more items in the list, next() returns
   //    NULL.
   //---------------------------------------------------------------------------

   IAAtom * first();
   IAAtom * next();

   //---------------------------------------------------------------------------
   // Debugging help
   //---------------------------------------------------------------------------

   void dbgInfo(ostream &o) const;

 private:
   IAFactSetPvt *p;      // Impl-specific private object.
};


#endif
