#ifndef IATK_COLLECT_H
#define IATK_COLLECT_H
/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/library/c/collect.h, tk_library, abe.l6
// VERSION: 1.14 97/06/20 10:44:13
//
// INCLUDE AS:  #include <iatk/collect.h>
//
// PURPOSE:  Define the IALibInferenceCollector object.
//
// DESCRIPTION:
//    An IALibInferenceCollector groups objects for inferencing.
//    The objects that can be grouped include IALibInferenceRuleSets,
//    IALibInferenceLTFactSets, and IALibInferenceCollectors.
//    (An IALibInferenceCollector can contain other IALibInferenceCollectors
//    objects.)  This grouping of objects is similar to and can be implemented
//    on many platforms as hierarchical directories.
//
// CLASSES DECLARED:
//    IALibInferenceCollector - Group objects for inferencing
//
// NOTES:
//    IALibInferenceCollectorImp is the related class used to implement
//    the IALibInferenceCollector class.
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ---   ------ --- ------ -----------
// $xn = yymmdd iii pnnnnn (abstract of feature/defect)
// $a1 = 960911 tlw F00296 Add friend class IALibLogRecordSet and
//                         support new Collector cursor scope for LogRecordSets
// $a2 = 961029 tlw F00445 Support new IALibrary class; constructors and destructor
//                         are now private
// $a3 = 961121 tlw F00445 Add size method
// $a4 = 961211 tlw F00530 Library concurrency - add extractVersion method
// $a5 = 970113 tlw D00567 Add friend class IALibCacheElement
// $a6 = 970114 tlw D00572 Add new cursor methods that return pointer to ContentsElement
// $a7 = 970120 tlw D00572 Add new find method that returns pointer to ContentsElement
// $a8 = 970519 jeg F00813 Support for Microsoft Visual C++ compiler
// $a9 = 970620 jeg D00888 Rename getimpl() to getImplP()
//
//******************************************************************************


#include <iatk/types.h>
#include <iatk/contele.h>
#include <iatk/libconst.h>

class IALibrary;
class IALibInferenceCollectorImp;
class IALibInferenceCollectorProv;                                                                      //$a8

class IAEXPCLASS IALibInferenceCollector {                                                              //$a8

  public:

    // Prototype for member functions
    const char* extractName() const;           // Get the Collector Name
    IALibrary* extractLibraryPtr() const;      // Get the pointer to the associated IALibrary object       $a2

    void setMetadataName(const char mdn='0');            // Set the Collector metadata name
    char extractMetadataName() const;                    // Get the Collector metadata name

    void setMetadata(const char* mdata);                 // Set the Collector metadata
    const char* extractMetadata() const;                 // Get Collector metadata

    unsigned int extractVersion() const;                 // Get Collector version                         $a4

    void setCursorElementScope(IALibInferenceCollectorCursorScope cescope);
                                 // Define the type of ContentsElement the cursor will point at

    int get();                   // Get contents and the named metadata from Store
    int getWithoutMetadata();    // Get only the contents from Store

    void del();                  // Delete the Collector from Store
    IABool delMetadata(char mdname);   // Delete the named Collector metadata

    int put();                   // Put contents and the named metadata to Store
    int putWithoutMetadata();    // Put only the contents to Store

    IABool find( IALibInferenceContentsElement& contElt, const char* ename,
                 IALibInferenceContentsElementType etype, const char mdname = ' ');
            // Find a ContentsElement and return a reference to it

    IALibInferenceContentsElement* find( const char* ename, IALibInferenceContentsElementType etype,
                                         const char mdname = ' ');                                      //$a7
            // Find a ContentsElement and return a pointer to it                                          $a7

    IABool firstElement(IALibInferenceContentsElement& contElt);
            // Position cursor to first ContentsElement in the Collector and return reference to it
    IALibInferenceContentsElement* firstElement();                                                      //$a6
            // Position cursor to first ContentsElement in the Collector and return pointer to it         $a6
    IABool lastElement(IALibInferenceContentsElement& contElt);
            // Position cursor to last ContentsElement in the Collector and return reference to it
    IALibInferenceContentsElement* lastElement();                                                       //$a6
            // Position cursor to last ContentsElement in the Collector and return pointer to it          $a6
    IABool nextElement(IALibInferenceContentsElement& contElt);
            // Position cursor to next ContentsElement in the Collector and return reference to it
    IALibInferenceContentsElement* nextElement();                                                       //$a6
            // Position cursor to next ContentsElement in the Collector and return pointer to it          $a6
    IABool previousElement(IALibInferenceContentsElement& contElt);
            // Position cursor to previous ContentsElement in the Collector and return reference to it
    IALibInferenceContentsElement* previousElement();                                                   //$a6
            // Position cursor to previous ContentsElement in the Collector and return pointer to it      $a6

    int size() const;  // Return the number of contents elements in the collector                          $a3

    void debug_print(ostream& outfile);          // Print the Collector - debugging

  private:

    IALibInferenceCollector(IALibInferenceCollectorProv* cimpl, IALibrary* _lib);                       //$a8

    ~IALibInferenceCollector();          // Destructor                                                     $a2

    IALibInferenceCollectorProv *impl;   // Implementation-specific private object pointer              //$a8
    IALibrary                   *lib;                                                                   //$a8

    // Method is used by IATK code to get to wrapper implementation                                     //$a8
    IALibInferenceCollectorProv* getImplP() const;                                                      //$a9

    // Method only to be used by the following classes
    // to access the pointer to the IALibInferenceCollectorImp.
    IALibInferenceCollectorImp* getImpl() const;
    friend class IALibInferenceCollectorImp;
    friend class IALibInferenceLTFactSet;
    friend class IALibInferenceLTFactSetImp;
    friend class IALibInferenceRuleSet;
    friend class IALibInferenceRuleSetImp;
    friend class IALibLogRecordSet;                                                                      //$a1
    friend class IALibLogRecordSetImp;                                                                   //$a2
    friend class IALibrary;                                                                              //$a2
    friend class IALibCacheElement;                                                                      //$a5

};

#endif
