#ifndef IATK_HANDLES_H
#define IATK_HANDLES_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/ae-api/api/handles.h, tk_aeapi, abe.l6
// VERSION: 1.12 97/04/02 12:32:40
//
// INCLUDE AS:  <iatk/handles.h>
//
// PURPOSE:  Defines classes that act as the primary reference handles
//           to assorted objects within the IATK A/E framework.
//
//
// CLASSES DECLARED:
//    IAAdapterHandle - Opaque handle for an adapter component.
//    IAEngineHandle  - Opaque handle for an engine component.
//    IALibraryHandle - Opaque handle for a library component.
//    IAProcHandle    - Opaque handle for a adapter named procedure.
//
// NOTES:
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   =====  === ====== ===========
// New = 960607 JMG f10    Initial implementation of adapter/engine intf
//
// $E1 = 960622 JMG f82    Pre-alpha cleanup spree
//     - Added EventRec as friend of IAAgentHandle to allow EvenRec code
//       to get to IAAgent object via a handle.
//     - Adjusted IAProcHandle per IAProcToken changes.
//
// $E2 = 961112 JMG f477   Reduce use of collection classes in A/E API code
//     - Moved AdapterList, EngineList and EventList to their own header file
//     - Changed pub/private access on IAAdapterHandle::isValid to facilitate
//       re-implementation of AdapterList using the Sequence class.
//
// $E3 = 961119 JMG f480   Implement conduct-set loading while agent is running
//     - Added IALibraryHandle, which is used as part of the new form of
//       the IAAgent::loadConductSet() interface.
//
// $E4 = 961125 JMG f491   Clean up for externalizing IAAgent interface.
//
// $D1 = 961128 JMG d492   Move IAProcHandle to handles.h; needed for IAEngine
//
// $F1 = 961204 AC  f495   Make IAEngineImp a fried to IALibraryHandle, so
//                         engines can have accessability to library functions
//
//******************************************************************************

#include <iatk/types.h>                                                         //@E1
#include <iatk/proctokn.h>                                                      //@E1
#include <stddef.h>

class IAAdapterImp;
class IAEngineImp;
class IALibrary;                                                                //@E3


//##############################################################################
//
// IAAdapterHandle
//
//##############################################################################

class IAAdapterHandle {
 public:
   IAAdapterHandle() : impl(NULL) {}

   int operator ==(const IAAdapterHandle &rhs) const
      {return impl == rhs.impl;}
   int operator !=(const IAAdapterHandle &rhs) const
      {return !(*this == rhs);}

   int isValid() const                               // Chg'ed from priv to pub //@E2
      {return (impl != NULL);}                                                  //@E2

 private:
   friend class IAAgentImp;
   friend class IARtRegistry;
   friend class IAAdapterImp;
   friend class IAEngineImp;
   friend class IAProcHandle;
   friend class IAAdapterList;  // Only for impl ctor and implPtr()             //@E2

   IAAdapterHandle(IAAdapterImp *_impl)
      : impl(_impl) {}


   IAAdapterImp * implPtr() const        {return impl;}
   IAAdapterImp * operator ->() const    {return impl;}

   IAAdapterImp  *impl;
};


//##############################################################################
//
// IAEngineHandle
//
//##############################################################################

class IAEngineHandle {
 public:
   IAEngineHandle() : impl(NULL) {}

   int operator ==(const IAEngineHandle &rhs) const
      {return impl == rhs.impl;}
   int operator !=(const IAEngineHandle &rhs) const
      {return !(*this == rhs);}

 private:
   friend class IAAgentImp;
   friend class IARtRegistry;
   friend class IAAdapterImp;
   friend class IAEngineImp;
   friend class IAEngineList;   // Only for impl ctor and implPtr()             //@E2

   IAEngineHandle(IAEngineImp *_impl)
      : impl(_impl) {}

   int isValid() const
      {return (impl != NULL);}

   IAEngineImp * implPtr() const {return impl;}
   IAEngineImp * operator ->() const {return impl;}

   IAEngineImp   *impl;
};

                                                                              //#B@E3
//##############################################################################
//
// IALibraryHandle
//
//##############################################################################

class IALibraryHandle {
 public:
   IALibraryHandle() : lib(NULL) {}

   int operator ==(const IALibraryHandle &rhs) const
      {return lib == rhs.lib;}
   int operator !=(const IALibraryHandle &rhs) const
      {return !(*this == rhs);}

   int isValid() const
      {return (lib != NULL);}

 private:
   friend class IAAgentImp;
   friend class IAEngineImp;                                                    //$F1

   IALibraryHandle(IALibrary *_lib)
      : lib(_lib) {}

   IALibrary * libPtr() const        {return lib;}
   IALibrary * operator ->() const   {return lib;}

   IALibrary  *lib;
};
                                                                              //#E@E3

//##############################################################################
//
// IAProcHandle
//
//##############################################################################

class IAProcHandle {

 public:
   IAProcHandle()
      : kind(IANoProc), adapter(IAAdapterHandle()), token() {}                  //@E1

   ~IAProcHandle() {}

 private:
   friend class IAAdapterImp;   // Allow access to constructor
   friend class IARtRegistry;   // Allow access to extractors & ctor
   friend class IAEngineImp;    // Allow access to extractors & ctor

   IAProcHandle(const IAProcType &k, const IAAdapterHandle &a,
                const IAProcToken &t)
      : kind(k), adapter(a), token(t) {}

   IAProcType  procType()  const {return kind;}
   IAProcToken procToken() const {return token;}
   IAAdapterHandle adapterHdl() const {return adapter;}

   IAProcType       kind;        // Sensor, BooleanSensor or Effector
   IAAdapterHandle  adapter;     // Adapter providing attached procedure
   IAProcToken      token;       // That adapter's token for this procedure
};

#endif // IATK_HANDLES_H
