#ifndef IATK_IAERRORS_H
#define IATK_IAERRORS_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/misc/api/iaerrors.h, tk_aeapi, abe.l6
// VERSION: 1.32 97/06/19 16:09:18
//
// INCLUDE AS:  <iatk/iaerrors.h>
//
// PURPOSE:  Defines IATK error/exception class interface.
//
// CLASSES DECLARED:
//    IAError - Object used to return errors in IATK interfaces
//
// NOTES:
//    Be sure to update IAError.java whenever IA_E_ constants are changed
//
//******************************************************************************

//* Version History ************************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// NEW = 9605?? jmg        Create temporary IAError class
// --- = 960604 jph        Flesh out original temporary IAError class
// --- = 960618 jph        Add copy constructor so IAErrors can be copied and
//                         thrown as exceptions.
// --- = 960622 jmg f82    Add #defines for new errors now detected in
//                         adapter interface code.
// --- = 960624 jmg f113   Cleanup.
//
// $I1 = 960912 jeg f315   Add setter methods to enable IAError to be used as
//       a status parameter.
//
// $I2 = 960920 jeg f334   Split IAErrors into wrapper and implementation
//       classes.
// $I3 = 960930 jeg f360   ! function now returns IABool (for consistency)
// $I4 = 961008 jeg d401   New define for error
// $I5 = 961009 JMG d408   Make MS Visual C++ API library a DLL
// $I6 = 961119 TLW f485   New defines for Library errors
// $I7 = 961119 JMG f480   Implement conduct-set loading while agent is running
// $I8 = 961212 AC  f539   Add #define for ...UNLOAD_CONDUCTSET_FAILED
// $I9 = 970122 TLW d541   Categorize the error message numbers
// $IA = 970123 TLW d588   Get iaerrors.h in synch with iaerror.java
// $IB = 970124 TLW d591   Library error codes
// $IC = 970128 TLW d611   Library error codes
// $ID = 970221 TLW d660   Library error codes
// $IE = 970616 JMG f872   Add in some A/E API errors
// $IF = 970620 JMG f883   Additional error for IAAgent::reset()
//
//******************************************************************************

#include <iatk/types.h>                                                         //@I1

class ostream;    // Predeclare this, since it's used below
class IAErrorImpl;                                                              //@I2

//#############################################################################
//
// Class IAError
//
// An IAError object contains error information, which includes an
// error number and a message associated with the error.
//
//#############################################################################


class IAEXPCLASS IAError {                                                      //@I5
 public:
   //--------------------------------------------------------------------------
   // Simple Constructors, Destructor and Assignment operator
   //--------------------------------------------------------------------------

   IAError(int msgnumber, const char *msgpattern, ...);
      //                                                                        //@I2
      // Constructs an IAError with an error code and a message
      //
      // Input:
      //    msgnumber =
      //       Integer containing the error code
      //
      //    msgpattern =
      //       String containing the message associated with the error.
      //       The string can be a simple message, or can be a format-string
      //       which contains escape sequences and format specifications
      //       within the message (like ANSI C library function "printf" -
      //       see printf for format specifiers).
      //       Total length of message, including substitutions, can not
      //       exceed 500 characters.
      //
      //    (optional)
      //    ... =
      //       argument list.  If the msgpattern contains format
      //       specifications, the argument list is used to specify
      //       the values that are to be substituted into the string
      //       (like printf).  If the msgpattern does not contain
      //       substitutions, this argument is not used.
      //
      // Examples:
      //    IAError rc(99,"This is a simple message with no format specification");
      //    IAError rc(0,"No error, but include a message anyway");
      //
      //    int num = 30;
      //    char *text= "MESSAGE";
      //    IAError rc(99,
      //      "This contains string %10s integer %d using format specifications",
      //      text, num);
      //
      //                                                                        //@I2

   IAError(int msgnumber=0);
      // Constructs an IAError with an error code                               //@I2
      //
      // Input:
      //    (optional)
      //    msgnumber =
      //       Integer containing the error code.  If no number is
      //       provided, the default is 0 (no error).
      //
      // Examples:
      //    IAError rc();       // Define with error code 0 (no error)
      //    IAError rc(0);      // Define with error code 0 (no error)
      //    IAError rc(99);     // Define with error code 99 (error)
      //                                                                        //@I2

   IAError(const IAError &copy);

   ~IAError();

   IAError & operator=(const IAError &from);                                    //@I2

   //--------------------------------------------------------------------------
   // Attribute access functions
   //
   // msgnumber = Returns the error code
   // errormsg  = Returns the message associated with the error
   // isOk      = Returns true if error code is 0 (no error),
   //             returns false if error code is non-zero (error).
   // !         = Returns true if error code is non-zero (error),
   //             returns false if error code is zero (no error).
   //--------------------------------------------------------------------------

   int msgnumber() const;
   const char *errormsg() const;

   IABool isOk(void);                                                           //@I1

   IABool operator !() const;                                                   //@I3

   //--------------------------------------------------------------------------
   // Setter functions - see corresponding constructors for information
   // on parameters.
   //--------------------------------------------------------------------------

   void set(int msgnumber, const char *msgpattern, ...);                        //@I1
   void set(int msgnumber);                                                     //@I1

   //--------------------------------------------------------------------------
   // Debugging help
   //--------------------------------------------------------------------------

   void dbgInfo(ostream &o) const;                                              //@I2

  private:                                                                      //@I2
    IAErrorImpl *impl;       // Implementation-specific private object          //@I2

};

IAEXPORT1 ostream &operator <<(ostream &stream,
                     const IAError &error);

//-----------------------------------------------------------------------------
// Define IATK error message numbers
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// 0 - 99: General errors                                                       //@I9
//-----------------------------------------------------------------------------
#define IA_E_OK                                       0   // No error occurred
#define IA_E_CALL_FAILED                              1
#define IA_E_INTERNAL_ERROR                           2
#define IA_E_INVALID_PARM                             3
#define IA_E_WRONG_STATE                              4
#define IA_E_GENERAL_PLACEHOLDER_1                    5
#define IA_E_GENERAL_PLACEHOLDER_2                    6
#define IA_E_GENERAL_PLACEHOLDER_3                    7
#define UNKNOWN_MSG_NUMBER                           99                         //@IA

//-----------------------------------------------------------------------------
// 100 - 199: Agent errors                                                      //@I9
//-----------------------------------------------------------------------------
#define IA_E_AGENT_HANDLE_BAD                        100
#define IA_E_ADAPTER_HANDLE_BAD                      101
#define IA_E_ENGINE_HANDLE_BAD                       102
#define IA_E_MODULE_NAME_BAD                         103
#define IA_E_ENGINE_KIND_BAD                         104
#define IA_E_ADAPTER_DOMAIN_BAD                      105
#define IA_E_CANT_ADD_ADAPTER_NOW                    106
#define IA_E_CANT_ADD_ENGINE_NOW                     107
#define IA_E_CANT_CONNECT_A_E_NOW                    108
#define IA_E_CANT_LOAD_COND_SET_NOW                  109
#define IA_E_CANT_START_AGENT_NOW                    110
#define IA_E_CANT_STOP_AGENT_NOW                     111
#define IA_E_CONDUCTSET_LOAD_ERR                     112
#define IA_E_NO_ENGINES_LOADED                       113                        //@I7
#define IA_E_NO_ADAPTERS_LOADED                      114                        //@I7
#define IA_E_ADAPTERS_NOT_CONNECTED                  115                        //@I7
#define IA_E_ENGINE_UNLOAD_CONDUCTSET_FAILED         116                        //@I8
#define IA_E_ENGINE_DRAIN_FAILED                     117                        //@IE
#define IA_E_ADAPTER_QUIESCE_FAILED                  118                        //@IE
#define IA_E_ADAPTER_RESET_FAILED                    119                        //@IE
#define IA_E_ADAPTER_RESTART_FAILED                  120                        //@IE
#define IA_E_CSCHANGE_EVENT_FAILED                   121                        //@IE
#define IA_E_CONFIG_EVENT_FAILED                     122                        //@IE
#define IA_E_ENGINE_RESET_FAILED                     123                        //@IF
#define IA_E_AGENT_PLACEHOLDER_1                     124  // Placeholders
#define IA_E_AGENT_PLACEHOLDER_2                     125
#define IA_E_AGENT_PLACEHOLDER_3                     126
//-----------------------------------------------------------------------------
// 200 - 299: Library errors                                                    //@I9
//-----------------------------------------------------------------------------
#define IA_E_LIBRARY_MAP_BAD                         200                        //@I6
#define IA_E_LIBRARY_TYPE_BAD                        201                        //@I6
#define IA_E_LIBRARY_NAME_BAD                        202                        //@I6
#define IA_E_LIBRARY_MODULE_LOAD_FAIL                203                        //@I6
#define IA_E_LIBRARY_FIND_FAIL                       204                        //@I6
#define IA_E_LIBRARY_IMPLEMENTATION_NOT_BUILT        205                        //@I6
#define IA_E_LIBRARY_VERSION_ERROR                   206                        //@IB
#define IA_E_LIBRARY_COLLECTOR_EXISTENCE_ERROR       207                        //@I9
#define IA_E_LIBRARY_COLLECTOR_METADATA_EXISTENCE_ERROR 208                     //@I9
#define IA_E_LIBRARY_INVALID_PARM                    209                        //@IB
#define IA_E_LIBRARY_FILE_READ_ERROR                 210                        //@IB
#define IA_E_LIBRARY_FILE_WRITE_ERROR                211                        //@IC
#define IA_E_LIBRARY_RULE_LTFACT_INVALID             212                        //@IB
#define IA_E_LIBRARY_CURSOR_INVALID                  213                        //@IC
#define IA_E_LIBRARY_LOGRECORDSET_ERROR              214                        //@IC
#define IA_E_LIBRARY_FILE_OPEN_ERROR                 215                        //@JG
#define IA_E_LIBRARY_CANNOT_DELETE_COLLECTOR         216                        //@ID
#define IA_E_LIBRARY_PLACEHOLDER_1                   217  // Placeholders
#define IA_E_LIBRARY_PLACEHOLDER_2                   218
#define IA_E_LIBRARY_PLACEHOLDER_3                   219                        //@ID

//-----------------------------------------------------------------------------
// 300 - 399: Adapter/Engine errors                                             //@I9
//-----------------------------------------------------------------------------
#define IA_E_INVALID_KIF_SYNTAX                      300
#define IA_E_INVALID_PREDICATE                       301
#define IA_E_INVALID_SYMBOL                          302
#define IA_E_INVALID_STRING                          303
#define IA_E_INVALID_INTEGER                         304
#define IA_E_INVALID_REAL                            305
#define IA_E_INVALID_TERM_NUMBER                     306
#define IA_E_CANT_REGISTER_NOW                       307
#define IA_E_PROC_NAME_BAD                           308
#define IA_E_PROC_SIGNATURE_BAD                      309
#define IA_E_PROC_SIGNATURE_ITEM_BAD                 310
#define IA_E_PROC_SIGNATURE_HAS_VARS                 311
#define IA_E_PROC_ALREADY_REGISTERED                 312
#define IA_E_CANT_NOTIFY_NOW                         313
#define IA_E_NOT_ATTACHED                            314
#define IA_E_WRONG_PROC_TYPE                         315
#define IA_E_INVALID_EVENT_HEADER                    316
#define IA_E_CANT_REUSE_EVENT                        317                        //@I4
#define IA_E_CANT_DELIVER_EVENT                      318                        //@I4
#define IA_E_NULL_KIF_STRING                         319
#define IA_E_INVALID_LOGICAL_VARIABLE                320
#define IA_E_TERM_TYPE_MISMATCH                      321
#define IA_E_INVALID_FACT                            322
#define IA_E_INVALID_EVENT_TYPE                      323
#define IA_E_INVALID_EVENT_SELECTOR                  324
#define IA_E_CANT_NOTIFY_FOR_SELECTOR                325
#define IA_E_ENGINE_LOAD_CONDUCTSET_FAILED           326                        //@I7
#define IA_E_ENGINE_THREW_EXCEPTION                  327                        //@I7
#define IA_E_ADAPTER_THREW_EXCEPTION                 328                        //@I7
#define IA_E_AE_PLACEHOLDER_1                        329  // Placeholders
#define IA_E_AE_PLACEHOLDER_2                        330
#define IA_E_AE_PLACEHOLDER_3                        331

//-----------------------------------------------------------------------------
// 400 - 499: Java Adapter errors                                               //@IA
//-----------------------------------------------------------------------------
#define NEWADAPTER_EXCEPTION                         400
#define PERFORMACTION_EXCEPTION                      401
#define ANSWERQUERY_EXCEPTION                        402
#define TESTCONDITION_EXCEPTION                      403
#define JAVA_SOCKET_INITIALIZING_ERROR               404
#define JAVA_CLIENT_SOCKET_EXCEPTION                 405
#define JAVA_ADAPTER_ARGUMENT_ERROR                  406
#define SOCKET_FUNCTION_ID_ERROR                     407
#define SOCKET_READSTRING_ERROR                      408
#define SOCKET_READINTEGER_ERROR                     409
#define SOCKET_SENDINTEGER_ERROR                     410
#define SOCKET_SENDSTRING_ERROR                      411
#define SOCKET_SENDVOID_ERROR                        412
#define UNKNOWN_ADAPTER_FUNCTION_ERROR               413
#define ADAPTER_CALLBACK_EXCEPTION                   414
#define JAVA_ADAPTER_ID_ERROR                        415
#define JAVA_ADAPTER_RUNTIME_EXCEPTION               416


//-----------------------------------------------------------------------------
// Define a macro which can be used to catch and describe an IAError object
// when one is thrown as an exception.  Don't code a semicolon after an
// invocation of this macro.
#define CATCH_IAERRORS()            \
   catch(const IAError &exc) {      \
      cerr << exc << endl;          \
      }

#endif // IATK_IAERRORS_H
