#ifndef IATK_RULESET_H
#define IATK_RULESET_H
/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/library/c/ruleset.h, tk_library, abe.l6
// VERSION: 1.14 97/06/19 18:12:00
//
// INCLUDE AS:  #include <iatk/ruleset.h>
//
// PURPOSE:  Define the IALibInferenceRuleSet object.
//
// DESCRIPTION:
//    An IALibInferenceRuleSet is a set of Rules grouped
//    together to be used for inferencing.
//
// CLASSES DECLARED:
//    IALibInferenceRuleSet - Rule Set
//
// NOTES:
//    IALibInferenceRuleSetImp is the related class used to implement
//    the IALibInferenceRuleSet class.
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ---   ------ --- ------ -----------
// $xn = yymmdd iii pnnnnn (abstract of feature/defect)
// $a1 = 960920 tlw F00296 Add signatures to find() and first/last/next/previousRule()
//                         to return a pointer to a Rule
// $a2 = 960923 tlw F00337 Add updateRule() method
// $a3 = 961011 tlw F00414 Add currentRule() method
// $a4 = 961029 tlw F00445 Support new IALibrary class; constructor and destructor
//                         are now private
// $a5 = 961030 tlw F00445 Change method names so no duplicates
// $a6 = 961121 tlw F00445 Add size method
// $a7 = 961210 tlw D00502 Make copy constructor and assignment operator private
// $a8 = 970107 tlw D00559 Add extractVersion method
// $a9 = 970124 tlw D00600 Default setRulesMetadataName() parm to '0'
// $aa = 970519 jeg F00813 Support for Microsoft Visual C++ compiler
//
//******************************************************************************

#include <iatk/types.h>
#include <iatk/rule.h>

/*
#ifndef IATK_RULESETCONSTANTS
#define IATK_RULESETCONSTANTS
char* const ACTIVERS = "A";  	     // Active RuleSet
char* const INACTIVERS = "I";      // Inactive RuleSet
#endif
*/
class IALibrary;                                                                                        //$aa
class IALibInferenceRuleSetProv;                                                                        //$aa

class IAEXPCLASS IALibInferenceRuleSet {                                                                //$aa

  public:

    // Prototype for member functions
    const char* extractName() const;           // Get the RuleSet Name
    const char* extractID() const;             // Get the RuleSet ID

    void setMetadataName(const char mdn='0');    // Set the RuleSet metadata name
    char extractMetadataName() const;            // Get the RuleSet metadata name

    void setMetadata(const char* mdata);         // Set the RuleSet-level metadata
    const char* extractMetadata() const;         // Get RuleSet-level metadata

    unsigned int extractVersion() const;         // Get RuleSet version                                   $a8

    void setRulesMetadataName(const char rmdname='0');  // Set the Rules-level metadata name              $a9
    const char extractRulesMetadataName() const;   // Get the Rules-level metadata name

    void activate();                 // Activate the RuleSet
    void inactivate();               // Inactivate the RuleSet

    IABool isActive() const;               // Is the RuleSet active?

    int get();                     // Get Rules and the named metadata from Store
                                   // Return value = 0 indicates success
                                   // Return value != 0 indicates failure
    int getWithoutMetadata();      // Get only the Rules from Store
                                   // Return value = 0 indicates success
                                   // Return value != 0 indicates failure

    void del();                    // Delete the RuleSet from Store
    IABool delMetadata(char mdname);   // Delete the RuleSet Metadata from Store
    void delRulesMetadata(char mdname);      // Delete the Rules Metadata from Store

    int put();                     // Put Rules and the named metadata to Store
                                   // Return value = 0 indicates success
                                   // Return value > 0 indicates position of Rule that failed validation;
                                   //   (no Rules have been put)
    int putWithoutMetadata();              // Put only the Rules to Store
                                           // Return value = 0 indicates success
                                           // Return value > 0 indicates position of Rule that failed validation;
                                           //   (no Rules have been put)

    IABool find(const char* ruleName, IALibInferenceRule& rule);  // Find a Rule given its name and return a reference to it
    IALibInferenceRule* find(const char* ruleName);               // Find a Rule given its name and return a pointer to it $a1

    void addRule(const IALibInferenceRule& rule);              // Add a Rule to the end of the RuleSet, positioning cursor to added Rule
    void addRuleAsNext(const IALibInferenceRule& rule);        // Add a Rule (after the current position) to the RuleSet, positioning cursor to added Rule
    void addRuleAsPrevious(const IALibInferenceRule& rule);    // Add a Rule (before the current position) to the RuleSet, positioning cursor to added Rule
    int updateRule(IALibInferenceRule& rule);    // Update the Rule at the current cursor position, keeping the cursor at the current Rule $a2
    IABool deleteRule();                   // Delete Rule at current cursor position by setting "Delete" field

    IABool firstRule(IALibInferenceRule& rule);      // Position cursor to first Rule in the RuleSet and return reference to Rule
    IALibInferenceRule* firstRule();      // Position cursor to first Rule in the RuleSet and return pointer to Rule $a1$a5
    IABool lastRule(IALibInferenceRule& rule);       // Position cursor to last Rule in the RuleSet and return reference to Rule
    IALibInferenceRule* lastRule();       // Position cursor to last Rule in the RuleSet and return pointer to Rule  $a1$a5
    IABool nextRule(IALibInferenceRule& rule);       // Position cursor to next Rule in the RuleSet and return reference to Rule
    IALibInferenceRule* nextRule();       // Position cursor to next Rule in the RuleSet and return pointer to Rule  $a1$a5
    IABool previousRule(IALibInferenceRule& rule);   // Position cursor to previous Rule in the RuleSet and return reference to Rule
    IALibInferenceRule* previousRule();   // Position cursor to previous Rule in the RuleSet and return pointer to Rule $a1$a5
    IABool currentRule(IALibInferenceRule& rule) const;   // Return reference to current Rule                    $a3
    IALibInferenceRule* currentRule();   // Return pointer to current Rule                                 $a3$a5

    int size() const;  // Return the number of undeleted rules in the RuleSet                              $a6

    void debug_print(ostream& outfile) const;   // Print the RuleSet - debugging

  private:

    IALibInferenceRuleSet (IALibInferenceRuleSetProv* rsimpl, IALibrary* _lib);                         //$aa

    ~IALibInferenceRuleSet();         // Destructor

    IALibInferenceRuleSetProv *impl;   // Implementation-specific private object pointer                //$aa
    IALibrary                 *lib;                                                                     //$aa

    friend class IALibrary;                                                                              //$a4

};
#endif
