#ifndef IATK_TRIGEVNT_H
#define IATK_TRIGEVNT_H

/*** Copyright ****************************************************************/
/*                                                                            */
/* IBM Agent Building Environment                                             */
/*                                                                            */
/* (C) Copyright IBM Corp. 1996                                               */
/* All Rights Reserved.                                                       */
/*                                                                            */
/* Licensed Materials - Property of IBM                                       */
/*                                                                            */
/* US Government Users Restricted Rights - Use, duplication or                */
/* disclosure restricted by GSA ADP Schedule Contract with IBM Corp.          */
/*                                                                            */
/******************************************************************************/

//******************************************************************************
//
// FILE:    src/ae-api/api/trigevnt.h, tk_aeapi, abe.l6
// VERSION: 1.14 97/06/03 10:50:09
//
// INCLUDE AS:  #include <iatk/trigevnt.h>
//
// PURPOSE:  Defines classes related to toolkit trigger events.
//
// CLASSES DECLARED:
//    IAEventHeader   - Standard descriptive information about an event
//    IATriggerEvent  - Header and facts consituting an event to be processed
//
// THREADING NOTES:
// - Concurrency control is the caller's responsibility; the implementation
//   of these classes does not perform its own locking.  All of the classes
//   require exclusive access to the object being operated upon (the "this"
//   object) as well as any others identified in function parameter lists.
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   =====  === ====== ===========
// New = 960622 JMG f82    Newly split from former apiobjs.h.
// $E1 = 960922 JMG f341   Add rule-set selector to event header
// $E2 = 961008 WES f406   Removal of throws from TriggerEvent and EventHeader
//                         constructors
// $B1 = 961009 JMG d408   Make MS Visual C++ API library a DLL
// $E3 = 961209 WES f512   Add throws for error conditions
// $E4 = 970417 JEG f771   Add event completion status
// $E5 = 970512 WES        eventNumber needed
//
//******************************************************************************

#include <string.h>
#include <time.h>
#include <iatk/types.h>


//#############################################################################
//
// Class IAEventHeader
//
// An IAEventHeader object contains standard header data that describes a
// particular trigger event.  This header is (partially) created by the
// Adapter that initiates event processing.  The header (within a trigger
// event) is completed by IAAdapter::notify().  A copy of the completed
// header is a parameter on many calls made during event processing.
//
//#############################################################################

class IAError;                                                                  //@E1
class IAEventHeaderPvt;
class IAUniqueId;
class ostream;

class IAEXPCLASS IAEventHeader {                                                //@B1
 public:
   //--------------------------------------------------------------------------
   // Simple Constructors, Destructor and Assignment operator
   //--------------------------------------------------------------------------

   IAEventHeader();
      // Header built by default constructor is not valid for use in a
      // trigger event.  Use setType below to build a usable                    //@E2
      // event header.

   ~IAEventHeader();

   IAEventHeader(const IAEventHeader &from);
   IAEventHeader & operator=(const IAEventHeader &rhs);


   //--------------------------------------------------------------------------
   // Initialize a (partially-filled in) event header                           //@E2
   //--------------------------------------------------------------------------

   void setType(const char *type);  // throw(IAError)                     //@E3
      //
      // Initializes an IAEventHeader.                                          //@E2
      //
      // Input:
      //    *type =
      //       String containing the domain-specific identifier for the type
      //       of event being generated, eg. "NewNews". This type identifier
      //       is part of the fully-qualified name for the event.
      //
      //       The fully-qualified name for the event is formed by concat-
      //       enating the domain name of the initiating adapter, a colon,
      //       and the type identifier specified on this constructor.  For
      //       example, the fully qualified name of a NewNews event generated
      //       by an NNTP adapter is "NNTP:NewNews".
      //
      // Exceptions:                                                      //@E3
      //      IAError                                                     //@E3
      //      Thrown if type is null or points to a string that           //@E3
      //      isn't a valid event type identifier.                              //@E2
      //
      // Notes:
      // - In addition to the type identifier, the event header also contains
      //   the domain name of the initiating adapter, an event id in the form
      //   of an IAUniqueId, and an event timestamp.  The event id is generated
      //   automatically by this constructor.  The domain name and event time
      //   stamps are filled in (for the event header within a IATriggerEvent)
      //   by IAAdapter::notify().
      //
      // - Adapters may find it convenient to store and retrieve event-related
      //   based on the event's id; the event id generated for this header can
      //   be accessed using the getId() member function.

   void setStatus(IAEventCompletionStatus status);                              //@E4

   //--------------------------------------------------------------------------
   // Attribute access functions
   //
   // getId     = Accesses the event id in this header.
   // getType   = Returns the event type from this header.
   // getDomain = Returns the event domain name from this header.
   // getTime   = Returns the event timestamp from this header.
   //--------------------------------------------------------------------------

   const IAUniqueId & getId() const;
   const char * getType() const;

   // eventNumbers begin with 1 and are increasing
   long eventNumber() const;                                              //@E5

   const char * getDomain() const;
      // Note: Returns "" when called for an event header that has not
      // yet been completed by IAAdapter::notify().

   time_t getTime() const;
      // Note: Returns zero when called for an event header that has not
      // yet been completed by IAAdapter::notify().

   IAEventCompletionStatus getStatus() const;                                   //@E4

   IABool isValid() const;
                                                                              //#B@E1
   //--------------------------------------------------------------------------
   // Manipulators for rule-set selector
   //
   // getScope  = Set the scope token in the event header
   // getScope  = Returns the scope token in the event header
   //--------------------------------------------------------------------------

   void  setSelector(const char * selector);  // throw(IAError)           //@E3

   const char * getSelector() const;
      // Note: Returns "" when called for an event header for which
      // the setSelector function hasn't been called.
                                                                              //#E@E1

   //--------------------------------------------------------------------------
   // Debugging help
   //--------------------------------------------------------------------------

   void dbgInfo(ostream &o) const;


   //--------------------------------------------------------------------------
   // Methods intended for use only by the toolkit itself
   //--------------------------------------------------------------------------

   IAEventHeaderPvt * implPtr() const {return p;}

 private:
   IAEventHeaderPvt  *p;    // Impl-specific private object
};


//#############################################################################
//
// Class IATriggerEvent
//
// An IATriggerEvent object contains the data provided to Engines on
// IAAdapter::notify() calls: an event header for the event, plus a set
// of short term facts that are "inputs" for the processing of the event.
//
//#############################################################################

class IATriggerEventPvt;
class IAFactSet;

class IAEXPCLASS IATriggerEvent {                                               //@B1
 public:
   //--------------------------------------------------------------------------
   // Constructors, destructors and Assignment operator
   //--------------------------------------------------------------------------

   IATriggerEvent();
      // Trigger event built by default constructor is not valid for use in
      // a notify() call.  Use setType below to build a usable
      // trigger event.

   ~IATriggerEvent();

   IATriggerEvent(const IATriggerEvent &from);
   IATriggerEvent& operator =(const IATriggerEvent &rhs);

   void setFactSet(const IAFactSet &fs);                                        //@E2
      //
      // Input:
      //    fs =
      //       Set of short-term facts for the event.  These facts are used
      //       as event-specific "inputs" for the processing of this event.
      //
      //       When the this setter isn't used, the trigger event is given an
      //       empty fact set.  Facts also can be added by getting a reference
      //       to the trigger event's fact set using getFactSet() and then
      //       using IAFactSet::add().
      //
   void setType(const char *eventType); // throw(IAError)                 //@E3 //@E2
      //
      // Input:
      //    *eventType =
      //       String containing the domain-specific identifier for the type
      //       of event being generated, eg. "NewNews".  his arg is used in
      //       constructing the event header contained within the trigger
      //       event.  See comments above for the IAEventHeader class.
      // Exceptions                                                       //@E3 //@E2
      //      IAError.                                                          //@E2
      //      Thrown if type is null or points to a string that           //@E3
      //      isn't a valid event type identifier.                              //@E2

   //--------------------------------------------------------------------------
   // Extractor Methods
   //--------------------------------------------------------------------------

   IAEventHeader & getHeader() const;
   IAFactSet &     getFacts()  const;

   IABool isValid() const;

   //--------------------------------------------------------------------------
   // Debugging help
   //--------------------------------------------------------------------------

   void dbgInfo(ostream &o) const;

 private:
   IATriggerEventPvt *p;  // Impl-specific private object.
};

#endif // IATK_TRIGEVNT_H
