/*************************************************************************
** File:         sampmail.cpp
** Description:  Implementation of a sample email adapter
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
#ifdef __OS2__
   #include <netdb.h>      // Needed for TCP/IP gethostname() call
   #include <fstream.h>    // File I/O streams
#endif
#ifdef __WINDOWS__
   #include <winsock.h>    // Winsock
#endif
#ifdef _AIX
   #include <stdio.h>      // Needed for popen and fprintf
#endif
#include <sampmail.h>
#include <iexcbase.hpp>

// Define tokens for any sensors and effectors
const int SampMailAdapter::EMailFromAddressIs_Token = 1;
const int SampMailAdapter::EMail_Token              = 2;
const int SampMailAdapter::EMailWithSubject_Token   = 3;
const int SampMailAdapter::EMailRelayHostIs_Token   = 4;

// fromAddress is set by EMailFromAddressIs effector (OS/2, Windows)
I0String SampMailAdapter::fromAddress       = "";

// name for a temporary file to hold email to be sent (OS/2)
I0String SampMailAdapter::tempEmailFileName = "iagmtmp0.tmp";

// name of the host to use to relay message (Windows)
I0String SampMailAdapter::relayHost         = "";

// ----------------------------------------------------------------------
// Default constructor
// ----------------------------------------------------------------------
SampMailAdapter::SampMailAdapter():IAAdapter() {}

//-----------------------------------------------------------------------
// identify - Register our Sensors and Effectors
//
// Invoke registerProcedure for each Sensor and Effector that we provide,
// so that they may be invoked by the inference engine later.  If any of
// these attempts to register a procedure fails, return a false (zero)
// to indicate that our identify() failed; otherwise return true (non-zero).
//
// registerProcedure is a service method provided by the IAAdapter class
// and inherited by SampMailAdapter(); the adapter writer does not have
// to provide it.
//
// To add a new Sensor or Effector to this adapter:
// . Add a "<procName>_Token" constant to sampmail.h
// . Assign the new "<procName>_Token" a value in sampmail.cpp
// . Add a registerProcedure() call to SampMailAdapter::identify()
// . Add code to run sensor/effector to SampMailAdapter::answerQuery() or
//   SampMailAdapter::performAction(), as appropriate
//-----------------------------------------------------------------------
IABool SampMailAdapter::identify() {

   // Register our Sensors
   // No sensors to register

   // Register our Effectors
   registerProcedure("EMailFromAddressIs",  IAEffectorProc,
      "String", EMailFromAddressIs_Token);
   registerProcedure("EMail",               IAEffectorProc,
      "String String", EMail_Token);
   registerProcedure("EMailWithSubject",    IAEffectorProc,
      "String String String", EMailWithSubject_Token);
   registerProcedure("EMailRelayHostIs",    IAEffectorProc,
      "String", EMailRelayHostIs_Token);

   return true;                   // Success
}

//-----------------------------------------------------------------------
// performAction - Execute Effector procedure
//
// See the SampMailAdapter::identify() prolog for details on the steps
// required to add a new Effector.
//-----------------------------------------------------------------------
void SampMailAdapter::performAction(IAProcToken token,
   const IAEventHeader &hdr, const char *binding) {

   // Build the atom from the binding string
   IAAtom atom;
   atom.setKIF(binding);

   if (token == EMailFromAddressIs_Token)
      // For the EMailFromAddressIs effector,
      //    term 1 is the email address to use on the email's "From:" field
      fromAddress = atom.getStringTerm(1);

   else if (token == EMailRelayHostIs_Token)
      // For the EMailRelayHostIs effector,
      //    term 1 is the hostname of the SMTP relay machine
      relayHost = atom.getStringTerm(1);

   else if ((token == EMail_Token) || (token == EMailWithSubject_Token)) {
      // For EMail and EMailWithSubject effectors,
      //    term 1 is email address to send to
      //    term 2 is material to send
      // EMailWithSubject has a 3rd term
      //    term 3 is a subject line to send
      I0String itemToMail = "";
      if (token == EMailWithSubject_Token) {
         itemToMail += "Subject: ";
         itemToMail += atom.getStringTerm(3);
         itemToMail += "\n";
      }
      itemToMail += "\n**** Attached text forwarded by SampMail adapter ****\n\n";
      itemToMail +=  atom.getStringTerm(2);

#ifndef _AIX
      // On AIX, let sendmail determine the From: address.  For OS/2
      // and Windows, let's make sure we have one.
      if ((fromAddress == "") || (fromAddress.indexOf("@") == UINT_MAX)) {
         // if "From:" address hasn't been set by EMailFromAddressIs
         //   effector, set to "user@<local host name>"
         char* hname = new char[100];
         ::gethostname(hname,100);
         if (fromAddress == "") fromAddress = "user";
         fromAddress += "@";
         fromAddress += hname;
         delete [] hname;
      }
#endif

#ifdef __OS2__
      // Use OS/2's sendmail command to send a temporary file

      // Create the temporary file to email
      fstream tempMailFile;
      tempMailFile.open(tempEmailFileName, ios::out | ios::trunc);
      tempMailFile << itemToMail << flush;
      tempMailFile.close();

      // Send the piece of email
      int rc = ::system("sendmail -af " + tempEmailFileName +
                        " -f " + fromAddress + " " + atom.getStringTerm(1));
      if (rc) {
          cerr << "Unable to EMail to " << atom.getStringTerm(1)
               << "; sendmail rc is " << rc << "." << endl;
          cerr << "SampMail effector failed . . ." << endl;
      }

      // Dump the temporary file
      ::system("erase "+tempEmailFileName);

#elif _AIX
      // Use AIX's sendmail command to send

      // Build the sendmail command string, including recipient
      I0String smCmd("sendmail -t ");
      smCmd += atom.getStringTerm(1);

      // Use popen to run sendmail, piping the item to mail to it
      FILE* fp = popen(smCmd,"w");
      fprintf(fp,itemToMail);
      fprintf(fp,"\n.\n");   // Mark end of item
      int rc = pclose(fp);

      if (rc) {
          cerr << "Unable to EMail to " << atom.getStringTerm(1)
               << "; sendmail rc is " << rc << "." << endl;
          cerr << "SampMail effector failed . . ." << endl;
      }
      else cout << "Mail sent to " << atom.getStringTerm(1)
                << " by SampMail adapter." << endl;

#elif __WINDOWS__
      // Conduct an SMTP dialog with an SMTP relay host to send
      // the mail.

      if (relayHost=="") { // Need a relay host to send mail to
         cerr << "Unable to send mail without establishing a relay "
              << "system with EMailRelayHostIs effector." << endl
              << "SampMail effector failed . . ." << endl;
         return;
      }

      // Get a socket descriptor
      int sock = ::socket(AF_INET, SOCK_STREAM, 0);
      if (sock == -1)  {
         cerr << "Socket creation failed, error code = "
              << WSAGetLastError()  << "." << endl
              << "SampMail adapter failed . . ." << endl;
         return;
      }

      // Build address structure for relayHost
      struct sockaddr_in sa;        // Address structure for relayHost
      memset(&sa, 0, sizeof sa);    // Clear address structure
      sa.sin_family = AF_INET;      // Address family -- Internet
      sa.sin_port = htons(25);      // Use SMTP port number
      struct hostent *nameent;      // For gethostbyname() result
      nameent = ::gethostbyname(relayHost);
      if (!nameent) {
         cerr << "Unable to resolve host name " << relayHost
              << " from EMailRelayHostIs effector, " << endl
              << "Error code = "
              << WSAGetLastError() << "." << endl
              << "SampMail adapter failed . . ." << endl;
         return;
      }
      sa.sin_addr = *(struct in_addr *) nameent->h_addr; // Host addr

      // Connect to relayHost
      if (::connect(sock, (sockaddr*)&sa, sizeof sa))  {
         cerr << "Socket connect failed, error code = "
              << WSAGetLastError() << "." << endl
              << "SampMail adapter failed . . ." << endl;
         return;
      }

      int bufSize = 250;                // buffer size
      char* buf=new char[250];          // buffer for response

      try {  // Conduct SMTP dialog
         // Get connect message
         sendCommand("RECV ONLY", sock, "220", buf, bufSize);

         // Identify ourself
         I0String fromHost =
                       fromAddress.subString(fromAddress.indexOf("@")+1);
         sendCommand("HELO "+fromHost+"\n", sock, "250", buf, bufSize);

         // Whom mail is from
         sendCommand(I0String("MAIL From:<")+fromAddress+">\n",
                     sock, "250", buf, bufSize);

         // Whom mail is to
         sendCommand(I0String("RCPT To:<")+atom.getStringTerm(1)+">\n",
                     sock, "250 251", buf, bufSize);

         // Send the piece of mail
         sendCommand("DATA\n", sock, "354", buf, bufSize);
         sendCommand(itemToMail+"\n.\n", sock, "250", buf, bufSize);

         // End dialog
         sendCommand("QUIT\n", sock, "221", buf, bufSize);
         cout << "Mail sent to " << atom.getStringTerm(1)
              << " by SampMail adapter." << endl;
      }

      catch(int i) {
         cerr << "SampMail adapter failed:  Error " << i
              << " thrown by SampMailAdapter::sendCommand()." << endl;
      }

      delete [] buf;               // Don't need buffer anymore
      ::closesocket(sock);         // Close socket, ignoring errors
#else
      cerr << "The SampMail adapter does not support this platform."
           << endl << "SampMail adapter failed . . ." << endl;
#endif
   }

   else ITHROW(IInvalidParameter("SampMailAdapter::performAction() called"
            " with an unrecognized token: " + I0String(token), 0));

   return;
}

//-----------------------------------------------------------------------
// sendCommand
//   - Send a command (cmd) on the requested socket (sock) and store the
//     response in a buffer (buf of size bufSize).  If the response does
//     not begin with one of the expected response codes (expectedCodes),
//     throw the response code received.  Otherwise, return 0.
//-----------------------------------------------------------------------
int SampMailAdapter::sendCommand(const I0String &cmd,
                                 int sock,
                                 const I0String &expectedCodes,
                                 char* buf,
                                 int bufSize) throw(int) {

// sendCommand function is only used/needed on Windows.
#ifdef __WINDOWS__

   int rc = 0;

   // "RECV ONLY" value for cmd means no command to send.  Otherwise,
   // send the command on the socket.
   if (cmd != "RECV ONLY") {
      // Send the command
      rc = ::send(sock,(char *)cmd,cmd.length(),0);
      if (rc < 0) {
         cerr << "Socket send failed, error code = "
              << WSAGetLastError() << "." << endl;
         throw rc;
      }
   }

   // Receive the response from the socket and place in buf
   rc = ::recv(sock,buf,bufSize,0);
   if (rc <= 0) {
      cerr << "Socket recv failed, error code = "
           << WSAGetLastError() << "." << endl;
      throw -1;
   }

   // If code from SMTP response does not appear in list of expected
   // responses, throw an error.
   //
   // Single-line response = <3-digit-response-code><space >...<crlf>
   //
   // Multi-line  response = <3-digit-response-code><hyphen>...<crlf>
   //                        <3-digit-response-code><hyphen>...<crlf>
   //                        <3-digit-response-code><space >...<crlf>

   I0String response(buf);   // convert to I0String for parsing
   if (expectedCodes.wordIndexOfPhrase(response.subString(0,3)) == 0) {
      cerr << "Bad response to " << cmd
           << "Response received: " << response.subString(0,rc);
      throw (int) response.asInt();  // throw just response code
   }

#endif

   return 0;
}

//=======================================================================
// Global function newAdapter()
//
// Function newAdapter() is a "factory" function for creating new
// instances of SampMail Adapters.
//
// Input:
//    *domain  = Domain name that identifies the type of concrete
//               Adapter requested.  Only "SampMail" is recognized.
//    *parms   = Adapter start-up parameters obtained from the Agent
//               configuration file.  These are not used by SampMail.
//
// The adapter instance created by a call to this function is owned by
// the IATK framework and should NOT ever be deleted by code in the
// adapter module.
//=======================================================================
IAAdapter * newAdapter(const char *domain, void *parms) {

   if (strcmp(domain, "SampMail")) {
      ITHROW(IInvalidParameter("newAdapter() does not support domain "
         + I0String(domain) + "; only \"SampMail\" is valid.", 0));
   }

   IAAdapter *adapter = new SampMailAdapter();

   if (!adapter) {
      ITHROW(IException("newAdapter() is unable to create a new"
         " SampMailAdapter object."));
   }
   return adapter;
}

#ifdef __WINDOWS__
/* ------------------------ TCPIPinit class ------------------------ */
class TCPIPinit {
 private:
   int init;
 public:
   TCPIPinit();
   ~TCPIPinit() {if (init) WSACleanup();}
   };

TCPIPinit::TCPIPinit(): init(0)
   {
   WORD requestedVersion = MAKEWORD(1, 1);
   WSADATA wsaData;
   if (WSAStartup(requestedVersion, &wsaData)) {
      cerr << "The available sockets DLL only supports higher"
           << " levels of the WINSOCK API than those supported by the"
           << " SampMail adapter." << endl;
      return;
      }
   if (wsaData.wVersion != requestedVersion) {
      WSACleanup();
      cerr << "The available sockets DLL does not support the WINSOCK API"
           << " version required by the SampMail adapter." << endl;
      return;
      }
   init = 1;
   }

// Static object to initialize/cleanup windows socket environment
static TCPIPinit initobj;

#endif

