/*************************************************************************
** File:         iagent.cpp
** Description:  Implementation of a sample agent
**               Main agent program to read configuration file,
**               connect engines and adapters, and start the agent running.
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/

#include <iapl/common.h>

#include <iostream.h>
#include <fstream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <i0string.hpp>
#include <signal.h>
#include <ctype.h>
#include <locale.h>
#include <time.h>
#include <errno.h>

#include <iexcept.hpp>           // For catching VisualAge exceptions
#include <itrace.hpp>            // For tracing

#include <iatk/agent.h>
#include <iatk/iaerrors.h>
#include <iatk/handles.h>

//#include <iapl/astring.h>         // Wrapper on VisualAge C++ I0String class
#include <iapl/serializ.h>        // Serialization class (semaphores, mutexes, etc.)
#include <iapl/athreads.h>        // Threading class
#include <iapl/aexcept.h>         // Exception class to throw

#ifdef IA_UNIX
   #include <unistd.h>           // For isatty()
#endif
#ifdef IA_OS2
   #include <io.h>               // For isatty()
   #define INCL_DOSERRORS
   #include <os2.h>
#endif
#ifdef IA_WIN32
   #include <io.h>               // For isatty()
#endif
#ifdef IA_AIX
#include <new.h>
#endif

// Classes for internal use
class config_file_parser {
 public:
   config_file_parser();
   config_file_parser(const char *filename);
   void read(const char *filename);

 private:
   int recno;
   const char *filename;

   void cfg_error(const char *error);
   void parse_engine(AString &line);
   void parse_adapter(AString &line);
   void parse_library(AString &line);
   void parse_conductset(AString &line);
   };

// IAAgent has pure virtual functions, so we must inherit from the class
// and "implement" them.  Since we don't need to make our agent do
// anything new, our implementation is minimal.
class Agent : public IAAgent {
   public:
   Agent() {}
   ~Agent(){}
   };

// List of engines we are configuring
static struct cfg_engine {
   AString type;                 // Engine type
   AString lmodname;             // Name of loadable module for engine
   AString parms;                // Parameter string for engine
   IAEngineHandle handle;        // Engine handle
   cfg_engine *next;             // Pointer to next engine in chain
   } *engine_list = NULL, *engine_list_last;

// List of adapters we are configuring
static struct cfg_adapter {
   AString domain;               // Adapter domain name
   AString lmodname;             // Name of loadable module for adapter
   AString parms;                // Parameter string for adapter
   IAAdapterHandle handle;       // Adapter handle
   cfg_adapter *next;            // Pointer to next adapter in chain
   } *adapter_list = NULL, *adapter_list_last;

// Info about the library we are configuring
static struct cfg_library {
   AString type;                 // Library DLL type (type of library system)
   AString dllname;              // Library DLL name
   AString map;                  // Name of library map
   AString collector;            // Name of top collector
   int levels;                   // Number of collector levels
   } *library = NULL;

// List of inferencing sets we are configuring
static struct cfg_conduct {
   AString engtype;              // Engine type this conduct set applies to
   AString name;                 // Conduct set name
   AString selector;             // Selector associated with set
   cfg_conduct *next;            // Pointer to next adapter in chain
   } *conduct_list = NULL, *conduct_list_last;

// Prototypes for internal functions
static void parse_options(int argc, char **argv);
static void check_definitions();
static void load_engines();
static void load_adapters();
static void make_connections();
static void load_library();
static void load_conduct_sets();
static void wait_for_end();
static long accept_commands(void *);
static void process_command(const char *command);
static void handler(int sig);
static void wait_end_signal(Bool givemsg);
static long monitor_stats(void *arg);
static void sayrate();
static void declare_end();

static void loadCS(AString csname,AString selector);
static void unloadCS(AString csname,AString selector);
static void reset(AString selector);
IAEngineHandle our_engine;

static int getNextPossiblyQuotedParm(AString &fromString, AString &intoString);
static int indexOfQuote(AString &inString, int startpos=0);
static void displayHelp(void);


#ifdef IA_AIX
void failed_new_handler();
#endif


// Structure to hold command-line options
static struct Opts {
   AString filename;             // Filename to use as config file
   Bool check;                   // Just check config file
   Bool daemon;                  // Run as background daemon
   Bool verbose;                 // Print registry information
   Bool trace;                   // Enable run-time tracing
   Bool monitor;                 // Enable agent throughput monitoring

   Opts() :                      // Initialize options to defaults
      filename("iagent.cfg"),    // Default config file name
      check(false),              // Default is to run agent after checking
      daemon(false),             // Run as background daemon
      verbose(false),            // Don't print registry information
      trace(false),              // Don't trace internal events
      monitor(false)             // Don't monitor throughput
         {}
   } opts;

static Agent agent;              // Create the agent object
static IALibraryHandle libHdl;   // Create a handle for the agent's library

static AMutex stop_mx;           // Mutex to serialize the stop_agent flag
static ACondVar stop_cv(stop_mx);   // Condition var signalled when stop_agent
                                    // is changed by a signal handler.
static Bool stop_agent;          // Flag that says agent should be stopped

static time_t started_at;        // Time that we started the agent

int realmain(int argc, char **argv) {

   // Initialize the locale settings to the user's preference.  Do this first
   // because it might affect parsing.
   char *loc = setlocale(LC_ALL, "");
#ifndef IA_OS2    // VisualAge on OS/2 returns a null pointer when it shouldn't
   if (!loc) {perror("Error establishing locale"); return 1;}
#endif

   // Parse the options from the command line
   parse_options(argc, argv);
   // The opts structure now defines defaulted or specified options.

   // Activate tracing if we've been requested to
   if (opts.trace) {
      ITrace::writeToStandardOutput();    // Set trace output to stdout
      ITrace::enableTrace();              // Turn on tracing
      }

   // Now parse the specified or defaulted configuration file.
   config_file_parser cfg(opts.filename);

   // Make sure every engine and every adapter have at least one connection
   check_definitions();

   // Add the engines to the agent
   load_engines();

   // Add the adapters to the agent
   load_adapters();

   // Connect engines to adapters as appropriate
   make_connections();

   // Print registry information, if desired by user.  It's best to do this
   // before loading the conduct set so that user has a reference of the
   // agent registry to use when debugging conduct-set parsing problems.
   if (opts.verbose) agent.dbgInfo(cout);

   // Load the library configuration
   load_library();

   // Load the conduct sets
   load_conduct_sets();

   // Stop now if we're just checking the configuration file
   if (opts.check) {
      cout << "Configuration check is successful" << endl;
      return 0;
      }

   // Start the agent
   try {
      agent.start();
      }
   catch(IAError &e) {
      cerr << e.errormsg() << endl;
      cerr << "Error starting agent." << endl;
      return 10;
      }
   started_at = time(NULL);

   // Start statistics monitoring thread if requested
   if (opts.monitor)
      AThread stats_thread(monitor_stats, NULL);

   // Wait until the agent should be stopped
   wait_for_end();

   // Shutdown the agent engines and adapters

   try {
      agent.stop();
      }
   catch(IAError &e) {
      cerr << e.errormsg() << endl;
      cerr << "Error stopping agent." << endl;
      return 10;
      }


   return 0;
   }

int main(int argc, char **argv) {
   int status;                            // For return status
#ifdef IA_AIX
   set_new_handler(&failed_new_handler);  // Establish a new handler
#endif
   try {
      status = realmain(argc, argv);      // Do actual processing
      }
   catch (const IAError &exc) {           // Catch things derived from IAError
      AString me = AThreadId::current().format() + ": ";
      cerr <<'\n'<<me<< "Caught an exception that is derived from IAError.\n";
      cerr <<me<< "The error message is:\n";
      cerr <<me<< ' ' << exc.errormsg() << endl;
      status = 201;
      }
   catch (const AException &exc) {        // Catch stuff derived from AException
      AString me = AThreadId::current().format() + ": ";
      cerr <<'\n'<<me<<"Caught an exception that is derived from AException.\n";
      cerr <<me<< "Below is the exception information.\n";
      cerr <<me<< "Exception type is:  " << exc.name() << '\n';
      cerr <<me<< "Message is:\n";
      cerr <<me<<' '<< exc.getErrorText() << endl;
      cerr <<me<< "Location is:  ";
      const AExceptionLocation loc = exc.getLocation();
      const char *p = loc.getFileName();
      if (p) cerr << "file " << p << "; ";
      p = loc.getFunctionName();
      if (p) cerr << "function " << p << "; ";
      unsigned long num = loc.getLineNumber();
      if (num) cerr << "line " << num;
      cerr << endl;
      status = 202;
      }
   catch(const IException &exc) {         // Catch VisualAge exception hierarchy
      int i, msgs=exc.textCount();
      AString me = AThreadId::current().format() + ": ";
      cerr <<'\n'<<me<<"Caught an exception that is derived from IException.\n";
      cerr <<me<< "Below is the exception information.\n";
      cerr <<me<< "Exception type is:  " << exc.name() << '\n';
      cerr <<me<< "Messages are:\n";
      for (i=msgs-1; i>=0; --i) cerr <<me<<' '<< exc.text(i) << endl;
      cerr <<me<< "Location is:  ";
      const IExceptionLocation *loc = exc.locationAtIndex(0);
      const char *p = loc->fileName();
      if (p) cerr << "file " << p << "; ";
      p = loc->functionName();
      if (p) cerr << "function " << p << "; ";
      unsigned long num = loc->lineNumber();
      if (num) cerr << "line " << num;
      cerr << endl;
      status = 203;
      }
   catch (...) {                          // Catch all other exceptions
      AString me = AThreadId::current().format() + ": ";
      cerr <<me<< "Unexpected error:  Caught an exception not derived from\n";
      cerr <<me<< "any supported exception class.  Terminating process." <<endl;
      status = 204;
      }
   return status;
   }

// Parse options from the command line
void parse_options(int argc, char **argv) {
   int i;
   char oc, *p;

   // Options are initialized to defaults already by Opts constructor

   // Start parse command-line options.  Sadly, we can't assume we have
   // the getopt() library routine available.
   for (i=1; i<argc; ++i) {      // For each argument
      if (*argv[i] != '-' && *argv[i] != '/') break;  // End of options
      p = argv[i]+1;             // Start of option
      while (*p) {               // For each option letter
         switch (oc = *p++) {    // Decode option
            case 'c':            // Option "check"
               opts.check = true; break;
            case 'd':            // Option "daemon"
               opts.daemon = true; break;
            case 'v':            // Option "verbose"
               opts.verbose = true; break;
            case 't':            // Option "trace"
               opts.trace = true; break;
            case 'm':            // Option "monitor" (undocumented)
               opts.monitor = opts.daemon = true; break;
                                 // -m implies -d, since prompt would scroll
                                 // off the screen anyway.
            case '?':            // Option for usage information
               cout << "Usage:  iagent [options]\n"
                      " options:   -fFILENAME  name of configuration file.\n"
                      "                        Default is iagent.cfg.\n"
                      "            -c          Check configuration file,\n"
                      "                        but don't start agent.\n"
                      "            -d          Run as background program.\n"
                      "            -v          Print agent registry.\n"
                      "            -t          Activate internal tracing.\n"
                      "            -m          Activate throughput "
                                                               "monitoring.\n"
                      "            -?          Print usage information.\n";
               exit(0);          // Quick exit after giving usage info

            case 'f':            // Option "file" to name configuration file
               if (!*p) {        // Filename doesn't immediately follow -f
                  if (++i >= argc) {
                     cerr << "Configuration filename is missing." << endl;
                     exit(2);    // Return an error
                     }
                  p = argv[i];   // Point to file name
                  }
               opts.filename = p;   // Save filename
               p = strchr(p, '\0'); // Terminate option parsing for this arg
               break;

            default:
               cerr << "Unknown option \"" << oc << "\" specified.  "
                    << "Issue \"iagent -?\" for usage information." << endl;
               exit(2);          // Exit with option syntax error
            }
         }
      }

   // Done parsing options
   if (i>=argc) return;          // No more argument left

   // argv[i] is the first positional parameter.
   // Presently, we accept no positional parameters.
   cerr << "Parameter \"" << argv[i] << "\" is unrecognized.  "
        << "Issue \"iagent -?\" for usage information." << endl;
   exit(2);
   }

// -------------------- Code to parse configuration file ---------------------
config_file_parser::config_file_parser() {}

config_file_parser::config_file_parser(const char *filename) {
   read(filename);
   }

void config_file_parser::read(const char *ifilename) {
   // A filename of "-" says to use standard input.  Check for that.
   istream *cfgstream;           // Will point to input stream object
   ifstream cfgfile;             // For config file, if it's a regular file

   filename = ifilename;
   if (!strcmp(filename, "-")) cfgstream = &cin;
   else {                        //  It's a regular file
      cfgfile.open(filename);    // Open configuration file
      if (!cfgfile) {
         fprintf(stderr, "Unable to open configuration file %s: %s\n",
                                                   filename, strerror(errno));
         exit(3);
         }
      cfgstream = &cfgfile;      // Use this file for input
      }

   for (recno=1; *cfgstream; ++recno) {   // Read configuration file
      AString line = AString::lineFrom(*cfgstream);   // Read one line
      if (!line.length() || line[0]=='#' || line[0]=='*') continue;
                                 // Ignore comments and blank lines
      if (isspace(line[0])) cfg_error("Keyword must start in column 1");

      AString keyword = line.word(1).lowerCase();
                                 // Keyword defining type of line
      unsigned int w2 = line.indexOfWord(2);
      if (w2!=-1) line.remove(0, w2);
      line.strip();              // Now line contains second and rest of words

      // Route according to the type of line
      if      (keyword=="engine")     parse_engine(line);
      else if (keyword=="adapter")    parse_adapter(line);
      else if (keyword=="library")    parse_library(line);
      else if (keyword=="conductset") parse_conductset(line);
      else if (keyword=="run") break;  // Ignore rest of config file
      else cfg_error(AString("Unrecognized keyword \"")+keyword+"\"");
      }
   }

/* Print error message and exit program */
void config_file_parser::cfg_error(const char *error) {
   if (!strcmp(filename, "-")) cerr << "Input error: ";
   else cerr << "Error on line " << recno << " of configuration file "
             << filename << ":\n ";
   cerr << error << "." << endl;
   exit(3);
   }

// Parse "engine" statements
// format is:      engine type dllname [parms]
void config_file_parser::parse_engine(AString &line) {
   if (line.numWords() < 2)
      cfg_error("engine statement format is:  engine type dllname [parms]");
   cfg_engine *eng = new cfg_engine;   // Get storage to remember engine info
   if (!eng) {cerr << "Unable to acquire free storage." << endl; exit(9);}
   eng->type = line.word(1);           // First argument is engine type
   eng->lmodname = line.word(2);       // Second arg is name of loadable module
   // Isolate parameter string
   unsigned int parms = line.indexOfWord(3);
   if (parms == -1) parms = line.length();
   line.remove(0, parms);
   line.strip();
   eng->parms = line;                  // Save away parameter string
   eng->next = NULL;                   // Will be last in list

   // Append this engine description to the list of engines
   if (!engine_list) engine_list=eng;  // First engine
   else engine_list_last->next = eng;  // Append new engine to end
   engine_list_last = eng;             // Make this the last engine
   }

// Parse "adapter" statements
// format is:      adapter domain dllname [parms]
void config_file_parser::parse_adapter(AString &line) {
   if (line.numWords() < 2)
      cfg_error("adapter statement format is:  adapter domain dllname [parms]");
   cfg_adapter *adp = new cfg_adapter; // Get storage to remember adapter info
   if (!adp) {cerr << "Unable to acquire free storage." << endl; exit(9);}
   adp->domain = line.word(1);         // First argument is adapter domain name
   adp->lmodname = line.word(2);       // Second arg is name of loadable module
   // Isolate parameter string
   unsigned int parms = line.indexOfWord(3);
   if (parms == -1) parms = line.length();
   line.remove(0, parms);
   line.strip();
   adp->parms = line;                  // Save away parameter string
   adp->next = NULL;                   // Will be last in list

   // Append this adapter description to the list of adapters
   if (!adapter_list) adapter_list = adp; // First adapter
   else adapter_list_last->next = adp; // Append new adapter to end
   adapter_list_last = adp;            // Make this the last adapter
   }

// Parse "library" statement
// format is:      library type dllname map top-collector-name #-collector-levels
void config_file_parser::parse_library(AString &line) {
   if (library)
      cfg_error("Only one library statement allowed");
   if ((line.numWords() < 4) || (line.numWords() > 5))
      cfg_error("library statement format is:  "
                           "library libtype dllname map collector levels");
   library = new cfg_library;           // Get storage to remember library info
   if (!library) {cerr << "Unable to acquire free storage." << endl; exit(9);}
   library->type      = line.word(1);   // First arg is type of library
   library->dllname   = line.word(2);   // Second arg is name of dll
   library->map       = line.word(3);   // Third argument is library map name
   if ((library->map)[(library->map).length()-1] != FILEPATH_SEPCHR)
      library->map += FILEPATH_SEP;   // Append final path separator
   library->collector = line.word(4);   // Fourth argument is collector name
   if (line.numWords() == 5)            // If levels was specified...
      library->levels = (line.word(5)).asInt(); // 5th arg= # collector levels
   else library->levels = 2;            // Otherwise default to 2 levels
   }

// Parse "conductset" statements
// format is:      conductset engine-type name selector
//
// selector is optional
//
// use double quotes around name or selector if they contain imbedded blanks
//    (e.g. conductset raise "conduct set name" "my conduct set")
//
// use a pair of double quotes to imbed a single double-quote within a quoted
//    name or selector. (e.g. conductset raise mycs "John ""Ace"" Doe"  passes
//    selector of   John "Ace" Doe    to the engine).
//
void config_file_parser::parse_conductset(AString &line) {
   AString csline = line;

   if (csline.numWords() < 2)
      cfg_error("conductset statement format is:  conductset engine-type "
                                                 "set-name selector-name");
   cfg_conduct *con = new cfg_conduct;  // Get storage to remember set info
   if (!con) {cerr << "Unable to acquire free storage." << endl; exit(9);}

   con->engtype = csline.word(1);       // First argument is engine type

   // Take a fast path through here if no quotes contained in conductset stmt
   if (csline.indexOf('\"') == -1) {    // No quotes anywhere in statement
      if (csline.numWords() > 3)
         cfg_error("conductset statement format is:  conductset engine-type "
                                                    "set-name selector-name");
      con->name = csline.word(2);       // Second word is conduct set name
      con->selector = csline.word(3);   // Third word is conduct set name
      }
   // Take the slow path to figure out/parse quoted parameters
   else {                               // At least one quoted parameter...
     csline.remove(0,csline.indexOfWord(2)); // Remove previously processed parms

     if (getNextPossiblyQuotedParm(csline,con->name))
        cfg_error("conductset set-name is missing ending quote");
     if ((con->name).length() == 0)
        cfg_error("conductset set-name (null) is invalid");
     if (csline[0] != ' ')              // No blank between parms
        cfg_error("conductset set-name has extraneous data following ending quote");
     while ((csline.length() > 0) && (csline[0] == ' ')) { // Remove blanks up to next parm
        csline.remove(0,1); }
     if (csline.length() >0) {          // If selector specified
        if (getNextPossiblyQuotedParm(csline,con->selector))
           cfg_error("conductset selector-name is missing ending quote");
        csline.strip();                 // Remove ending white space
        if (csline.length() > 0)        // If extraneous data, error
           cfg_error("conductset statement has extraneous data following selector-name");
        }
     }

   con->next = NULL;                    // Will be last in list

   // Append this conduct set description to the list
   if (!conduct_list) conduct_list=con; // First conduct set
   else conduct_list_last->next = con;  // Append new conduct set to end
   conduct_list_last = con;             // Make this the last conduct set
   }
// ------------------ End of code to parse configuration file -----------------

// Make sure configurations are correct and consistent
void check_definitions() {
   // Make sure at least one engine and one adapter have been configured
   if (!engine_list) {cerr << "No engine is configured." << endl; exit(4);}
   if (!adapter_list) {cerr << "No adapters are configured." << endl; exit(4);}
   if (!library) {cerr << "No library is configured." << endl; exit(4);}

   // Make sure that the engine types specified on conductset statement all
   // refer to engines which have been configured.
   if (!conduct_list) return;
   cfg_conduct *con;
   cfg_engine *eng;
   for (con=conduct_list; con; con=con->next) {
      for (eng=engine_list; eng && eng->type!=con->engtype; eng=eng->next) ;
      if (!eng) {
         cerr << "Conduct set with name \"" << con->name << "\" refers to "
              << "engine type \"" << con->engtype << "\",\n"
              << "which doesn't match the type of any configured engine."
              << endl;
         exit(4);
         }
      }

   // Make sure every engine and every adapter have at least one connection.
   // Happens for free now, since we connect every engine with every adapter
   }

// Add engines to the agent
void load_engines() {
   cfg_engine *eng;

#if 0    // For internal testing, don't make this check yet
   // For now, make sure only one engine is configured
   if (engine_list.next)
      {cerr << "Only one engine can be configured." << endl; exit(5);}
#endif

   for (eng=engine_list; eng; eng=eng->next) {
      try {
         agent.addEngine(eng->type, eng->lmodname, eng->parms, eng->handle);
         }
      catch(IAError &e) {
         cerr << e.errormsg() << endl;
         cerr << "Error loading engine with type = " << eng->type <<
                 ", loadable module name = " << eng->lmodname << '.' << endl;
         exit(5);
         }
      our_engine = eng->handle;
      }
   }

// Add adapters to the agent
void load_adapters() {
   cfg_adapter *adp;
   for (adp=adapter_list; adp; adp=adp->next)
      try {
         agent.addAdapter(adp->domain, adp->lmodname, adp->parms,
                                                      adp->handle);
         }
      catch(IAError &e) {
         cerr << e.errormsg() << endl;
         cerr << "Error loading adapter with domain = " << adp->domain <<
                 ", loadable module name = " << adp->lmodname << '.' << endl;
         exit(5);
         }
   }

// Connect engines to adapters as appropriate
void make_connections() {
   // For now, connect all engines to all adapters
   cfg_engine *eng;
   cfg_adapter *adp;

   for (eng=engine_list; eng; eng=eng->next)
      for (adp=adapter_list; adp; adp=adp->next)
         try {
            agent.connect(adp->handle, eng->handle);
           }
         catch(IAError &e) {
           cerr << e.errormsg() << endl;
           cerr << "Error connecting adapter to engine:\n"
                << " engine:  type   = " << eng->type <<
                        ", loadable module name = " << eng->lmodname << '\n'
                << " adapter: domain = " << adp->domain <<
                        ", loadable module name = " << adp->lmodname << endl;
            exit(5);
            }
   }

// Load the library configuration
void load_library() {
#if defined(IA_OS2) || defined(IA_WIN32)
   library->dllname.lowerCase();
#endif
   if (!library->dllname.size()) {
      cerr << "The library DLL name must be specified as \"iaglibf\"." << endl;
      exit(6);
      }
   agent.addLibrary(library->map,library->collector,       // In: map, top
                    library->dllname,library->type,        // In: dll, type
                    library->levels,                       // In: levels
                    libHdl);                               // Out: lib handle
   }

// Load the conduct sets into the appropriate engines
void load_conduct_sets() {
   IAError rc;
   cfg_conduct *con;
   cfg_engine *eng;
   if (!conduct_list) return;
   for (con=conduct_list; con; con=con->next)
      for (eng=engine_list; eng; eng=eng->next)
         if (con->engtype == eng->type) {
            try {
               agent.loadConductSet(libHdl,
                                    eng->handle,
                                    con->name,
                                    con->selector);
               }
            catch (IAError &e) {
               cerr << e.errormsg() << endl;
               cerr<< "Error loading conduct set with name = " << con->name <<endl;
               exit(5);
               }
            }
   }

// While the agent is running, we wait.  If standard input is a tty, and if the
// daemon option wasn't specified, we allow the user to enter commands.  See
// the list below for the recognized commands.
void wait_for_end() {
   // Set up a signal handler so we can perform an orderly shutdown if receive
   // SIGINT (ctl-C), SIGTERM (a "kill" operation), or SIGHUP (shutdown).
   stop_agent = false;           // Initialize flag saying agent should run
                                 // We don't need the mutex to touch this now,
                                 // because the signal handler can't be running.
#if !defined(IA_VAC) || defined(IA_UNIX)
   if (SIG_ERR==signal(SIGINT, handler) || SIG_ERR==signal(SIGTERM, handler)
#else
   // OS/2 & Win32 VisualAge compilers have bug re typedefs that forces us to:
   if (SIG_ERR==signal(SIGINT,  (_SigFunc) handler) ||
       SIG_ERR==signal(SIGTERM, (_SigFunc) handler)
#endif
#ifdef SIGHUP
       || SIG_ERR==signal(SIGHUP, handler)
#endif
      ) {perror("Error establishing signal handler"); exit(7);}

   // See if we should do an interactive wait or a non-interactive wait
   if (opts.daemon || !isatty(0)) {
      wait_end_signal(true);     // Wait for terminating signal
      return;
      }

#ifdef IA_WIN32
   // On Windows, we avoid reading commands from the initial thread, because
   // if we catch a signal (e.g., SIGINT), the signal handler would get control
   // and set stop_agent=true, but Windows won't interrupt the outstanding
   // read from stdin.  We therefore make sure the initial thread is merely
   // waiting for a termination request, and read commands from another thread.
   // When the initial thread wakes up and returns from main(), the thread from
   // which we're reading commands will simply die.
   AThread do_cmds(accept_commands, NULL);   // Create thread to read commands
   wait_end_signal(false);       // Wait for terminating signal
#else
   accept_commands(NULL);        // Read and process commands
#endif

   // Stop catching signals
   signal(SIGINT, SIG_DFL); signal(SIGTERM, SIG_DFL);
#ifdef IA_UNIX
   signal(SIGHUP, SIG_DFL);
#endif
   }

// Read and process commands from stdin
long accept_commands(void *) {
   // Interactive wait.  Write a header and a prompt for each input line.
   // Accept the following commands:
   displayHelp();
   while (!stop_agent) {
      // We use ANSI I/O below because with C++ I/O, there no way to tell the
      // difference between EOF and being interrupted by a signal.
      fputs("\niagent> ", stdout); fflush(stdout);
      char input[200];           // Input area for command
      size_t len;
      if (fgets(input, sizeof input, stdin)) {
         if ((len=strlen(input)) && input[len-1]=='\n') input[len-1]=='\0';
                                 // Eliminate trailing newline
         process_command(input); // Handle command
         }
      else
         if (feof(stdin)) {
            cout << "Received EOF on standard input." << endl;
            wait_end_signal(true);  // Wait for terminating signal
            return 0;
            }
         else {
            #ifdef IA_OS2
               if (errno==EOS2ERR && _doserrno==ERROR_INTERRUPT) errno=EINTR;
            #endif
            if (errno!=EINTR) perror("Error reading command");
            }
      }
   return 0;
   }

// Non-interactive wait.  Loop on a condition variable until the signal
// handler wakes us up.
void wait_end_signal(Bool givemsg) {
   if (givemsg) cout << "Waiting for signal to terminate." << endl;
   stop_mx.lock();         // Acquire mutex, which serializes stop_agent
   while (!stop_agent) stop_cv.wait(); // Wait until stop_agent gets set
   stop_mx.unlock();       // No longer need mutex
   }

// Process a command input while the agent is running
void process_command(const char *input) {
   AString command(input);
   AString verb = command.word(1).lowerCase();  // Isolate command verb
   if ((verb=="quit") || (verb=="stop") || (verb=="end") ) declare_end();
   else if (verb=="stats") sayrate();
   else if ((verb=="help") || (verb=='\?')) displayHelp();
   else if (verb=="reset") {
        AString selector;
        command.remove(0,command.indexOfWord(2)); // Remove previously processed parms
        if (!getNextPossiblyQuotedParm(command,selector)) {
          reset(selector);
          }
       else {
          cout << "Command format is " << verb << " selector" <<endl;
          }
      }
   else if ((verb=="loadcs") || (verb=="unloadcs")) {
        if (command.numWords() >= 2) {
           AString csname;
           AString selector;
           // Take a fast path through here if no quotes contained in conductset stmt
           if (command.indexOf('\"') == -1) {   // No quotes anywhere in statement
              csname = command.word(2);         // Second word is conduct set name
              selector = command.word(3);       // Third word is conduct set name
              if (verb == "loadcs") loadCS(csname,selector);
              if (verb == "unloadcs") unloadCS(csname,selector);
              }
           // Take the slow path to figure out/parse quoted parameters
           else {                               // At least one quoted parameter...
             command.remove(0,command.indexOfWord(2)); // Remove previously processed parms
             if (!getNextPossiblyQuotedParm(command,csname)) {
                while ((command.length() > 0) && (command[0] == ' ')) { // Remove blanks up to next parm
                   command.remove(0,1); }
                if (!getNextPossiblyQuotedParm(command,selector)) {
                  if (csname.length() > 0) {
                    if (verb == "loadcs") loadCS(csname,selector);
                    if (verb == "unloadcs") unloadCS(csname,selector);
                    }
                  else
                   cout << "Conduct set name cannot be null" <<endl;
                  }
                else {
                   cout << "Command format is " << verb << " csname selector" <<endl;
                   }
                }
             else {
                cout << "Command format is " << verb << " csname selector" <<endl;
                }
             }
           }
         else
            cout << "Command format is " << verb << " csname selector" <<endl;
         }
     else
        if (verb != "")
           cout << "Command \"" << verb << "\" is unrecognized.  Ignoring..." <<endl;
   }


// Background thread which periodically prints agent throughput statistics
#define STATS_INTERVAL  (3)      // Seconds between statistics printing
long monitor_stats(void *) {
   stop_mx.lock();         // Acquire mutex, which serializes stop_agent
   time_t wakeup = time(NULL) + STATS_INTERVAL; // First wakeup time
   while (!stop_agent) {
      Bool pop = stop_cv.timed_wait(wakeup);
                           // Wait until stop_agent gets set or for 3 seconds.
      if (stop_agent) continue;     // Do nothing else if stopping
      if (pop) {           // Timer expired
         sayrate();        // Give statistics on throughput rate
         wakeup = time(NULL) + STATS_INTERVAL;  // Next wakeup time
         }
      }
   stop_mx.unlock();       // No longer need mutex
   return 0;
   }

// Output agent throughput statistics
void sayrate() {
   time_t now = time(NULL);
   long   run_time = now - started_at;
   double rate;

   if (run_time >= 1)
      rate = (double) agent.eventsCompleted() / run_time;
   else
      rate = 1.0;  // Fudge something
   cout << "Agent throughput rate is " << rate
        << " events/second." << endl;
   }

// Indicate that agent should end
void declare_end() {
   stop_mx.lock();         // Get serialization on stop_agent
   stop_agent = true;      // Set flag telling main routine to stop
   stop_cv.broadcast();    // Signal condition variable to wake up threads
   stop_mx.unlock();       // Release serialization
   }

// Catch and process SIGINT, SIGTERM, or SIGHUP
void handler(int sig) {
   int saveerr=errno;      // Save errno in case interrupted code needs it

   cout << "\nTerminating agent due to signal" << endl;

   declare_end();          // Start shutting down the agent

   // Do NOT reestablish signal handler.  If a second signal of the same
   // type is received, let it terminate the process.  Presumably our attempt
   // to terminate gracefully didn't work.

   errno=saveerr;
   }

#ifdef IA_AIX
// Our "new" handler.  Invoked if an operator new() fails.  This is needed
// because when a new() fails, a null pointer is returned on AIX when the
// shared library (ibmcls) is used.  Other platforms and libraries throw an
// exception for this condition.  We throw our own exception on AIX.
void failed_new_handler() {
   ATHROW(AException("operator new() failed; no storage is available."));
   }
#endif

void loadCS(AString csname,AString selector)
{
   cout << "\nDynamically loading conduct set..." << endl;
   try {
      agent.loadConductSet(libHdl, our_engine, csname, selector);
   }
   catch (IAError &e) {
      cout << e.errormsg() << endl;
      cout << "\nConduct set load failed..." << endl;
      return;
   }
   cout << "\nConduct set load is complete..." << endl;
}

void unloadCS(AString csname,AString selector)
{
   cout << "\nDynamically unloading conduct set..." << endl;
   try {
      agent.unloadConductSet(libHdl, our_engine, csname, selector);
   }
   catch (IAError &e) {
      cout << e.errormsg() << endl;
      cout << "\nConduct set unload failed..." << endl;
      return;
   }
   cout << "\nConduct set unload is complete..." << endl;
}

void reset(AString selector)
{
   cout << "\nResetting agent for selector..." << endl;
   try {
      agent.reset(selector);
   }
   catch (IAError &e) {
      cout << e.errormsg() << endl;
      cout << "\nReset failed..." << endl;
      return;
   }
   cout << "\nReset is complete..." << endl;
}

// Get next parameter either quoted or not
// Input:
//         fromString - String to parse parm from
//         intoString - String where parsed parm is placed
// Output:
//         0 = parm successfully parsed
//             fromString - updated to remove parm that was parsed out of it
//             intoString - contains parsed parm
//         1 = parm failed parsing - unbalanced quotes
//             fromString - unchanged
//             intoString - unchanged
//        99 = parm failed parsing
//             fromString - unchanged
//             intoString - unchanged
//
int getNextPossiblyQuotedParm(AString &fromString, AString &intoString)
{
   int result = 0;

   if (fromString[0] == '\"') {                   // If parm starts with "
      int pos = indexOfQuote(fromString,1);       // Find next "
      if (pos == -1) return 1;                    // No ending quote
      intoString = fromString.subString(1,pos-1); // Set parm
      fromString.remove(0,pos+1);                 // Remove processed parm
      return 0;                                   // Success
      }
   else {                                         // Parm not quoted
      intoString = fromString.word(0);            // Set parm
      fromString.remove(0,intoString.length()); // Remove processed parm
      return 0;                                   // Success
      }
   return 99;                                     // Failure
}

// Find position of next non-imbedded quote within a string
// Input:
//         inString - String to be searched
//         startpos - 0-based index within string where search should begin
// Output:
//         -1 = quote not found
//        >-1 = index in string where quote was found (0-based)
//              if the string contained imbedded quotes, inString has been
//              modified to convert it to one quote.
int indexOfQuote(AString &inString, int startpos)
{
   int pos = startpos;
   while ((pos = inString.indexOf('\"',pos)) != -1) {  // Found quote
      if (inString[pos+1] == '\"') {  // If embedded quote...
         inString.remove(pos,1);      // Remove extra quote
         pos++;                       // ...move past it
         }
      else break;                   // ...else found ending quote
   }
   return pos;
}

void displayHelp(void)
{
   cout << "\nEnter command.  The following commands are recognized:\n"
           "   help or \?                - Display this help information\n"
           "   quit                     - End the agent program\n"
           "   stats                    - Print agent throughput statistics\n"
           "   loadcs   csname selector - load conduct set on-the-fly\n"
           "   unloadcs csname selector - unload conduct set on-the-fly\n"
           "   reset selector           - reset agent for selector"
           << endl;
}

