/*************************************************************************
** File:         RemAgent.java
** Description:  Sample Agent for Reminders
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/

/* RemAgent.java
 * This class provides reminders to clients in the form of EMAIL or direct
 * message generation.  Clients are able to
 * set or edit reminders and receive and manipulate messages.
 */
package ibm.AgentBuilder.samples.ReminderAgent;
import ibm.AgentBuilder.aeapi.IAAgent;
import ibm.AgentBuilder.aeapi.IAAdapterHandle;
import ibm.AgentBuilder.aeapi.IAEngineHandle;
import ibm.AgentBuilder.aeapi.IALibraryHandle;
import ibm.AgentBuilder.misc.IAError;
import ibm.AgentBuilder.Adapter.IAAdapter;
import ibm.AgentBuilder.Library.*;
import ibm.AgentBuilder.misc.*;
import java.net.*;
import java.io.*;
import java.util.*;

public class RemAgent extends IAAgent {

   static RemAgent theAgent;            //IAAgent implementation..
   Server server;
   public static IAEngineHandle  engineHandle;
   public static IALibraryHandle libHandle;
   private IALibrary appLibrary;
   private IALibInferenceCollector appTopCollector;
   private String cfgFileName;
   static String ruleSetMetadataString =
          "BEGIN PREDICATES" + "\n" +
             "EventName(string);" + "\n" +
             "AlarmId(string);" + "\n" +
             "EventTime(string,string);" + "\n" +
             "addTimeInterval(string, string, Integer, string, string, string);" + "\n" +
             "currentTimeBetween(string,string,string,string);" + "\n" +
             "setRandomStartBoundedIntervalAlarm" +
                    "(integer,string,string,string,integer,string,string);" + "\n" +
             "setEndBoundedIntervalAlarm" +
                    "(integer,string,string,string,string,string);" + "\n" +
             "setRandomBoundedIntervalAlarm" +
                    "(integer,string,string,string,string,string,integer,string,string);" + "\n" +
             "EMailWithSubject(string, string,string);" + "\n" +
             "SetMessage(string);" + "\n" +
          "END PREDICATES" + "\n" +
          "BEGIN SENSORS" + "\n" +
             "SENSOR addTimeInterval Time:addTimeInterval;" + "\n" +
             "SENSOR currentTimeBetween Time:currentTimeBetween;" + "\n" +
          "END SENSORS" + "\n" +
          "BEGIN EFFECTORS" + "\n" +
             "EFFECTOR setRandomStartBoundedIntervalAlarm " +
                "Time:setRandomStartBoundedIntervalAlarm;" + "\n" +
             "EFFECTOR setEndBoundedIntervalAlarm " +
                "Time:setEndBoundedIntervalAlarm;" + "\n" +
             "EFFECTOR setRandomBoundedIntervalAlarm " +
                "Time:setRandomBoundedIntervalAlarm;" + "\n" +
             "EFFECTOR EMailWithSubject SampMail:EMailWithSubject;" + "\n" +
             "EFFECTOR SetMessage Messages:SetMessage;" + "\n" +
          "END EFFECTORS";

   //--------------------------------------------------------------------------//
   // Implementation of IAAgent virtual functions                              //
   //--------------------------------------------------------------------------//

   // Constructor
   RemAgent(String fn) throws Exception {

      super();
       cfgFileName = new String(fn);
      // our concrete implementation of abstract IAAgent - AnAgent
      theAgent=this;

      // Setup and configure the agent
      if (!setupAgent()) throw new Exception("Agent setup failed");

      // start the agent - IAAgent start
      // once all components added and adapters and engines are connected)
      try {
         theAgent.start();
      } catch (Exception e) {
         System.out.println("Agent program exception " + e.getMessage());
         e.printStackTrace();
         throw e;
      }

      // Start the reminders server, listen on default port
      server = new Server(0,appLibrary);
   }

   // End of IAAgent virtual function implementation

   // Main: Expects the configuration file name as an argument
   public static void main(String args[]) {
      if (args.length == 1) {

         // Get filename to use for list of adapters is input to this agent
         // i.e, the configuration file for this agent filename.cfg specified
         // by the user.
         System.out.println("filename= "+ args[0]);
         try {
            new RemAgent(args[0]);
         } catch (Exception e) {
           System.exit(1);
         }
      } else {
         System.out.println("Usage: java RemAgent configuration_file_name");
         System.exit(1);
      }
   }

   public boolean setupAgent() {

      // Handles for: 1)Inference Engine, 2)Persistent storage library

      engineHandle = new IAEngineHandle();
      libHandle = new IALibraryHandle();
      BufferedReader input = null;

      try {
         input = new BufferedReader(new FileReader(cfgFileName));

         if (!addEngine(engineHandle)) return false;
         if (!addLibrary(input, libHandle)) return false;
         if (!addAdapters(input)) return false;

         preloadCS();                      // Load existing conduct sets

      } catch (Exception e) {
         System.out.println("Setting up agent exception "+e.getMessage());
         e.printStackTrace();
         return false;
      }
      return true;
   }   // end setup()

   private boolean addEngine(IAEngineHandle engineHandle) {
      // ADDING THE ENGINE:
      // First add the engine - we only have RAISE, so this is hard-coded
      // kind = "RAISE" , module_name="iagerais", empty string for parms
      System.out.println("Adding engine RAISE");
      try {
         theAgent.addEngine("RAISE","iagerais","",engineHandle);
      } catch (IAError e) {
         System.out.println("Engine setup exception "+e.getMessage());
         e.printStackTrace();
         return false;
      }
      return true;
   }   // end addEngine

   private boolean addLibrary(BufferedReader input,
                              IALibraryHandle libHandle) {
      //Declare library parms

      String map = null; // Map to library within file system
      String top = null; // Name of top collector in library (top directory)
      String library_type = null; // Library implementation is a file system
      String library_implementation_dll = null; // Dll containing implementation of library
      int numLevels; // Number of Collectors levels in the Library

      // fill library parms from cfg file.  Cfg file is also used to add Adapters.
      // The cfg file is: cfgFileName.cfg
      // that is provided as a command line parameter input
      // The contents of the library and adapter lines in the cfg file is the same
      // as described for the standard ABE sample agent - refer to
      // Configuration File for the Sample Agent Programs in the ABE User's Guide.

      String line;
      StringTokenizer tokens;
      String currentToken=new String();

      try {

      // Read, tokenize and store values for library parms from cfg file

         line = input.readLine();
         if ( null != line ) {
            tokens = new StringTokenizer(line); //default delimiter is " \t\n\r"

            // read line data w/o \r\n
            if (tokens.hasMoreTokens())
               currentToken = new String(tokens.nextToken());
            System.out.println("key= " + currentToken);
            if (currentToken.equals("library")) {

               // next token is library type
               library_type=new String(tokens.nextToken());

               // next token is library module name
               library_implementation_dll=new String(tokens.nextToken());

               // next token library map
               map=new String(tokens.nextToken());
               if (map.lastIndexOf(File.separator) < (map.length()-1)) map +=File.separator;
               // next token library top collector
               top=new String(tokens.nextToken());

               // next token number of Collector levels
               numLevels=(Integer.decode(new String(tokens.nextToken()))).intValue();

               // Add library to the engine
               System.out.println("Adding library");
               theAgent.addLibrary(map, top, library_implementation_dll,
                               library_type, numLevels, libHandle);

               // libHandle returned by addLibrary represents the new instance of the
               // library added to the current agent.

               // Previously the addLibrary established the library component of the
               //   agent.  Now we are going to establish an interface to the same
               //   library for direct use by the application for adding new rules
               //   etc. while the agent is running.  The anchor for the application
               //   use of the library is an IALibrary object.  We construct it; then
               //   use IALibary.attachImplementation() to complete its association with
               //   the same inferencing library that the agent is employing.  Because the
               //   library is truly the same as the one used by the agent, we can also
               //   employ the same parameter values as used for addLibary above.

               // Create library object
               appLibrary = new IALibrary();

               appLibrary.attachImplementation(map,
                                   library_type,
                                   library_implementation_dll,
                                   top,
                                   numLevels);

               // Create and access the top collector object

                  appTopCollector = appLibrary.getTop();
                  appTopCollector.get();

            } else {
                System.out.println("\nLibrary statement must be first line in " + cfgFileName);
                return false;
            }  // End of token != Library
         } else {
            System.out.println("\nLibrary statement must be first line in " + cfgFileName);
            return false;
         }  // End of (line == null)
      }
      catch (IAError e) {
         System.out.println("Library setup exception "+e.getMessage());
         e.printStackTrace();
         return false;
      }
      // Catch fileNotFoundException, NumberFormatException, and IOException
      catch (Exception e) {
         System.out.println("Library setup exception "+e.getMessage());
         e.printStackTrace();
         return false;
      }
      return true;
   }  // end of addLibrary

   private boolean addAdapters(BufferedReader input) {
      // ADDING THE ADAPTERS.
      IAAdapterHandle aHandle;
      String domain;
      String moduleName;
      String parms;
      String conductSet;
      String selector;

      String line;
      StringTokenizer tokens;
      String currentToken=new String();

      try {
         boolean endAdapters=false;

         // loop and read adapters names from file,
         // add each adapter and connect it to the engine
         // format of cfg file is:
         // key1<tab>domain1<tab>module_name1<tab>parms1
         // key2<tab>domain2<tab>module_name2<tab>parms2
         // etc.. - again the same as the configuration file for the ABE sample agent.
         // ex: Adapter<tab>Time<tab>iagatime
         while (!endAdapters) {
            System.out.println("Read first adapter");
            line = input.readLine();
            if ( null == line ) break;
            tokens = new StringTokenizer(line); //default delimiter is " \t\n\r"

            // get keyword ("adapter")
            // read line data w/o \r\n
            if (tokens.hasMoreTokens())
               currentToken = new String(tokens.nextToken());
            System.out.println("key= " + currentToken);
            if (currentToken.equals("adapter")) {

               // next token is adapter domain
               domain=new String(tokens.nextToken());

               // next token is adapter module name
               moduleName=new String(tokens.nextToken());

               // next token is parms for the adapter (if they exist)

               // Note:  The tokenizer is parsing on 4 values (blank, \n, \t,
               // \r) so we don't really know how the adapter parm data is
               // delimitted.  However, since the Message adapter is the only
               // one with parms for this agent, and it's parms are expected
               // to be blank-delimited, we can safely blank-delimit the parm
               // list that we build.

               if (tokens.hasMoreTokens()) {
                  parms=new String(tokens.nextToken());
                  while (tokens.hasMoreTokens()){
                    parms += " " + tokens.nextToken();
                    }
                  }
               else parms=new String("");

               // add the adapter and connect it to the engine
               aHandle=new IAAdapterHandle();
               theAgent.addAdapter(domain,moduleName,parms,aHandle);
               theAgent.connect(aHandle,engineHandle);
            }
            else endAdapters=true;
         } //end-while
      }
      catch (Exception e) {
         System.out.println("Adapters setup exception "+e.getMessage());
         e.printStackTrace();
         return false;
      }
      return true;
   }  // end addAdapters

   boolean preloadCS() {
      IALibInferenceCollector user_collector = null;
      IALibInferenceContentsElement current_collector_element = null;
      IALibInferenceContentsElement current_ruleset_element = null;
      IALibInferenceRuleSet user_ruleset = null;

      try {
        //   Find all user collectors, and their rulesets so that we can
        //   initialize the agent.  That is, we need to establish
        //   alarms for all of the reminders that we had in the
        //   system when the server was last up.
        //
        String user_name; //used for user collector, ruleset, and
                          //selector names (common name for all)
        boolean looper;
        String csname;
        //
        appTopCollector.setCursorElementScope(IALibrary.IALibICCS_collectors);
        current_collector_element = appLibrary.new_IALibInferenceContentsElement();
        current_ruleset_element = appLibrary.new_IALibInferenceContentsElement();
        for(
        looper = appTopCollector.firstElement(current_collector_element);
           looper==true;) {
        //
        user_name = current_collector_element.extractElementName();
        user_collector =
            appLibrary.new_IALibInferenceCollector(appTopCollector,user_name);
        user_collector.get();
        //
        //  We cannot assume that finding a collector means
        //  that a ruleset exists because we could
        //  have a Collector with only messages held for the
        //  user to retrieve (has deleted the reminder rules
        //  that generated the message!
        //
        user_collector.setCursorElementScope(IALibrary.IALibICCS_rulesets);
        //
        if (user_collector.find(current_ruleset_element,user_name,
                              IALibrary.IALibICET_ruleset))
        {
        //
        //   We use user_name above and below because
        //   user name, user collector name, ruleset name,
        //   and selector name are all the same!
        //
        user_ruleset = appLibrary.new_IALibInferenceRuleSet(user_collector,
                             user_name);
        csname = File.separator + user_name + File.separator + user_name;

        // Note:  We use the application library object to 'find' which
        //        rule sets to load, however we pass the Agent's library
        //        object handle on the load conduct set since the agent is
        //        doing the load.
        theAgent.loadConductSet(libHandle, engineHandle,
                                 csname, user_name);
        //
        //  Where lHandle and eHandle were established in the
        //  setupAgent() method earlier.  Selector
        //  name is the same as collector name (user_name)
        //
        } // end of the if
        looper = appTopCollector.nextElement(current_collector_element);
        } // end of the for loop
      } catch (IAError e) {
         System.out.println("Caught IAError: " + e.errormsg());
         return false; // Unable to pre-load existing conduct sets
      }
   return true;
   } // end preloadCs()


} //end class RemAgent

// The Reminders server class:
// 1. Creates a server socket thread that listens for connections from clients
// 2. Once a connection request from a client is received, a Connection
//    thread is started to handle all comm. with client.
//
class Server extends Thread {
   public final static int DEFAULT_PORT = 6789;
   protected int port;
   protected ServerSocket listen_socket;
   protected IALibrary library;

   // Exit with an error message when an error occurs.
   public static void fail(Exception e, String msg) {
      System.err.println(msg + ": " + e);
      System.exit(1);
   }

   // Creat a ServerSocket to listen for connections on;
   // Start the thread.
   public Server(int port, IALibrary library) {
      if (port == 0) port = DEFAULT_PORT;
      this.port = port;
      this.library = library;
      try {
         listen_socket = new ServerSocket(port);
      }
      catch (IOException e) {
         fail(e, "Exception creating server socket");
      }
      this.start();
   }

   // The body of the server thread. Loop forever, listening for and
   // accepting connections from clients. For each connection,
   // create a Connection object to handle communication through the new
   // socket to the client.
   public void run() {
      Socket client_socket;
      Connection c;
      try {
         while(true) {
            System.out.println("Server: listening on port " + port);
            client_socket = listen_socket.accept();
            c = new Connection(client_socket,library);
         }
      }
      catch (IOException e) {
         fail(e, "Exception while listening for connections");
      }
   }

}

// This class is the thread that handles all communications with a client:
// 1. Opens a socket with the client
// 2. Starts a thread that loops forever, and gets commands from client
//
class Connection extends Thread {
   protected Socket client;
   protected BufferedReader in;
   protected PrintWriter out;
   private RemMessageVector msgCollection = new RemMessageVector();
   private boolean debug = false;

   IALibrary library = null;    // Library object

   // Declare the library objects where rules are saved
   // Collector object
   public IALibInferenceCollector top_collector = null;

   // User's collector object
   public IALibInferenceCollector user_collector = null;

   // Current contents element object
   public IALibInferenceContentsElement current_collector_element = null;

   // Current contents element object
   public IALibInferenceContentsElement current_ruleset_element = null;

   // User's rule set object
   public IALibInferenceRuleSet user_ruleset = null;

   // User's rule object
   public IALibInferenceRule user_rule = null;

   // Number of rules in user's ruleset
   int number_of_rules = 0;

   // User ID
   String userID = "";

   // Initialize the streams and start the threads
   public Connection(Socket client_socket,
                     IALibrary library) {
      client = client_socket;
      this.library = library;
      try {
        top_collector = library.getTop();
      } catch (IAError e) {
         System.err.println("Exception while accessing library.");
         return;
      }
      try {
         in = new BufferedReader(new InputStreamReader(client.getInputStream()));
         out = new PrintWriter(client.getOutputStream(), true);
      }
      catch (IOException e) {
         try {client.close();}
         catch (IOException e2){};
         System.err.println("Exception while getting socket streams.");
         return;
      }
      this.start();
   }

   // Provide the service:
   // Read a line, form a command, and respond..
   public void run() {
      String line, cmd, remID, eventDateString;
      String remsStr, rulesStr, remStr, rule1Str, rule2Str, v1, v2, fn;
      boolean newFlg;
      int i, mode = 5;
      Hashtable remDB = new Hashtable();
      String dbStr = new String();


      // Connectionless mode - provide service and return
      // Read in the commands from the client:
      // 1. GET: Return to the client, all reminders for the specified user
      //         in persistant storage
      // syntax= GET:USERID
      // 2. PUT:Stores a reminder in persistant storage and creates rules
      // syntax= PUT:USERID:REMSTRING
      // 2. REMOVE:Removess a reminder and associated rules from storage
      // syntax= REMOVE:USERID:REMID
      // 3. MODIFY: Same as a remove followed by a put.
      // 4. GETMSGS: Return to the client all messages for the specified
      //             user in persistent storage, and markr returned messages
      //             as read.
      // syntax=GETMSGS:USERID
      // 5. GETMSGSNOMARK: Return to the client all messages for the specified
      //             user in persistent storage, but do not mark returned
      //             messages as read.
      // syntax=GETMSGSNOMARK:USERID
      // 6. DELMSGS: Removes specified messages for the specified user from
      //             persistent storage.
      // syntax=DELMSGS:USERID:MSGID,MSGID,MSGID,...
      // 7. MARKMSGSASNEW: Marks all messages for the specified user
      //                   as not read.
      // syntax=MARKMSGSASNEW:USERID:MSGID,MSGID,...
      // 8. STATUSMSGS: Returns true if new messages exist for the specified
      //                user in persistent storage.
      // syntax=STATUSMSGS:USERID

      try {
         for(;;) {

            // Read in the message line from the client
            line = in.readLine();
            System.out.println("Client: " + line);
            if (line == null) {
               out.println("ERROR:01");
               System.out.println("Server: null command from client!");
               break;
            }

            // 1. Parse the command: cmnd:userID:v1
            StringTokenizer st = new StringTokenizer(line, ":");
            cmd = st.nextToken().toUpperCase();
            System.out.println("Server: CMD= " + cmd);
            if(!st.hasMoreTokens()&& !cmd.equals("QUIT")) {
               out.println("ERROR:02");
               System.out.println("Server: Message without user ID..");
                break;
            }

            // 2nd. token should be the user's ID
            if( !cmd.equals( "QUIT" ) )
              {
              userID = st.nextToken();
              System.out.println("Server: " + cmd + " " + userID);
              }

            if (cmd.indexOf("MSGS") == -1 && !cmd.equals( "QUIT" )) {
              // Setup rules access.. mode will tell us if new user,
              // if user has a collector, or any rules
              mode = setupRulesAccess(userID,cmd);
              System.out.println("mode= " + mode);
              if ( mode > 2 ) {

                 // modes 5, 4 are errors: can't create; 3 is new collector
                 // top collector, user's collector, or user's rule-set respectively.
                 out.println("ERROR:03");
                 break;
              }
            }

            // service the command
            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% GET %%%%%%%%%%%%%%%%%%%%%%%%
            if (cmd.equals("GET")) {
               System.out.println("Server: GET command.. ");
               if (0 == mode) {

                  // user already has collector & a rule set; get the rules
                  remsStr = getUserRules();
                  if (null != remsStr) {


                     //return reminders to the client
                     out.println("GET:" + remsStr);
                     System.out.println("User " + userID + " reminders: ");
                     System.out.println(remsStr);
                  } else {

                     // user has no saved reminders
                     // respond with LOGON cmnd w/o rules
                     out.println("GET:");
                  }
               } else if (2 == mode) {

                  // user has an empty rule set.
                  // respond with LOGON cmnd w/o rules
                  out.println("GET:");
                  System.out.println("Server: empty ruleset.. ");
               }
               break;

            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% PUT %%%%%%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("PUT")) {
               System.out.println("Server: PUT command.. ");
               if(!st.hasMoreTokens()) {
                  out.println("ERROR:05");
                  System.out.println("Server: PUT without parms..");
                  break;
               }

               //Command format is: PUT:remMsgStr where remMsgStr is:
               // ID$remID$remStr
               v1 = st.nextToken();
               StringTokenizer st1 = new StringTokenizer(v1, "$");
               if(!st1.nextToken().equals("ID")) {
                  System.out.println("Server: PUT command error: ");
                  System.out.println("Server: No ID field received in reminder.");
                  out.println("ERROR:06");
                  break;
               }

               // Get reminder ID, reminder string, convert to KIF rule
               // format, and save in library.
               remID = st1.nextToken();
               System.out.println("Server: ID= " + remID);
               remStr = st1.nextToken();

               //If a brand new reminder, create a new rem ID
               if ( remID.equals("000") ) {
                  remID = createUniqueID();

                  // update the reminder string with the new ID
                  remStr = "ID=" + remID + "#" + remStr;
               }
               System.out.println(remStr);
               if (0 == user_ruleset.size()) {
                  newFlg = true;
               }
               else {
                  newFlg = false;
               }
               if (saveReminderInLibrary(remID, remStr, newFlg)) {

                  // Get the rules string (meta data) and return it..
                  System.out.println("Get user rules..");
                  remsStr = getUserRules();
                  System.out.println("To the Client >>>>>>>>>>>>>>>>>>>>>");
                  System.out.println("PUT:" + remsStr);
                  out.println("PUT:" + remsStr);
               } else {
                  System.out.println("Error saving Reminder..");
                  out.println("ERROR:07");
                  break;
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% MODIFY %%%%%%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("MODIFY")) {
               System.out.println("Server: MODIFY command.. ");

               //Command format is: MODIFY:userID:remMsgStr where remMsgStr is:
               // ID$remID$remStr
               v1 = st.nextToken();
               StringTokenizer st1 = new StringTokenizer(v1, "$");
               if(!st1.nextToken().equals("ID")) {
                  System.out.println("Server: MODIFY command error: ");
                  System.out.println("Server: No ID field received in reminder.");
                  out.println("ERROR:08");
                  break;
               }

               // Get reminder ID, reminder string, convert to KIF rule
               // format, and save in library.
               remID = st1.nextToken();
               System.out.println("Server: ID= " + remID);
               remStr = st1.nextToken();

               // restore the reminder ID to reminder string
               remStr = "ID=" + remID + "#" + remStr;
               System.out.println(remStr);

               // 1st remove existing rule, then save new one
               newFlg = false;
               if (deleteReminderFromLibrary(remID)) {
                  if (saveReminderInLibrary(remID, remStr, newFlg)) {

                     // Get the rules string (meta data) and return it..
                     System.out.println("Get user rules..");
                     remsStr = getUserRules();
                     System.out.println("To the Client >>>>>>>>>>>>>>>>>>>>>");
                     System.out.println("MODIFY:" + remsStr);
                     out.println("MODIFY:" + remsStr);
                  } else {
                     System.out.println("Error saving Reminder..");
                     out.println("ERROR:08");
                     break;
                  }
               } else {
                  System.out.println("Error deleting reminder..");
                  out.println("ERROR:09");
                  break;
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% REMOVE %%%%%%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("REMOVE")) {

               // SYNTAX= REMOVE:USER_ID:REMINDER_ID
               System.out.println("Server: REMOVE command.. ");
               if(!st.hasMoreTokens()) {
                  System.out.println("Server: REMOVE without reminder ID..");
                  out.println("ERROR:10");
                  break;
               }

               remID = st.nextToken();
               System.out.println("Server: REMOVE: " + userID + " " + remID);

               if (deleteReminderFromLibrary(remID)) {
                  String csName = File.separator + userID + File.separator + userID;

                  // check if ruleset is empty
                  if (0 == user_ruleset.size()) {

                     // unload conduct set and delete collector
                     try {

                        // We reset the user BEFORE we try to delete
                        // the collector.  This is to make sure no new
                        // messages get generated after/during the collector
                        // delete.

                        RemAgent.theAgent.reset(userID);

                        // Delete rule set and rule set metadata from
                        // persistent storage

                        user_ruleset.del();

                        // Now delete the user collector, since it is empty
                        // Check to see if any messages for this user (any
                        // collector metadata) and delete collector from
                        // persistent storage if no messages (since no rules
                        // exist either).
                        // Note that our version of the metadata may be
                        // outdated (the metadata on persistent storage may
                        // have changed since we did our 'get' of it).

                        if (user_collector.extractMetadata().equals("")) {
                           try {
                             synchronized (top_collector) {
                               user_collector.del();
                               }
                           } catch (IAError e) {
                           if (e.msgnumber() !=
                               IAError.IA_E_LIBRARY_VERSION_ERROR)
                             throw e;    // Not version mismatch, throw error
                           }
                        }
                     } catch (IAError e) {
                        System.out.println("Remove Rule- Caught IAError: " + e.errormsg());
                        out.println("ERROR:11");
                        break;
                     }
                  } else {
                     try {
                        user_ruleset.put();

                        // load the new conduct set
                        RemAgent.theAgent.reset(userID);
                        RemAgent.theAgent.loadConductSet(RemAgent.libHandle,
                                     RemAgent.engineHandle, csName, userID);
                     } catch (IAError e) {
                        System.out.println("Remove Rule- Caught IAError: " + e.errormsg());
                        out.println("ERROR:12");
                        break;
                     }
                  }

                  // Get the rules string (meta data) and return it..
                  System.out.println("Get user rules..");
                  remsStr = getUserRules();
                  System.out.println("To the Client >>>>>>>>>>>>>>>>>>>>>");
                  System.out.println("REMOVE:" + remsStr);
                  out.println("REMOVE:" + remsStr);
               } else {
                  System.out.println("Server: Error deleting reminder..");
                  out.println("ERROR:13");
                  break;
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%%%%% GET ALL MESSAGES %%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("GETMSGS")) {
               System.out.println("Server: GET ALL MESSAGES command.. ");
               // user already has collector & a message set; get the msgs
               remsStr = convertNLs(getMsgs(userID, true));
               if (null != remsStr) {

                  //return reminders to the client
                  out.println(cmd + ":" + remsStr + "\n" );
                  System.out.println("User " + userID + " messages: ");
                  System.out.println(remsStr);
               } else {

                  // user has no saved reminders
                  // respond with current command name with no messages
                  out.println(cmd + ":");
               }
               break;

            // %%%%%%%%%%%%%%%%%%%%%%%% GET MESSAGES NO MARK %%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("GETMSGSNOMARK")) {
               System.out.println("Server: GET ALL MESSAGES NO MARK command.. ");
               // user already has collector & a message set; get the msgs
               remsStr = convertNLs(getMsgs(userID, false));
               if (null != remsStr) {

                  //return reminders to the client
                  out.println(cmd + ":" + remsStr + "\n" );
                  System.out.println("User " + userID + " messages: ");
                  System.out.println(remsStr);
               } else {

                  // user has no saved reminders
                  // respond with current command name with no messages
                  out.println(cmd + ":");
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%%%%% DELETE MESSAGES %%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("DELMSGS")) {
               System.out.println("Server: DELETE ALL MESSAGES command.. ");
               // user already has collector & a message set; delete  msgs
               v1 = st.nextToken();
               remsStr = deleteMsgs(userID, v1);
               if (null != remsStr) {

                  //return reminders to the client
                  out.println(cmd + ":" + remsStr + "\n" );
                  System.out.println("User " + userID + " messages: ");
                  System.out.println(remsStr);
               } else {

                  // user has no saved reminders
                  // respond with current command name with no messages
                  out.println(cmd + ":");
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%%%% MARK MESSAGES NEW %%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("MARKMSGSASNEW")) {
               System.out.println("Server: GET NEW MESSAGES command.. ");
               // user already has collector & a message set; delete  msgs
               v1 = st.nextToken();
               remsStr = markMsgs(userID, v1);
               if (null != remsStr) {

                  //return reminders to the client
                  out.println(cmd + ":" + remsStr + "\n" );
                  System.out.println("User " + userID + " messages: ");
                  System.out.println(remsStr);
               } else {

                  // user has no saved reminders
                  // respond with current command name with no messages
                  out.println(cmd + ":");
               }
               break;
            // %%%%%%%%%%%%%%%%%%%%%%%% QUERY IF NEW MESSAGES %%%%%%%%%%%%%%%%%
            } else if (cmd.equals("STATUSMSGS")) {
               System.out.println("Server: QUERY NEW MESSAGES command.. ");
               // user already has collector & a message set; get the msgs
               remsStr = newMsgs(userID);
               if (null != remsStr) {

                  //return reminders to the client
                  out.println(cmd + ":" + remsStr + "\n" );
                  System.out.println("User " + userID + " messages: ");
                  System.out.println(remsStr);
               } else {

                  // user has no saved reminders
                  // respond with current command name with no messages
                  out.println(cmd + ":");
               }
               break;


            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% QUIT %%%%%%%%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("QUIT")) {
              // The msgConnection code uses this to terminate the read loop
              // from the socket...
               System.out.println("QUIT: client terminated connection.. ");
               break;

            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% OK %%%%%%%%%%%%%%%%%%%%%%%%%%%%
            } else if (cmd.equals("OK")) {
               System.out.println("OK: client received response.. ");
               break;

            // all other commands..
            // %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% UNKNOWN %%%%%%%%%%%%%%%%%%%%%%%%
            } else {
               System.out.println("Server: unknown command.. ");
               out.println("ERROR:14");
               break;
            }
         } //end for
      } catch (IOException e){};
      try {client.close();} catch (IOException e2){};
      System.out.println("Done! Close the connection..");
   } //end run() thread


   // *************************************************************************
   // Setup user's rules access from persistant storage..
   int setupRulesAccess(String usr, String cmd) {
   boolean newUser = true;

      try {

       synchronized (top_collector) {
         // Indicate we only want to look at collectors within the top
         // collector (we're going to loop through our user's collectors).
         top_collector.setCursorElementScope(
                   IALibrary.IALibICCS_collectors);

         // Create the user's collector object
         // Using the parent collector (in this case top)
         // And the name of the collector (pulled from contents element)
         user_collector =
                 library.new_IALibInferenceCollector(top_collector, usr);

         // Contents element object: Place for find to put it's stuff
         IALibInferenceContentsElement contents_element =
                              library.new_IALibInferenceContentsElement();

         if (top_collector.find(contents_element, usr,
                           IALibrary.IALibICET_collector)) {

            // User Collector exists. Populate it with data from persistant store
            // If collector metadata exists (user has messages pending),
            // we need to load it into the user_collector so that when it
            // comes time to 'delete' the rules from persistent storage, we
            // can tell if there is metadata or not.
            user_collector.setMetadataName('1');
            try {
              user_collector.get();
            } catch (IAError e) {
              if (e.msgnumber() !=
                  IAError.IA_E_LIBRARY_COLLECTOR_METADATA_EXISTENCE_ERROR)
                throw e;           // Other get problem, rethrow...
            }
            newUser = false;
            System.out.println("Collector exists");
         } else {
            if (cmd.equals("GET")) return 2;   // Don't create user for get
            else {
            // User Collector does not exist. Create it in persistant storage
            System.out.println("Collector doesn't exist. Create one..");
            user_collector.put();
            }
         }
       }   // End of synchronized block for top_collector
      } catch (IAError e) {
         System.out.println("Caught IAError: " + e.errormsg());
         System.out.println("Could not create user's collector!");
         return 4; //could not create a user's collector in persistant storage
      }

      //user has a collector, give user a rule Set
      System.out.println("Collector Name: " +
                                   user_collector.extractName());
      try {

         // Indicate we only want to access rule sets in the user's collector.
         user_collector.setCursorElementScope(
                    IALibrary.IALibICCS_rulesets);

         // Get empty rule set element object
         current_ruleset_element =
             library.new_IALibInferenceContentsElement();

         // Create rule set object indicating parent collector
         // and rule set name

         // We create the rule set without a metadata name in order to
         // "speed up" any future "put" of the ruleset for the case when the
         // rules already exist.
         // In the case where the collector doesn't exists, or the
         // rule set doesn't exist, then we'll explicitly set the
         // metadata name.  This will cause the initial metadata will be put
         // into persistent storage on any subsequent user put request.
         // Once the metadata is initially placed on persistent storage, we
         // don't change it, so there's no need to re-put it and/or get it.

         user_ruleset = library.new_IALibInferenceRuleSet(user_collector, usr,
                         "", '0' );


         // Populate rule set object with data from persistant store
         if (newUser) {
            user_ruleset.setMetadataName('0');
            user_ruleset.setMetadata("CONDUCTSET " + usr + "\n" + RemAgent.ruleSetMetadataString);
            return 2;
         } else {
            if (user_collector.find(current_ruleset_element,usr,
                                    IALibrary.IALibICET_ruleset))
              {
              user_ruleset.get();
              }
            else {
               // This can happen if the user collector exists for the purpose
               // of housing user messages, but the user rules have been
               // deleted.
               System.out.println("User has an empty ruleSet!");
               user_ruleset.setMetadataName('0');
               user_ruleset.setMetadata("CONDUCTSET " + usr + "\n" + RemAgent.ruleSetMetadataString);
               return 2;
            }
         }
      } catch (IAError e) {

         // Could not give user a new rule set
         System.out.println("Caught IAError: " + e.errormsg());
         System.out.println("No ruleSet in persistant storage!");
         return 3; //new collector for user but no ruleset in persistant storage
      }
      return 0;
   }  //end setupRulesAccess

   // *************************************************************************
   boolean saveReminderInLibrary(String remID, String remStr, boolean newFlg) {

      Reminder rem = new Reminder(remStr);

      // Put the specific rule in the Library; // ruleStr is in KIFF format
      try {
         System.out.println("Create the rule..");

         // rule_name, rule_id, KIFF_string, metadata
         user_rule = library.new_IALibInferenceRule(
                                      remID + "Set1", remID + "Set1",
                                      rem.getAlarmSet1Rule(), remStr);
         user_ruleset.addRule(user_rule);

         user_rule = library.new_IALibInferenceRule(
                                      remID + "Set2", remID + "Set2",
                                      rem.getAlarmSet2Rule(), "");
         user_ruleset.addRule(user_rule);

         user_rule = library.new_IALibInferenceRule(
                                      remID + "Set3", remID + "Set3",
                                      rem.getAlarmSet3Rule(), "");
         user_ruleset.addRule(user_rule);

         user_rule = library.new_IALibInferenceRule(
                                      remID + "Pop", remID + "Pop",
                                      rem.getAlarmPopRule(), "");
         user_ruleset.addRule(user_rule);

         // save the rule
         System.out.println("Save the rules..");
         user_ruleset.put();
         String csname = File.separator + userID + File.separator + userID;
         System.out.println("Conduct Set Name= " + csname);
         if (newFlg) {
            RemAgent.theAgent.loadConductSet(RemAgent.libHandle,
                                        RemAgent.engineHandle, csname, userID);
            }
         else {
            RemAgent.theAgent.reset(userID);
            RemAgent.theAgent.loadConductSet(RemAgent.libHandle,
                                        RemAgent.engineHandle, csname, userID);
            }
      } catch (IAError e) {
               System.out.println("Add New Rule- Caught IAError: " + e.errormsg());
               return false;
      }
      return true;
   } //end saveReminderInLibrary()

   // ************************************************************************
   boolean deleteReminderFromLibrary(String remID) {

      //remove an existing rule; 1st. find the first rule
      try {
         user_rule = library.new_IALibInferenceRule(remID + "Set1");
         if (user_ruleset.find(remID + "Set1", user_rule)) {

            // delete the rule
            System.out.println("Delete the rule..");
            user_ruleset.deleteRule();
         } else {
            System.out.println("Removing existing reminder: Rule can not be found!.");
            return false;
         }
         //remove an existing rule; 1st. find the second rule              @bb
         user_rule = library.new_IALibInferenceRule(remID + "Set2");
         if (user_ruleset.find(remID + "Set2", user_rule)) {

            // delete the rule
            System.out.println("Delete the rule..");
            user_ruleset.deleteRule();
         } else {
            System.out.println("Removing existing reminder: Rule can not be found!.");
            return false;
         }
         //remove an existing rule; 1st. find the third rule               @bb
         user_rule = library.new_IALibInferenceRule(remID + "Set3");
         if (user_ruleset.find(remID + "Set3", user_rule)) {

            // delete the rule
            System.out.println("Delete the rule..");
            user_ruleset.deleteRule();
         } else {
            System.out.println("Removing existing reminder: Rule can not be found!.");
            return false;
         }
         if (user_ruleset.find(remID + "Pop", user_rule)) {

            // delete the rule
            System.out.println("Delete the rule..");
            user_ruleset.deleteRule();
         } else {
            System.out.println("Removing existing reminder: Rule can not be found!.");
            return false;
         }
      } catch (IAError e) {
         System.out.println("Remove Rule- Caught IAError: " + e.errormsg());
         return false;
      }
      return true;
   } //end deleteReminderFromLibrary
   // ************************************************************************

   // *************************************************************************
   // Get user's rules (rule set) from library (persistant storage)
   // Rules are returned as a tokenized string as indexed reminder strings:
   // index1Str;rem1Str;index2;rem2Str;..
   String getUserRules() {

      String remsStr = null;
      String ruleNameStr = null;
      int    ruleNameLen = 0;
      try {

         // Get number of rules in rule set
         number_of_rules = user_ruleset.size();
         System.out.println("Rule Set Name: " + user_ruleset.extractName()
                                          + " # of rules= " + number_of_rules);

         if ( 0 == number_of_rules ) return null;

         // Get first rule
         user_rule = user_ruleset.firstRule();
         ruleNameStr = user_rule.extractName();
         ruleNameLen = ruleNameStr.length();

         // Don't need to convert rules to reminders since the reminder
         // string is already stored in the meta data of the rule..
         if (ruleNameLen >= 4){
           if (ruleNameStr.substring(ruleNameLen-4).compareTo("Set1") == 0) {
             remsStr = user_rule.extractMetadata() + ";";
           }
         }

         // Loop through the rule set for each rule..
         // Concatinate in remsStr the rule-level metadata that is associated with each
         // of the "......Set1" named rules (metadata is only stored once - with the first
         // rule of a set of rules that represent each reminder.

         for (int i = 1; i < number_of_rules; i++) {

             // Get next rule
             user_rule = user_ruleset.nextRule();
             ruleNameStr = user_rule.extractName();
             ruleNameLen = ruleNameStr.length();

             if (ruleNameLen >= 4){
               if (ruleNameStr.substring(ruleNameLen-4).compareTo("Set1") == 0) {
                 remsStr = remsStr + user_rule.extractMetadata() + ";";
               }
             }
         }
      } catch (IAError e) {
         System.out.println("Caught IAError: " + e.errormsg());
         remsStr = null;
      }
      System.out.println("remsStr=  " + remsStr);
      return remsStr;
   } // end getUserRules()

   //Create unique Reminder ID from time & date
   String createUniqueID() {
      Date d = new Date();
      return (new Integer(d.hashCode())).toString();

   }


   //-------------------------------------------------------------------------
   // Method: getMsgs( String )
   //-------------------------------------------------------------------------
   public String getMsgs( String aUser, boolean mark ) {
     if ( debug ) System.out.println( "Connection::getMsgs - Start --->" );

     String answer;
     boolean retry;
     int rc = 0;
     int retryCount = 0;

     do {
       retry = false;
       if ( getUserMessages( aUser ) )
         {
         answer = msgCollection.print();
         if ( mark ) {
           msgCollection.markAllRead();
           if ( (rc = updateUserMsgs( aUser )) == 1 ) {
             retry = true;                // version conflict - retry
             retryCount++;
             }
           }
         }
       else                             // user has no messages!!
         answer = "500 ";
     } while (retry && (retryCount <= 2));

     if (rc > 0) {
       answer = "501 ";                 // error writing messages back to file
       }

     if ( debug ) System.out.println( "Connection::getMsgs - End   <---" );
     return( answer );
   }

   //-------------------------------------------------------------------------
   // Method: deleteMsgs( String, String )
   //-------------------------------------------------------------------------
   public String deleteMsgs( String aUser, String msgsToDelete ) {
     if ( debug ) System.out.println( "Connection::deleteMsgs - Start --->" );

     boolean retry;
     int rc = 0;
     int retryCount = 0;

     String  answer, aLine = new String(), aMsgID;
     int i;

     do {
       retry = false;
       if ( getUserMessages( aUser ) )
         {
         StringTokenizer sT = new StringTokenizer( msgsToDelete, "," );
         while ( sT.hasMoreTokens() )
           {
           aMsgID = sT.nextToken();
           for ( i = 0; i < msgCollection.size(); i++)
             {
             if ( ((RemMessage)msgCollection.elementAt( i )).msgID().equals( aMsgID ) )
               {
               if (debug)
                  System.out.println( "Connection::deleteMsgs - going to delete "
                                 + "message with ID = " + aMsgID);
               msgCollection.removeElementAt( i ) ;
               break;
               }
             }
           }
           if ( (rc = updateUserMsgs( aUser )) == 1 ) {
             retry = true;                // version conflict - retry
             retryCount++;
             }
         }
       else
         answer = "501 " + aUser + " " + msgsToDelete;
     } while (retry && (retryCount <= 2));

     if (rc == 0)
       answer = "220 ";
     else
       answer = "500 ";

     if ( debug ) System.out.println( "Connection::deleteMsgs - End   <---" );
     return( answer );
  }

   //-------------------------------------------------------------------------
   // Method: markMsgs( String, String )
   //-------------------------------------------------------------------------
   public String markMsgs( String aUser, String msgsToMark ) {
     if ( debug ) System.out.println( "Connection::markMsgs - Start --->" );

     boolean retry;
     int rc = 0;
     int retryCount = 0;

     String  answer, aLine = new String(), msgID;
     int i;

     do {
       retry = false;
       if ( getUserMessages( aUser ) )
         {
         StringTokenizer sT = new StringTokenizer( msgsToMark, "," );
         while ( sT.hasMoreTokens() )
           {
           msgID = sT.nextToken() ;
           for ( i = 0; i < msgCollection.size(); i++ )
             {
             if ( ((RemMessage)msgCollection.elementAt( i )).msgID().equals( msgID ) )
               {
               if (debug)
                 System.out.println( "Connection::markMsgs - going to mark as new "
                                     + msgID );
               ( (RemMessage)msgCollection.elementAt( i )).setStatus( 'n' ) ;
               break;
               }
             }
           }
           if ( (rc = updateUserMsgs( aUser )) == 1 ) {
             retry = true;                // version conflict - retry
             retryCount++;
             }
         }
       else
         answer = "501 " + aUser + " " + msgsToMark;
     } while (retry && (retryCount <= 2));

     if (rc == 0)
       answer = "220 ";
     else
       answer = "500 ";


     if ( debug ) System.out.println( "Connection::markMsgs - End   <---" );
     return( answer );
  }

   //-------------------------------------------------------------------------
   // Method: newMsgs( String )
   //-------------------------------------------------------------------------
   public String newMsgs( String aUser ) {
     if ( debug ) System.out.println( "Connection::newMsgs - Start --->" );

     String  answer;
     boolean newMsgs = false;

     if ( getUserMessages( aUser ) )
       {
       if ( msgCollection.size() > 0 )
         {
          for ( int i = 0; i < msgCollection.size(); i++ )
            {
            if ( ((RemMessage)msgCollection.elementAt( i )).status( ) == 'n' )
              {
              newMsgs = true;
              break;
              }
           }
          if ( newMsgs )
            answer = "220 ";          // User has new messages
          else
            answer = "500 ";          // User has messages, but none are new
         }
       else
         answer = "221 ";
       }
     else
       answer = "501 ";


     if ( debug ) System.out.println( "Connection::newMsgs - End   <---" );
     return( answer );
  }

  //-------------------------------------------------------------------------
  // Method: getUserMessages( String )  (was createCollection)
  //-------------------------------------------------------------------------
  public boolean getUserMessages( String aUser ) {
    if ( debug ) System.out.println( "Connection::getUserMessages - Start --->" );

    boolean      answer = true;

    msgCollection.removeAllElements();

    try {
      user_collector = library.new_IALibInferenceCollector(top_collector,aUser);
      user_collector.setMetadataName('1');
      user_collector.get();

      String                   aLine = new String();
      String                   msgs = new String(user_collector.extractMetadata());
      BufferedReader           dataIn = new BufferedReader(new StringReader( msgs ));
      RemMessage               theMessage;
      while ( true )
        {
        aLine = dataIn.readLine();
        if ( aLine == null )
          break;
        theMessage = new RemMessage( aLine );
        msgCollection.addElement( theMessage );
        }
    }
    catch (IOException e) {
       user_collector = null;
       answer = false;
    }
    catch (IAError e) {
       answer = false;
    }

    if ( debug ) System.out.println( "Connection::getUserMessages - End   <---" );
    return( answer );
  }

  //-------------------------------------------------------------------------
  // Method: updateUserMsgs( String ) (was writeCollection)
  // Returns:  0  = Message log successfully updated in library
  //           1  = Update failed, version conflict (msgs changed after they
  //                were retrieved - must retrieve again and retry)
  //           2  = Update failed, notify user
  //-------------------------------------------------------------------------
  public int updateUserMsgs( String aUser ) {
    if ( debug ) System.out.println( "Connection::updateUserMsgs - Start --->" );

    int      answer = 0;

    if (user_collector != null) { // The new was successful during creation...
      if (!msgCollection.isEmpty()) {   // If msgs remain, update pers stor
        try
          {
            user_collector.setMetadata( msgCollection.print() );
            synchronized (top_collector) {
              user_collector.put();
              }
          }
        catch( IAError e)
          {
          if (e.msgnumber() == IAError.IA_E_LIBRARY_VERSION_ERROR)
             answer = 1;
          else
             answer = 2;
          }
        } else {                          // No msgs left, clean up pers stor
           try {
             IALibInferenceContentsElement element =
                 library.new_IALibInferenceContentsElement();
             user_collector.setCursorElementScope(IALibrary.IALibICCS_rulesets);
             if (user_collector.find(element,user_collector.extractName(),
                                     IALibrary.IALibICET_ruleset))  {
                 // No messages, but reminders exist. Delete persistent msgs
                 user_collector.delMetadata('1');
             } else {
                 // No messages, no reminders, delete user collector.
                 synchronized (top_collector) {
                   user_collector.del();
                   }
             }
           } catch (IAError e) {
             answer = 2;                  // Update failed, notify user
           }
        }                                 // End of no msgs left
      }
    else {              // Collector creation failed, nothing to update
      answer = 2;
      }

    if ( debug ) System.out.println( "Connection::updateUserMsgs - End   <---" );
    return( answer );
  }

/*****************************************************************************
 *  End of methods pulled from the MsgCollection class.
 *****************************************************************************
 */

  private String convertNLs( String aline ) {
    String workline = "";
    int startIndex = 0;
    int index;

    while ((index = aline.indexOf('\\',startIndex)) != -1) {  // While escapes
       if (aline.charAt(index+1) == 'n') {                    // If \\n
         workline += aline.substring(startIndex, index);      // Get up to \\n
         workline += "\n";                                    // Add \n
       }
       startIndex = index+2;                                  // Bump past \\n
    }
    if (workline == "")                                       // If no escapes
       workline = aline;                                      // return orig
    else
       workline += aline.substring(startIndex);               // Get remaining
    return workline;                                          // return convrtd
  }

}  // end of class Connection

// Reminder class
class Reminder {
   String NL;
   String remStr;
   String remMsg;
   String idStr;
   String idStr2;
   String typeStr;
   String nameStr;
   String monthStr;
   String dayStr;
   String yearStr;
   String dateStr;
   String countStr;
   String countStr1;
   int icount;
   String viaStr;
   String userIDStr;
   String userEmailStr;
   String timeStr = "10:00:00";

   Reminder(String rem) {
      remStr = rem;
      StringTokenizer st = new StringTokenizer(rem, "#=");
      String attr, val;
      try {
         attr = st.nextToken();
         val  = st.nextToken();
         idStr = val;
         idStr2 = idStr + "Y";
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         typeStr = val;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         nameStr = val;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         monthStr = val;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         dayStr = val;
         if (1 == dayStr.length()) dayStr = "0" + dayStr;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         yearStr = val;
         dateStr = yearStr + "-" + monthToNum(monthStr) + "-" + dayStr;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         countStr = val;
         // Following is to bump the value in countStr by 1 because we start
         // alarms (from the rules) on the day before the lead time given by
         // the client (alarm does not go off until the following day).
         icount = Integer.parseInt(countStr);
         icount++;
         countStr1 = new String(Integer.toString(icount));
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         viaStr = val;
         if (viaStr.equals("E-Mail")) NL = "\\n";  // \n to e-mail
         else NL = "\\\\n";                        // \\n to msg adapter
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         userIDStr = val;
         System.out.println(attr + "\t" + val);
         attr = st.nextToken();
         val  = st.nextToken();
         userEmailStr = val;
         System.out.println(attr + "\t" + val);
      } catch(NoSuchElementException e) {
         System.out.println("Server: error decoding reminder String: " + e);
      }
      GregorianCalendar d = new GregorianCalendar();

      if ( d.get(Calendar.HOUR_OF_DAY)  < 10 )
        timeStr = "0";
      timeStr += (new Integer(d.get(Calendar.HOUR_OF_DAY) ).toString()) + ":";

      if ( d.get(Calendar.MINUTE) < 10 )
        timeStr += "0";
      timeStr += (new Integer(d.get(Calendar.MINUTE)).toString()) + ":";

      if ( d.get(Calendar.SECOND) < 10 )
        timeStr += "0";
      timeStr += (new Integer(d.get(Calendar.SECOND)).toString() );

      remMsg = createReminderMessage();
   }

   // Create message to be sent to the user for this reminder.
   String createReminderMessage() {
      String msgStr;
      if ( viaStr.equals("E-Mail") )
         {
         msgStr = "Hello " + userIDStr + ": " + NL + NL +
            "This is a reminder that on " +
            monthStr + " " + dayStr + NL +
            "you have the following " + typeStr +
            " event planned: " + NL + nameStr;
         }
      else
         {
         msgStr = "On " + monthStr + " " + dayStr +
                  " you have the following " + typeStr +
                  " event planned:<BR>" + nameStr;
         }
      return msgStr;  //temporarily until ABE fixes new-line in rule
   }

   // Create the first setAlarm rule and return a rule string

   String getAlarmSet1Rule() {
      String ruleStr =
         "(=> (AND (OR (EventName \"AGENT:STARTING\") " +
                      "(EventName \"AGENT:CSCHANGE\")) " +
                  "(addTimeInterval " + "\"" + dateStr + "\" " +
                        "\"" + "00:00:01" + "\"" + " -" +
                        countStr1 + " \"days\" ?date ?time))"  +
                  "(setRandomStartBoundedIntervalAlarm " +
                     "1 \"years\" ?date ?time 21600 " +
                     "\"" + idStr2 + "\"" + " \"\" ))";
      return ruleStr;
   }

   // Create the second setAlarm rule and return a rule string

   String getAlarmSet2Rule() {
      String ruleStr =
         "(=> (AND (AND (OR (EventName \"AGENT:STARTING\") " +
                  "(EventName \"AGENT:CSCHANGE\")) " +
                  "(addTimeInterval " + "\"" + dateStr + "\" " +
                        "\"" + "00:00:01" + "\"" + " -" +
                        countStr1 + " \"days\" ?date ?time)) "  +
                  "(currentTimeBetween ?date \"00:00:01\" " +
                        "\"" + dateStr + "\" " + "\"06:00:00\")) " +
                  "(setRandomBoundedIntervalAlarm 1 \"days\" " +
                        "?date \"00:00:01\" " +
                        "\"" + dateStr + "\" \"06:10:00\" 21600 " +
                        "\"" + idStr + "\"" + " \"\" ))";
      return ruleStr;
   }

   // Create the third setAlarm rule and return a rule string

   String getAlarmSet3Rule() {
      String ruleStr =
         "(=> (AND (AND (AND (EventName \"Time:Alarm\") " +
                  "(AlarmId " + "\"" + idStr2 + "\"" + ")) " +
                  "(EventTime ?d ?t)) " +
                  "(addTimeInterval ?d ?t " + countStr1 + " \"days\" " +
                        "?newdate ?newtime)) " +
                  "(setEndBoundedIntervalAlarm " +
                        "1 \"days\" ?newdate ?newtime " +
                        "\"" + idStr + "\"" +  " \"\" ))";
      return ruleStr;
   }


   // Create and return a rule string
   String getAlarmPopRule() {
      String ruleStr;
      if (viaStr.equals("E-Mail")) {
         ruleStr = "(=> (AND (EventName \"Time:Alarm\") " +
           "(AlarmId " + "\"" + idStr + "\"" + "))" +
           "(EMailWithSubject " + "\"" + userEmailStr + "\"" + " " +
           "\"" + remMsg + "\"" + " " + "\"" + nameStr + "\"" + "))";
      } else {
         ruleStr = "(=> (AND (EventName \"Time:Alarm\") " +
             "(AlarmId " + "\"" + idStr + "\"" + ")) " +
             "(SetMessage " + "\"" + remMsg + "\"))";
      }
      return ruleStr;
   }


   // Convert month string ("January - "December") to Numeric string
   String monthToNum(String m) {
      if (m.equals("January")) return "01";
      else if (m.equals("February")) return "02";
      else if (m.equals("March")) return "03";
      else if (m.equals("April")) return "04";
      else if (m.equals("May")) return "05";
      else if (m.equals("June")) return "06";
      else if (m.equals("July")) return "07";
      else if (m.equals("August")) return "08";
      else if (m.equals("September")) return "09";
      else if (m.equals("October")) return "10";
      else if (m.equals("November")) return "11";
      else if (m.equals("December")) return "12";
      else return null;
   }

   // Convert year string "1996" to abbreveated year "96" string
   String shortYear(String y) {
      if (4 <= y.length()) return y.substring(2);
      if (3 == y.length()) return y.substring(1);
      return y;
   }

} //end Reminder class
