/*************************************************************************
** File:         message.cpp
** Description:  Implementation of a message adapter
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
#include <message.h>
#include <iatk/types.h>

#ifdef _AIX
   #define IA_AIX
   #define IA_UNIX
#endif
#ifdef IA_UNIX
// Character which separates directory names in a file path specification
#define FILEPATH_SEP    "/"
#define FILEPATH_SEPCHR '/'
#else

#define FILEPATH_SEP    "\\"
#define FILEPATH_SEPCHR '\\'
#endif

// ----------------------------------------------------------------------
//  constructor
// ----------------------------------------------------------------------

MessageAdapter::MessageAdapter(char * parms):IAAdapter()
  {
    I0String libList(parms);
    I0String type=libList.word(1);
    I0String dllname=libList.word(2);
    I0String map=libList.word(3);
    if (map[map.length()-1] != FILEPATH_SEPCHR)
      map += FILEPATH_SEP;   // Append final path separator
    I0String collector=libList.word(4);
   int levels;
   if (libList.numWords() == 5)            // If levels was specified...
      levels = (libList.word(5)).asInt();    // 5th arg= # collector levels
   else levels = 1;                       // Otherwise default to 2 levels
   myLib=new IALibrary();
   myLib->attachImplementation(map,type,dllname,collector,levels);
   myCollector=myLib->getTop();
  }

MessageAdapter::~MessageAdapter() { delete myLib; }
//-----------------------------------------------------------------------
// identify - Register our Sensors and Effectors
//
// Invoke registerProcedure for each Sensor and Effector that we provide,
// so that they may be invoked by the inference engine later.  If any of
// these attempts to register a procedure fails, return a false (zero)
// to indicate that our identify() failed; otherwise return true (non-zero).
//
// registerProcedure is a service method provided by the IAAdapter class
// and inherited by SampMailAdapter(); the adapter writer does not have
// to provide it.
//
// To add a new Sensor or Effector to this adapter:
// . Add a "<procName>_Token" constant to sampmail.h
// . Assign the new "<procName>_Token" a value in sampmail.cpp
// . Add a registerProcedure() call to SampMailAdapter::identify()
// . Add code to run sensor/effector to SampMailAdapter::answerQuery() or
//   SampMailAdapter::performAction(), as appropriate
//-----------------------------------------------------------------------
IABool MessageAdapter::identify() {

   // Register our Sensors
   // No sensors to register

   // Register our Effectors
   registerProcedure("SetMessage",  IAEffectorProc,
      "String", SETMESSAGE_TOKEN);
   return true;                   // Success
}

//-----------------------------------------------------------------------
// performAction - Execute Effector procedure
//
//-----------------------------------------------------------------------
void MessageAdapter::performAction(IAProcToken token,
   const IAEventHeader &hdr, const char *binding) {

   // Build the atom from the binding string
   IAAtom atom;
   atom.setKIF(binding);

   if (token == SETMESSAGE_TOKEN) {
      // First build Reminder Message
      time_t temp;
      time(&temp);
      struct tm* timeptr;
      char dest[22];
      timeptr=localtime(&temp);
      I0String theTime("");
      I0String theDate("");
      I0String theZone("");
      if (timeptr==NULL) cout << "Bad Time" << endl;
      else
        {
         strftime(dest,sizeof(dest)-1,"%m/%d/%EY",timeptr);
         theDate=dest;
         strftime(dest,sizeof(dest)-1,"%H:%M",timeptr);
         theTime=dest;
         strftime(dest,sizeof(dest)-1,"%Z",timeptr);
         theZone=dest;
        }
      I0String messageId=((I0String)(long)temp)+hdr.getId().asString();
      I0String messageText= atom.getStringTerm(1);
      I0String wholeMessageLine="n "+messageId+" "+theDate+" "+theTime+" "+theZone+" "+messageText+"\n";
      // new up collector based on selector and set metadataname
      // figure out error handling
      IALibInferenceCollector* currentCollector=
           myLib->new_IALibInferenceCollector(myCollector,hdr.getSelector());
      currentCollector->setMetadataName('1');
      // Now do get / put sequence
      IABool recordWritten=false;
      IABool metadataExists=true;
      I0String totalMetadata;
      while (!recordWritten) {
         try {
            currentCollector->get();
         }
         catch (IAError &e) {

           if (e.msgnumber()==IA_E_LIBRARY_COLLECTOR_METADATA_EXISTENCE_ERROR)
              metadataExists=false;
           else {
              cerr << "Error on collector: " << hdr.getSelector() <<
                   " : " << e.errormsg() << endl;
              myLib->delete_IALibInferenceCollector(currentCollector);
              return;
           }
         }
         if (metadataExists)
            totalMetadata=currentCollector->extractMetadata()+wholeMessageLine;
         else totalMetadata=wholeMessageLine;
         currentCollector->setMetadata(totalMetadata);
         try {
            currentCollector->put();
            recordWritten=true;
         }
         catch (IAError &e) {
            if (e.msgnumber()!=IA_E_LIBRARY_VERSION_ERROR) {
               cerr << "Unexpected error on put " << e.errormsg() << endl;
               recordWritten=true;
            }
            else cout << "version error" << endl;
         }
      }
      // delete collector
      myLib->delete_IALibInferenceCollector(currentCollector);
   }   // end if
}

//=======================================================================
// Global function newAdapter()
//
// Function newAdapter() is a "factory" function for creating new
// instances of SampMail Adapters.
//
// Input:
//    *domain  = Domain name that identifies the type of concrete
//               Adapter requested.  Only "SampMail" is recognized.
//    *parms   = Adapter start-up parameters obtained from the Agent
//               configuration file.  These are not used by SampMail.
//
// The adapter instance created by a call to this function is owned by
// the IATK framework and should NOT ever be deleted by code in the
// adapter module.
//=======================================================================
IAAdapter * newAdapter(const char *domain, void *parms) {

   if (strcmp(domain, "Messages")) {
      cerr << "I was called to create an adapter of domain " << domain <<
         "only  \"Messages\" is valid." << endl;
      return NULL;
   }

   return new MessageAdapter((char *)parms);
}
