// -----------------------------------------------------------------------
// (C) Copyright IBM Corp, 1996
//
// File:        StockAdapter.java
// Version:     1.0 (sample) 6 August 1996
//
// Description: This class extends IAAdapter and performs all tasks required
//              to register and act as the StockAdapter Adapter.
//
// DISCLAIMER OF WARRANTIES:
// -------------------------
// The following [enclosed] code is sample code created by IBM
// Corporation.  This sample code is not part of any standard IBM product
// and is provided to you solely for the purpose of assisting you in the
// development of your applications.  The code is provided "AS IS",
// without warranty of any kind.  IBM shall not be liable for any damages
// arising out of your use of the sample code, even if they have been
// advised of the possibility of such damages.
// -----------------------------------------------------------------------
package ibm.AgentBuilder.samples.stock;

import ibm.AgentBuilder.aeapi.*;
import ibm.AgentBuilder.misc.IAError;
import ibm.AgentBuilder.Adapter.IAAdapter;

import ibm.AgentBuilder.samples.stock.*;
// import StockClassLoader;
// import MonitorThread;
// import MonitorSC;
// import ParseSC;
// import StockAdapter;

public class StockAdapter extends IAAdapter {

   // -----------------------------------------------------
   // New Adapter
   // -----------------------------------------------------
   // -----------------------------------------------------
   // Static routine called by the agent to construct this adapter
   // -----------------------------------------------------
   public static IAAdapter newAdapter(String domain, String parms) throws IAError {

     if (!domain.equals("StockAdapter")) {
        throw new IAError(newAdapterError,
           "newAdapter() does not support domain " +
                     domain  + "; only \"StockAdapter\" is valid.");
     }

     return new StockAdapter();
   }

   // -----------------------------------------------------
   // Constructor()
   // In this case, we simply call the IAAdapter constructor.
   // -----------------------------------------------------
   public StockAdapter() {
      super();
   }

   // -----------------------------------------------------
   // IDENTIFY() - Adapter Control Callback Function
   // General registration methods and other identification
   // tasks.
   // RegisterProcdure prototype is:
   //   RegisterProcedure(char procedureName,
   //                     IAProcType type,
   //                     char invocationSignature,
   //                     IAProcToken token)
   // -----------------------------------------------------
   public boolean identify() {

     // Registration tasks.
     try {

       // Register Effectors.
       registerProcedure("StockAdapterShow",
                          IAEffectorProc,
                          "String String String",
                          new IAProcToken(StockAdapterShow_Token));
       registerProcedure("StockAdapterMonitorClass",
                          IAEffectorProc,
                          "String",
                          new IAProcToken(StockAdapterMonitorClass_Token));
       registerProcedure("StockAdapterURL",
                          IAEffectorProc,
                          "String",
                          new IAProcToken(StockAdapterURL_Token));
       registerProcedure("StockAdapterParserClass",
                          IAEffectorProc,
                          "String",
                          new IAProcToken(StockAdapterParserClass_Token));
       registerProcedure("StockMonitor",
                          IAEffectorProc,
                          "String",
                          new IAProcToken(StockMonitor_Token));
       registerProcedure("StockAdapterProxy",
                          IAEffectorProc,
                          "String Integer",
                          new IAProcToken(StockAdapterProxy_Token));

       // Register Boolean Sensors.
       registerProcedure("StockCompare",
                          IABooleanSensorProc,
                          "String String String",
                          new IAProcToken(StockCompare_Token));
       registerProcedure("IntegerCompare",
                          IABooleanSensorProc,
                          "Integer String Integer",
                          new IAProcToken(IntegerCompare_Token));

       // Register NON Boolean Sensors.  i.e. Sensors that return a factset.
       // The one term in this sensor may be a variable for us to replace
       registerProcedure("GetStockEventCount",
                          IASensorProc,
                          "IntegerLvar",
                          new IAProcToken(GetStockEventCount_Token));

     // Errors?
     } catch (Exception e) {
       System.out.println("StockAdapter:identify() Exception: " +
                           e.getMessage());
       return false;
     }
     return true;
   }

   // -----------------------------------------------------
   // START() - Adapter Control Callback Function
   // General 'Begin active adapter processing' routine.
   // Start tells the adapter it is allowed to start notification of events
   // although before this point it can process sensors and effectors
   // Start will set a flag to allow us to issue notify(IATriggerEvent)
   // -----------------------------------------------------
   public boolean start() {
      started = true;
      return super.start();
   }


   // -----------------------------------------------------
   // STOP() - Adapter Control Callback Function
   // General 'Stop active adapter processing' routine.
   // Stop will set the flag to disallow event notification to the agent
   // by this adapter
   // -----------------------------------------------------
   public boolean stop() {
      started = false;
      return super.stop();
   }


   // -----------------------------------------------------
   // SHUTDOWN() - Adapter Control Callback Function
   // General 'Terminate processing' routine and cleanup.
   // May be left to default to super's routine.
   // -----------------------------------------------------
   public void shutdown() {
      monitorObj = null;
      parseObj = null;
      super.shutdown();
   }

   // -----------------------------------------------------
   // eventComplete(IAEventHeader) - Adapter Control Callback Function
   // This routine is called each time the agent finishes processing
   // an event from any domain (adapter) and can be used to flush data
   // related to a particular event
   // -----------------------------------------------------
   public void eventComplete(IAEventHeader hdr) {
      String myDomain;

      myDomain = hdr.getDomain();

      // check to see if this event complete is for an event in our domain
      // (our adapter) and then increment our count of events processed by
      // this adapter
      if (myDomain.equals("StockAdapter")) {
         eventCount++;
      }

   }

   // -----------------------------------------------------
   // PerformAction(IAProcToken, IAEventHeader, Atom)
   // Perform effector action.
   // In most cases, this routine would be designed to work
   // based on the value of the provided token (i.e. a 'switch'
   // statement).
   // -----------------------------------------------------
   public void performAction(IAProcToken token,
                             IAEventHeader hdr,
                             String binding)        throws IAError {

      IAAtom iAtom;

      try {
         // create an atom from the binding String in KIF format for ease
         // of processing
         iAtom = new IAAtom();
         iAtom.setKIF(binding);

         // switch on the IAProcToken to determine the correct effector
         // IAProcToken is set in the registerProcedure call
          switch(token.intValue()) {

             case StockAdapterShow_Token: {
                stockAdapterShow(iAtom);
                break;
             }
             case StockAdapterMonitorClass_Token: {
                stockAdapterMonitorClass(iAtom);
                break;
             }
             case StockAdapterURL_Token: {
                stockAdapterURL(iAtom);
                break;
             }
             case StockAdapterParserClass_Token: {
                stockAdapterParserClass(iAtom);
                break;
             }
             case StockMonitor_Token: {
                stockMonitor(iAtom);
                break;
             }
             case StockAdapterProxy_Token: {
                stockAdapterProxy(iAtom);
                break;
             }

             default:
                throw new Exception("Unknown Effector, token= " +
                                    String.valueOf(token.intValue()));
          }

      } catch (Exception e) {
         throw new IAError(performActionError,
               "StockAdapter:performAction() Exception: " +
                e.getMessage());
      }

     return;
   }

   // -----------------------------------------------------
   // answerQuery(IAProcToken, IAEventHeader, Atom)
   // Invoke a non BOOLEAN sensor procedure.
   // For each term that is registered as Lvar(logical variable),
   // if it is a variable in this invocation, we find all
   // values which make a true fact from the binding String
   // and construct the facts to return.
   // If no values for the variable make a true fact, return an empty
   // fact set.
   // Even though we registered the term as a variable it may not be
   // a variable on any particular invocation of answerQuery, in which
   // case we determine if the particular value of the term makes a true
   // fact of the binding String and return that fact, if not, we return
   // an empty fact set.
   //
   // -----------------------------------------------------
   public IAFactSet answerQuery(IAProcToken token, IAEventHeader hdr, String binding)
                  throws IAError {

      IAAtom    iAtom;
      IAAtom    rAtom;
      IAFactSet returnFactSet;

      try {
         switch(token.intValue()) {
            // in this case we have only one IAProcToken for answerQuery
            case GetStockEventCount_Token: {

               // Create an atom from the input binding String
               iAtom = new IAAtom();
               iAtom.setKIF(binding);

               // Create a true fact for GetStockEventCount
               rAtom = new IAAtom();
               rAtom.setPredicate(iAtom.getPredicate());
               rAtom.addTerm(eventCount);

               // our return fact set
               returnFactSet = new IAFactSet();

               // If the term was a variable just add the true fact we
               // created
               if (iAtom.isLogicalVariable(1, "integer")) {
                  returnFactSet.add(rAtom);

               // otherwise we need to check if the one and only one term
               // in the fact we created matches the value sent in the
               // binding String and add the fact if we have a match
               } else {
                  if (iAtom.getIntegerTerm(1) == rAtom.getIntegerTerm(1) ) {
                     returnFactSet.add(rAtom);
                  }
               }
               return returnFactSet;
            }
            default:
               throw new Exception("Unknown Sensor, token = " +
                                    String.valueOf(token.intValue()));
         }

      } catch (Exception e) {
          throw new IAError(answerQueryError,
                "StockAdapter:answerQuery() Exception: " +
                 e.getMessage());
      }

   }


   // -----------------------------------------------------
   // testCondition(IAProcToken, IAEventHeader, Atom)
   // Invoke a BOOLEAN sensor procedure.
   // In most cases, this routine would be designed to work
   // based on the value of the provided token (i.e. a 'switch'
   // statement).
   // An atom is constructed from the binding, so the individual
   // routines can extract the terms and determine whether the sensor
   // is true or false
   // -----------------------------------------------------
   public boolean testCondition(IAProcToken token, IAEventHeader hdr, String binding)
                  throws IAError {

      IAAtom iAtom;

      try {

        iAtom = new IAAtom();
        iAtom.setKIF(binding);

        switch(token.intValue()) {
          case StockCompare_Token: {
            return stockPriceCompare(iAtom);
          }
          case IntegerCompare_Token: {
            return integerCompare(iAtom);
          }
          default:
            throw new Exception("Unknown Boolean Sensor, token = " +
                                 String.valueOf(token.intValue()));
        }

      } catch (Exception e) {
          throw new IAError(testConditionError,
              "StockAdapter:testCondition() Exception: " +
               e.getMessage());
      }

   }

   // -----------------------------------------------------
   // Other functions as required.....
   // -----------------------------------------------------

   // Given a class name, create a new instance of our monitor object.
   private void stockAdapterMonitorClass(IAAtom iAtom) throws IAError {

      StockClassLoader classLoader;

      try {
        classLoader = new StockClassLoader();
        Class newClass = classLoader.loadClass(iAtom.getStringTerm(1));
        monitorObj = (MonitorSC)newClass.newInstance();
      } catch (Exception e) {
         throw new IAError(monitorClassLoadError,
            "StockAdapter:stockAdapterMonitorClass() Exception: " +
             e.getMessage());
      }

   }

   // Set the URL within the Monitor object.
   private void stockAdapterURL(IAAtom iAtom) throws IAError {

      try {
         monitorObj.setURL(iAtom.getStringTerm(1));
      } catch (Exception e) {
         throw new IAError(URLSetError,
              "StockAdapter:stockAdapterURL() Exception: " +
               e.getMessage());
      }

   }

   // Set the proxy within the Monitor object.
   private void stockAdapterProxy(IAAtom iAtom) throws IAError {

      try {
         monitorObj.setProxy(iAtom.getStringTerm(1));
         monitorObj.setProxyPort(iAtom.getIntegerTerm(2));
      } catch (Exception e) {
         throw new IAError(proxySetError,
             "StockAdapter:stockAdapterProxy() Exception: " +
              e.getMessage());
      }

   }


   // Given a class name, create a new instance of our parser object.
   private void stockAdapterParserClass(IAAtom iAtom) throws IAError {

      StockClassLoader classLoader;

      try {
         classLoader = new StockClassLoader();
         Class newClass = classLoader.loadClass(iAtom.getStringTerm(1));
         parseObj = (ParseSC)newClass.newInstance();
      } catch (Exception e) {
         throw new IAError(parserClassLoadError,
            "StockAdapter:stockAdapterParserClass() Exception: " +
             e.getMessage());
      }

   }

   // Given a String.. show it on the console for debug operations.
   private void stockAdapterShow(IAAtom iAtom) throws IAError  {
      System.out.println(iAtom.getStringTerm(1) + iAtom.getStringTerm(2) +
                         iAtom.getStringTerm(3));
   }

   // Issue a Monitor request for the given list of stock identifiers.
   private void stockMonitor(IAAtom iAtom) throws IAError {


      // First check to see if the adapter is started.  If it isn't
      // we won't start a monitor thread, because that would do a notify
      // Start an async thread to do the monitor call.
      // Pass a referance to our current monitor and parse object.
      if (started = true) {
         try {
            // MonitorThread creates facts and notifies the agent of the
            // Stock Price Event
            MonitorThread ourThread = new MonitorThread(this, monitorObj, parseObj, iAtom.getStringTerm(1));
            ourThread.start();
         } catch (Exception e) {
            throw new IAError(threadStartError,
               "StockAdapter:stockMonitor() Exception: " +
                e.getMessage());
         }
      }
      return;
   }

   // Given two operands and an operator, compare and return.
   private boolean stockPriceCompare(IAAtom iAtom) {

      String iOperator;
      Double opA;
      Double opB;

      try {
        // get the three terms from the atom: value1 operator value2
         opA = new Double(iAtom.getStringTerm(1));
         opB = new Double(iAtom.getStringTerm(3));
         iOperator = iAtom.getStringTerm(2);

      } catch (Exception e) {
         // Mal formed strings?
         System.out.println("StockAdapter:stockPriceCompare() Exception: " +
                             e.getMessage());
         return false;
      }

      if ( iOperator.equals("=") || iOperator.equals("==")) {
        return (opA.doubleValue() == opB.doubleValue());
      } else
      if ( iOperator.equals(">")) {
        return (opA.doubleValue() > opB.doubleValue());
      } else
      if ( iOperator.equals("<")) {
        return (opA.doubleValue() < opB.doubleValue());
      } else
      if ( iOperator.equals(">=")) {
        return (opA.doubleValue() >= opB.doubleValue());
      } else
      if ( iOperator.equals("<=")) {
        return (opA.doubleValue() <= opB.doubleValue());
      }

      return true;
   }

   // Given two operands and an operator, compare and return.
   private boolean integerCompare(IAAtom iAtom) {

      String iOperator;
      int opA;
      int opB;

      try {
        // get the three terms from the atom: value1 operator value2
         opA = iAtom.getIntegerTerm(1);
         opB = iAtom.getIntegerTerm(3);
         iOperator = iAtom.getStringTerm(2);
      } catch (Exception e) {
         // Mal formed strings?
         System.out.println("StockAdapter:integerCompare() Exception: " +
                             e.getMessage());
         return false;
      }

      if ( iOperator.equals("=") || iOperator.equals("==")) {
        return (opA == opB);
      } else
      if ( iOperator.equals(">")) {
        return (opA > opB);
      } else
      if ( iOperator.equals("<")) {
        return (opA < opB);
      } else
      if ( iOperator.equals(">=")) {
        return (opA >= opB);
      } else
      if ( iOperator.equals("<=")) {
        return (opA <= opB);
      }

      return true;
   }
   // used by MonitorThread to check if we are still enabled when notify fails
   boolean isStarted()
     { return started; }

   // -----------------------------------------------------
   // Java Class Private Data Area:
   // -----------------------------------------------------

   // Tokens for our eff/sen routines.
   final static private int StockAdapterMonitorClass_Token = 1;
   final static private int StockAdapterURL_Token = 2;
   final static private int StockAdapterParserClass_Token = 3;
   final static private int StockMonitor_Token = 4;
   final static private int StockCompare_Token = 5;
   final static private int StockAdapterShow_Token = 6;
   final static private int GetStockEventCount_Token = 7;
   final static private int IntegerCompare_Token = 8;
   final static private int StockAdapterProxy_Token = 9;

   // IAError numbers

   final static private int newAdapterError = 201;
   final static private int performActionError = 202;
   final static private int answerQueryError = 203;
   final static private int testConditionError = 204;
   final static private int parserClassLoadError = 205;
   final static private int monitorClassLoadError = 206;
   final static private int URLSetError = 207;
   final static private int proxySetError = 208;
   final static private int threadStartError = 209;

   // Our monitor and parser objects.
   private MonitorSC      monitorObj;
   private ParseSC        parseObj;

   boolean                started = false;
   int                    eventCount = 0;

}  // end of java class StockAdapter


// --------------------------------------------------------------------

