/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: alarm.cpp
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/alarm.cpp, tka_timer, abe.l6
// VERSION: 1.10.1.3 97/06/20 15:13:25
//
// PURPOSE:  Alarm class implementation
//
// DESCRIPTION:
//    Abstract class defining alarm used by set alarm effector, used in
//     openclass collection
//
// CLASSES DECLARED:
//    Alarm- Abstract class
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $A1   960909 WES  f304   Add support for selector field in alarms
// $A2   960914 WES  f304   respond to changes in IAAtom
// $A3   961206 WES  f512   React to IAAtom and IAFactSet throws
// $A4   970226 WES  f658   Additional mapping for alarms
// $A5   970225 WES  f664   Persistent alarms
// $A6   970512 WES         eventNumber
//******************************************************************************

#include <iapl/common.h>

static char _partid_[] = "@(#) 1.10.1.3 src/adapters/timer/alarm.cpp, tka_timer, abe.l6";

#include <iatk/iaerrors.h>
#include <string.h>
#include <alarm.h>

// 0 indicates persistent Version hasn't been restored yet                //@A5
int Alarm::persistentVersion=0;                                           //@A5
const int Alarm::version=1;                                               //@A5
                                                                          //@A5

// no other static variables need to be saved?                            //@A5
boolean Alarm::versionAdded=                                              //@A5
                     PersistentObject::addStaticVars("Alarm",             //@A5
                     &restoreStaticVars,                                  //@A5
                     &saveStaticVars);                                    //@A5



// only static variable from persistence is currently persistentVersion   //@A5
boolean Alarm::restoreStaticVars(PersistentFile& objectFile)              //@A5
    {                                                                     //@A5
      if (!objectFile.read(persistentVersion)) return false;              //@A5
      return objectFile.readEnd();                                        //@A5
    }                                                                     //@A5

// only static variable from persistence is currently persistentVersion   //@A5
boolean Alarm::saveStaticVars(PersistentFile& objectFile)                 //@A5
    {                                                                     //@A5
      if (!objectFile.write(version)) return false;                       //@A5
      return objectFile.writeEnd();                                       //@A5
    }                                                                     //@A5

boolean Alarm::save(PersistentFile &objFile) const                        //@A5
          {                                                               //@A5
            if (!alarmTime.save(objFile))                                 //@A5
              return false;                                               //@A5
            if (!objFile.write(countOfDuplicates))                        //@A5
              return false;                                               //@A5
            if (!alarmUid.save(objFile)) return false;                    //@A5
            if (!objFile.write(alarmCategory)) return false;                 //@A6
            if (!objFile.write(flag)) return false;                       //@A6
            return objFile.write(eventNumber);                            //@A6
          }                                                               //@A5

boolean Alarm::restore(PersistentFile &objFile)                           //@A5
          {                                                               //@A5
            int tempflag;                                                 //@A6
            if (persistentVersion > version) {                            //@A5
               cout << "Alarm Versions not compatible" << endl;           //@A5
               return false;                                              //@A5
            }                                                             //@A5
            if (!alarmTime.restore(objFile))                              //@A5
              return false;                                               //@A5
            if (!objFile.read(countOfDuplicates))                         //@A5
              return false;                                               //@A5
            if (!alarmUid.restore(objFile)) return false;                 //@A5
            if (!objFile.read(alarmCategory)) return false;                  //@A6
            // flag currently is not persistent, so read into temp        //@A6
            // variable                                                   //@A6
            if (!objFile.read(tempflag)) return false;                    //@A6
            return objFile.read(eventNumber);                             //@A6
          }                                                               //@A5

void Alarm::setTestStartTime(const AString& thisStartTime)
{
    alarmTime.setTime(thisStartTime);
}



//  ctor                                                                    @A1
Alarm::Alarm (const AlarmUniqueId& anId):                                 //@A4
   alarmTime(),countOfDuplicates(0),                                      //@A1
   alarmUid(anId),alarmCategory("all"),flag(0),eventNumber(0),               //@A6
   persistent(false)                                                      //@A6

{ }                                                                       //@A1

// Ctor - using time and alarm type                                         @A1
Alarm::Alarm (const IATime& aTime,
              const AlarmUniqueId& anId):                                 //@A4
       alarmTime(aTime),countOfDuplicates(0),                             //@A1
   alarmUid(anId),alarmCategory("all"),flag(0),eventNumber(0),               //@A6
   persistent(false)                                                      //@A6

{ }                                                                       //@A1

// Ctor - using start time, start date, and alarm type                      @A1
Alarm::Alarm (const AString& thisStartDate,
              const AString& thisStartTime,                               //@A1
              const AlarmUniqueId& anId):                                 //@A4
   alarmTime(getAlarmStartTime(thisStartDate,thisStartTime)),             //@A1
   countOfDuplicates(0),alarmUid(anId),                                   //@A4
   alarmCategory("all"),flag(0),eventNumber(0),                              //@A6
   persistent(false)                                                      //@A6

{ }                                                                       //@A1

// remove a token from the selector                                         @A1
// RC=0 token removed, no more exist (this alarm should be deleted)         @A1
// RC=1 token removed, rout tokens still exist                              @A1
int Alarm::removeSelectorToken(const AString& selectorToken)              //@A1
   {                                                                      //@A1
     // for performance of common case                                    //@A1
     if (selectorToken==alarmUid.getSelector())                           //@A4
        {                                                                 //@A1
          alarmUid.setSelector("");                                       //@A4
          return 0;                                                       //@A1
        }                                                                 //@A1
     else return 1;                                                       //@A1
// Need to change selector to alarmUid.getSelector() or setSelector()
//     for (int i=1;i<=selectorTokens.numWords();++i)                                                                      //@A1
//       {                                                                //@A1
//        // Do we have this token in the selector?                       //@A1
//        int tokenWordIndex=selector.wordIndexOfPhrase(selectorTokens.word(i)); //@A1
//       if (tokenWordIndex!= 0 )                                         //@A1                                          //@A1
//           {                                                            //@A1
//             int tokenIndex=
//             selector.indexOfWord(tokenWordIndex);       //@A1
//             int nextWord=selector.indexOfWord(tokenWordIndex+1);       //@A1
//            // if last word                                             //@A1
//             if (nextWord < 0) selector.remove(tokenIndex-1);           //@A1
//             // if not last word                                        //@A1
//             else selector.remove(tokenWordIndex,nextWord-tokenIndex);  //@A1
//           } // if                                                      //@A1
//       }  // for                                                        //@A1
//      if (selectorTokens.numWords()==0) return 0;                       //@A1
//      else return 1;
     }                                                                    //@A1

// Key Access - Key is IATime
// IATime const& Alarm::getAlarmTime() const {return alarmTime;}

// Alarm Methods

// set start date
   void Alarm::setStartDate(const AString& thisStartDate)
  {
     alarmTime.setDate(thisStartDate);
  }

// set start time
   void Alarm::setStartTime(const AString& thisStartTime)
  {
     alarmTime.setTime(thisStartTime);
  }

// decrease the count of the duplicate alarms
boolean Alarm::decrementDuplicateCount()
  {
   if (countOfDuplicates==0) return false;
   else countOfDuplicates--;
   return true; }

// Is the alarm given past the current time?
boolean Alarm::isExpired() const
    {
      IATime aTime;
      return (getAlarmTime()<=aTime);
    }

// Get the weekday fact
IAAtom Alarm::getWeekdayFact() const
   {
    IAAtom WeekdayFact;                                                   //@A2
    WeekdayFact.setPredicate("DayOfWeek");                                //@A3
    WeekdayFact.addTerm(getEventWeekday());
    return WeekdayFact;
   }

IAAtom Alarm::getAlarmIdFact() const
    {
      IAAtom AlarmIdFact;
      AlarmIdFact.setPredicate("AlarmId");                                //@A1
      AlarmIdFact.addTerm(getAlarmId());                                  //@A2
      return AlarmIdFact;
     }

IAAtom Alarm::getAlarmCategoryFact() const
    {
      IAAtom AlarmCategoryFact;
      AlarmCategoryFact.setPredicate("AlarmCategory");                             //@A1
      AlarmCategoryFact.addTerm(getAlarmCategory());                               //@A2
      return AlarmCategoryFact;
     }

// Get the alarm time fact
IAAtom Alarm::getAlarmTimeFact() const
   {
    IAAtom AlarmTimeFact;                                                 //@A2
    AlarmTimeFact.setPredicate("AlarmTime");                              //@A3
    AlarmTimeFact.addTerm(getEventDate());
    AlarmTimeFact.addTerm(getEventTime());
    return AlarmTimeFact;
   }

IAAtom Alarm::getIntervalsMissedFact() const
      {
        IAAtom IntervalsMissedFact;                                       //@A2
        IntervalsMissedFact.setPredicate("IntervalsMissed");              //@A3
        IntervalsMissedFact.addTerm(0);
        return IntervalsMissedFact;
      }

// default for the alarmTime is always now if start time not specified
// starttime=0, use this time; start date=0, use this date
IATime Alarm::getAlarmStartTime(const AString& thisStartDate,
                              const AString& thisStartTime)
{
 // Parse start times
 IATime alarmTime;
 AString dateDel(dateDelimeter);
 // Start time is not zero, set current time
 if (thisStartTime.indexOfAnyBut("0:")!=-1)
    alarmTime.setTime(thisStartTime);
 // Start date is not zero, set current time
 if (thisStartDate.indexOfAnyBut(dateDel+"0")!=-1)
    alarmTime.setDate(thisStartDate);

 return alarmTime;
}

// default for the startTime is time zero
// starttime=0, use zero time; start date=0, use zero date
IATime Alarm::getStartTime(const AString& thisStartDate,
                              const AString& thisStartTime)
{
 // Parse start times
 IATime alarmTime=IATime::getZeroTime();
 AString dateDel(dateDelimeter);
 // Start time is not zero, set current time
 if (thisStartTime.indexOfAnyBut("0:")!=-1)
    alarmTime.setTime(thisStartTime);
 // Start date is not zero, set current time
 if (thisStartDate.indexOfAnyBut(dateDel+"0")!=-1)
    alarmTime.setDate(thisStartDate);

 return alarmTime;
}

void Alarm::print(ostream& os) const
   {
     os << "Alarm Class: " << className()  << endl;                       //@A4
     os << "Alarm time:" << (char *)alarmTime.getDateTime() << endl;
     os << "Count of Duplicates: " << countOfDuplicates << endl;
     os << "UniqueID: " << alarmUid.getSelector() << " " <<               //@A4
        alarmUid.getId() << endl;                                         //@A4
   }
