#ifndef _ALARM_H
#define _ALARM_H

/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: alarm.h
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/alarm.h, tka_timer, abe.l6
// VERSION: 1.10.1.3 97/06/20 15:13:28
//
// INCLUDE AS:  #include <alarm.h>
//
// PURPOSE:  Alarm class definition
//
// DESCRIPTION:
//    Abstract class defining alarm used by set alarm effector, used in
//     openclass collection
//
// CLASSES DECLARED:
//    Alarm- Abstract class
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $A1   960909 WES  f304   Add support for selector in alarms
// $A2   961206 WES  f512   React to IAAtom and IAFactSet throws
// $A3 = 970225 WES  f658   Additional mapping for alarms
// $A4 = 970225 WES  f664   Persistent alarms
// $A5 = 970509 WES        reset effector support
//******************************************************************************

// Includes
#include <time.h>
#include <string.h>
#include <iapl/astring.h>
#include <iostream.h>
#include <iset.h>
#include <iatime.h>
#include <alarmuid.h>
#include <iatk/atom.h>
#include <persobj.h>
#include <persfile.h>

typedef AString EventType; // {oneTimeEvent,intervalEvent,identifiedEvent,identifiedUserEvent};
typedef int WeekDay;

class IAError;

class Alarm : public PersistentObject  {
   public:

   friend ostream& operator<<(ostream& os, const Alarm& a) {a.print(os);return os;}

      virtual ~Alarm() {}

      // Key Access - Key is IATime
      IATime const& getAlarmTime() const { return alarmTime; }

      // Are the two alarms equal?
      // Matching does not include selector                               //@A1
      virtual boolean matches(const Alarm* theAlarm) const                //@A3
        { return alarmUid==theAlarm->alarmUid; }                          //@A3

      boolean  matches(const AlarmUniqueId& alarmuid) const               //@A3
        { return alarmUid==alarmuid; }

      // Alarm Methods

      // Is this a recurring or interval alarm and is the next interval past the current
      // time < endTime
      virtual boolean isRecurring() const=0;

      virtual boolean isRecurring(IATime currentTime) const=0;            //@A4

      // Increment to next interval past current time and date
      // Return number of increments skipped
      // Assumption: alarm is incrementable not past ending time
      virtual int incrementAlarm() {return 0;}

      virtual int incrementAlarm(IATime currentTime) {return 0;}          //@A4

      // same result as increment alarm but may be more efficient if
      // hundreds of intervals have expired
      virtual int update(IATime currentTime) {return incrementAlarm(currentTime);} //@A5          //@A4

      virtual IAFactSet getFactSet() const=0;

      AString getEventType() const { return "Alarm";}

      // Is the alarm given past the current time?
      virtual boolean isExpired() const;

      // Create a copy of any type of alarm
      virtual Alarm* createCopy() const=0;

      virtual IntervalUnits getIntervalUnits() const=0;                   //@B2

      // Get the date of the alarm for the alarm time fact
      DateType getEventDate() const {return getAlarmTime().getDate();}

      // Get the date of the alarm for the alarm time fact
      TimeType getEventTime() const {return getAlarmTime().getTime();}

      // Get the timeStamp of the alarm
      TimeStamp getEventTimeStamp() const {return getAlarmTime().getDateTime();}

      // Get the weekday of the alarm
      WeekDay getEventWeekday() const {return getAlarmTime().weekday();}

      // Get the weekday fact
      IAAtom getWeekdayFact() const throw(IAError);                       //@A2

      IAAtom getAlarmIdFact() const throw(IAError);               //@A3

      IAAtom getAlarmCategoryFact() const throw(IAError);               //@A3

      // Get the alarmTime fact
      IAAtom getAlarmTimeFact() const throw(IAError);                     //@A2

      // Get the interval missed fact
      virtual IAAtom getIntervalsMissedFact() const throw(IAError);       //@A2

      // set test start time / used for testing
       void setTestStartTime(const AString& thisStartTime) throw(TimeFormatError);

      // Is a duplicate alarm valid - yes, for alarms with default identifiers
      // if alarmUid type is 'n', alarm is default identified
      boolean duplicatesAllowed() const {return alarmUid.getType()=='n';}

      // increase the count of the duplicate alarms
      void incrementDuplicateCount() {countOfDuplicates++;}

      // decrease the count of the duplicate alarms
      boolean decrementDuplicateCount();

      void setDuplicateCount(int duplicates) {countOfDuplicates=duplicates; } //@A5

      // duplicate events now never occur
      virtual boolean duplicateEventsPossible() const {return false;}

      // Do we have matching events? - Given that alarm time is the same
      // This function only has meaning when duplicateEventsAllowed
      virtual boolean eventMatches(const Alarm *) const {return false;}

      //All alarms have identifiers
      AString getAlarmId() const {return alarmUid.getId();}               //@A3

      //All alarms  have identifiers
      AlarmUniqueId const& getAlarmUniqueId() const {return alarmUid;}           //@A3

      // set Selector - no restriction here on selector                     @A1
      void setSelector(const AString& aSelector)
      { alarmUid.setSelector(aSelector); }                                //@A3

      void setEventNumber(long number) {eventNumber=number; }             //@A5

      long getEventNumber() {return eventNumber; }                        //@A5

      void setAlarmCategory(AString category) throw(AlarmCategoryError)            //@A5
        {
         if (category.length() > 1023)
            ITHROW(AlarmCategoryError((AString)"AlarmCategory name too long: \""+category+"\""));
         else if (category.numWords() != 1)
            ITHROW(AlarmCategoryError((AString)"AlarmCategory name contains blank: \""+category+"\""));
         else alarmCategory=category;
        }

      AString getAlarmCategory() const {return alarmCategory; }                 //@A5

      void setDeferredOn() { setFlagOn(1);}                               //@A5

      void setDeferredOff() {setFlagOff(1); }                             //@A5

      boolean isDeferred() { return allFlagsOn(1); }                    //@A5

      void setFlagOn(int mask) {flag=flag | mask; }                       //@A5

      void setFlagOff(int mask) {flag=(mask ^ -1 ) & flag; }              //@A5

      void setFlag(int newFlag) {flag=newFlag; }                          //@A5

      boolean allFlagsOn(int mask) { return ((mask & flag) == mask); }

      int getFlag() {return flag; }                                       //@A5

      void setPersistent(boolean newPersistent) { persistent=newPersistent; } //@A5

      boolean isPersistent() { return persistent; }

      // get Selector                                                       @A1
      AString getSelector() const
                         { return alarmUid.getSelector(); }               //@A3


      // boolean to check for Selector                                      @A1
      boolean hasSelector() const
                          { return alarmUid.getSelector()!=""; }          //@A3

      boolean isDummy() const {return alarmUid.isDummy(); }               //@A3

      // remove a token from the selector                                   @A1
      // currently only one token exists in selector                        @A1
      // RC=0 token removed, no more exist (this alarm should be deleted)   @A1
      // RC=1 token removed, selector tokens still exist                    @A1
      int removeSelectorToken(const AString& selector);                   //@A1

    // starttime=0, use this time; start date=0, use this date              @A1
    static IATime getAlarmStartTime(const AString& thisStartDate,         //@A1
                                    const AString& thisStartTime)         //@A1
          throw(TimeFormatError,DateFormatError);                         //@A1
                                                                          //@A1
    // starttime=0, use zero time; start date=0, use zero date              @A1
    static IATime getStartTime(const AString& thisStartDate,              //@A1
                               const AString& thisStartTime)              //@A1
          throw(TimeFormatError,DateFormatError);                         //@A1

    virtual AString className() const { return "Alarm";}                 //@A3

    virtual AString classId() const=0;                                   //@A4

    boolean isDummy() { return alarmUid.getType()=='d'; }                 //@A3

    // only static variable from persistence is currently persistentVersion //@A4
    static boolean restoreStaticVars(PersistentFile& objectFile);         //@A4

    // only static variable from persistence is currently persistentVersion //@A4
    static boolean saveStaticVars(PersistentFile& objectFile);            //@A4

    virtual boolean save(PersistentFile &objFile) const;                  //@A4

    virtual boolean restore(PersistentFile &objFile);                     //@A4

protected:

       // basis for default
      Alarm ():alarmTime(),countOfDuplicates(0),alarmUid(),alarmCategory("all"),
                         flag(0), eventNumber(0), persistent(false)  {}   //@A5

      //  ctor                                                              @A1
       Alarm (const AlarmUniqueId& alarmUid);                             //@A3

      // Ctor - using time and alarm type                                 //@A1
      Alarm (const IATime& aTime,
             const AlarmUniqueId& alarmUid);                              //@A3

      // Ctor - using start time, start date, and alarm type                @A1
      Alarm (const AString& thisStartDate, const AString& thisStartTime,  //@A1
             const AlarmUniqueId& alarmUid)                               //@A3
             throw(DateFormatError,TimeFormatError);                      //@A1

      // set start date
       void setStartDate(const AString& thisStartDate) throw(DateFormatError);

      // set start time
       void setStartTime(const AString& thisStartTime) throw(TimeFormatError);

      virtual void print(ostream& os) const;

      // Subclasses can set alarmTime
      void setAlarmTime(const IATime& thisTime) {alarmTime=thisTime;}

      // For use in building Fact Sets
      WeekDay getWeekDay() const {return alarmTime.weekday();}



private:
   IATime alarmTime;
   int countOfDuplicates;
   AlarmUniqueId alarmUid;                                                //@A3
   AString alarmCategory; // allows grouping of alarms for reset functions   //@A5
   int flag;       // used to mark alarm as to whether its deferred       //@A5
                   // currently 0 or 1 but room for expansion             //@A5
   long eventNumber; // event number alarm was saved on                   //@A5
   // next not saved persistently - extracted once from value of alarmCategory
   boolean persistent; // not saved
   static const int version;                                              //@A4
   static boolean versionAdded;                                           //@A4
   static int persistentVersion;                                          //@A4

};

#endif //_ALARM_DEF
