/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: alarmbin.cpp
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/alarmbin.cpp, tka_timer, abe.l6
// VERSION: 1.11.1.4 97/06/20 15:13:31
//
// PURPOSE:  Implementation of bounded interval alarm
//
// DESCRIPTION:
//    A bounded interval alarm - Interval alarm.  All non-identified
//    interval alarms start on predefined boundaries, starting at midnight
//    Bounded means that an ending time for discontinuing the alarm is
//    given when the alarm is constructed
//
// CLASSES IMPLEMENTED:
//    AlarmBoundedInterval
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $B1 = 960725 WES d219   Fix infinite loop when start > end times
//                         (both in future)
// $A1   960926 WES  f304   Add support for selector field in alarms
//
// $A2 = 970225 WES f658   Additional mapping for alarms
//
// $A3 = 970225 WES f664   Persistent alarms
//
//******************************************************************************

#include <iapl/common.h>

static char _partid_[] = "@(#) 1.11.1.4 src/adapters/timer/alarmbin.cpp, tka_timer, abe.l6";

#include <persfile.h>
#include <alarmbin.h>

// 0 indicates persistent Version hasn't been restored yet
int AlarmBoundedInterval::persistentVersion=0;                            //@A3
int AlarmBoundedInterval::version=1;                                      //@A3
const AString AlarmBoundedInterval::classid="bin";                              //@A3

// no other static variables need to be saved?                            //@A3
boolean AlarmBoundedInterval::versionAdded=                               //@A3
                     PersistentObject::addStaticVars("AlarmBoundedInterval", //@A3
                     &restoreStaticVars,                                  //@A3
                     &saveStaticVars);                                    //@A3

PersistentObject * AlarmBoundedInterval::alarmDefault=                               //@A3
                        PersistentObject::addObject(classid,                //@A3
                                             new AlarmBoundedInterval());   //@A3

// only static variable from persistence is currently persistentVersion   //@A3
boolean AlarmBoundedInterval::restoreStaticVars(PersistentFile& objectFile) //@A3
    {
      if (!objectFile.read(persistentVersion)) return false;              //@A3
      return objectFile.readEnd();                                       //@A3
    }

// only static variable from persistence is currently persistentVersion   //@A3
boolean AlarmBoundedInterval::saveStaticVars(PersistentFile& objectFile)  //@A3
    {
      if (!objectFile.write(version)) return false;                       //@A3
      return objectFile.writeEnd();                                      //@A3
    }

boolean AlarmBoundedInterval::save(PersistentFile &objFile) const         //@A3
  {                                                                       //@A3
    if (!AlarmInterval::save(objFile)) return false;                      //@A3
    if (!endTime.save(objFile)) return false;                             //@A2
    return objFile.write(remainder);                                      //@A2
  }                                                                       //@A3

boolean AlarmBoundedInterval::restore(PersistentFile &objFile)            //@A3
  {                                                                       //@A3
    if (persistentVersion > version) {                                    //@A3
      cout << "AlarmBoundedInterval versions not compatible" <<           //@A3
          endl;                                                           //@A3
      return false;                                                       //@A3
    }                                                                     //@A3
    if (!AlarmInterval::restore(objFile)) return false;                   //@A3
    if (!endTime.restore(objFile)) return false;                          //@A2
    return objFile.read(remainder);                                       //@A2
  }                                                                       //@A3


// Bounded Interval Alarm ctor with selector                                 @A1
AlarmBoundedInterval::AlarmBoundedInterval                                //@A1
            (const AString& alarmId,
             Interval thisInterval,const IntervalUnits& units,            //@A1
             const AString& thisStartDate,const AString& thisStartTime,   //@A1
             const AString& thisEndDate,const AString& thisEndTime,       //@A1
             const AString& selector):                                    //@A1
     AlarmInterval(thisInterval,units,thisStartDate,thisStartTime,        //@A1
         makeUid(alarmId,
                  thisInterval,units,thisStartDate,                       //@A2
                  thisStartTime,thisEndDate,thisEndTime,selector)),         //@A2
         endTime(makeEndTime(thisEndDate,thisEndTime)),                     //@A1
         remainder(getRemainder(thisInterval,units))
                                                                          //@A1
{                                                                         //@A1
}                                                                         //@A1

// set end time
// both 0: essentially no end time -
// end time 0 only - 00:00:00 of given date
// end date 0 only - todays date
IATime AlarmBoundedInterval::makeEndTime(const AString& thisEndDate, const AString& thisEndTime)
{
 // Parse end time
 IATime tempEndTime;
 AString dateDel(dateDelimeter);
 // 0 : use current date or time
 if (thisEndTime.indexOfAnyBut(":0")!=-1)
     tempEndTime.setTime(thisEndTime);
 if (thisEndDate.indexOfAnyBut(dateDel+"0")!=-1)
     tempEndTime.setDate(thisEndDate);
 return tempEndTime;
}

// Create a copy of any type of alarm
Alarm* AlarmBoundedInterval::createCopy() const
  {
   return new AlarmBoundedInterval(*this);
  }

// Alarm Methods

int AlarmBoundedInterval::incrementAlarm(IATime currentTime)    //@A2
 {
  --remainder;
  return AlarmRecurring::incrementAlarm(currentTime);                     //@A2
 }

int AlarmBoundedInterval::update(IATime thisTime)                       //@A5
  {
   IATime anAlarmTime=getAlarmTime();                                     //@A5
   int skipped=-1;
   long diff;
   int multiples,leftoverSeconds,increment;
   // years, months, days always use increment function
   if (theIntervalUnits=="years" || theIntervalUnits=="months" ||
       theIntervalUnits=="days") return incrementAlarm(thisTime);
   long intervalSeconds=anAlarmTime.convertInterval(aInterval,theIntervalUnits);
   if (anAlarmTime <= thisTime) {
      diff=thisTime.diffTime(anAlarmTime);
      multiples=diff/intervalSeconds;
      leftoverSeconds=diff%intervalSeconds;
      increment=diff-leftoverSeconds+intervalSeconds;
      anAlarmTime=anAlarmTime.addInterval(increment,"seconds");
      skipped=skipped+multiples+1;
      remainder=remainder-multiples-1;
      }
   // if (skipped==-1)                                                    //@D1
   //  cout << "incrementAlarm called when no increment needed" << endl;  //@D1
   // Set new alarm time
   setAlarmTime(anAlarmTime);
   return skipped;
  }

long AlarmBoundedInterval::getRemainder(
     Interval anInterval, const IntervalUnits& units)
    {
      if (units=="months")
         {
          return endTime.diffMonths(getAlarmTime())/anInterval;
         }
      else if (units=="years")
         {
           return endTime.diffYears(getAlarmTime())/anInterval;
         }
      // daylight savings time compensation
      else if (getAlarmTime().isdst()!=endTime.isdst() && (units=="days"
               || (units=="hours" && anInterval > 1)))
         {
           // end time is daylight, old wasn't - add an hour to adjust to
           // uniform interval
           if (endTime.isdst() > 0 && getAlarmTime().isdst()==0)
           {
             return ((long)endTime.diffTime(getAlarmTime())+60*60)/
                getAlarmTime().convertInterval(anInterval,units);
           }
           // end time isn't daylight, old was - subtract an hour to adjust
           // to uniform interval
           else if (endTime.isdst()==0 && getAlarmTime().isdst() > 0)
           {
            return ((long)endTime.diffTime(getAlarmTime())-60*60)/
                getAlarmTime().convertInterval(anInterval,units);
           }
          // no conversion
          else return ((long)endTime.diffTime(getAlarmTime()))/
                getAlarmTime().convertInterval(anInterval,units);
         }
      else
        return ((long)endTime.diffTime(getAlarmTime()))/
                getAlarmTime().convertInterval(anInterval,units);
    }

// override to check for alarm being set after end time
boolean AlarmBoundedInterval::isExpired() const
  { return (Alarm::isExpired() || endTime < getAlarmTime());}

// Is this a recurring or interval alarm and is the next
// interval past the current
// time < endTime
// override needed since end time defined
boolean AlarmBoundedInterval::isRecurring(IATime now) const               //@A3
  {
    AlarmBoundedInterval tempAlarm(*this);
    if (now > tempAlarm.endTime ) return false;
    else if (getAlarmTime() > tempAlarm.endTime) return false;            //@B1
    else if (getAlarmTime() > now ) return true;
    else if (now.addInterval(aInterval,theIntervalUnits) > tempAlarm.endTime)
         return false;
    return true;
    // tempAlarm.incrementAlarm(now);                                     //@A3
    // return tempAlarm.getAlarmTime() <= endTime;
  }

IAFactSet AlarmBoundedInterval::getFactSet() const
    {
      IAFactSet IntervalFacts;
      IntervalFacts.add(getAlarmTimeFact());
      IntervalFacts.add(getWeekdayFact());
      IntervalFacts.add(getAlarmIdFact());
      IntervalFacts.add(getIntervalFact());
      IntervalFacts.add(getAlarmCategoryFact());
      IntervalFacts.add(getIntervalsMissedFact());
      IntervalFacts.add(getRemainderFact());
      return IntervalFacts;
    }

IAAtom AlarmBoundedInterval::getRemainderFact() const
    {
      IAAtom RemainderFact;
      RemainderFact.setPredicate("Remainder");
      RemainderFact.addTerm((int)remainder);
      return RemainderFact;
     }

void AlarmBoundedInterval::print(ostream& os) const
   {
     AlarmInterval::print(os);
     cout << "end Time= " << endTime << endl;
     cout << "remainder=" << remainder << endl;
   }

