/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: alarmint.cpp
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/alarmint.cpp, tka_timer, abe.l6
// VERSION: 1.10.1.3 97/06/20 15:13:34
//
// PURPOSE:  Implementation of AlarmInterval class.
//
// DESCRIPTION:
//    An interval alarm - All non-identified
//    interval alarms start on predefined boundaries, starting at midnight.
//    May specify start time of alarm (used for calculating actual start
//    time as next interval on predefined boundary).
//
// CLASSES IMPLEMENTED:
//    AlarmInterval
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $D1   960823 WES  d280  Adjust start for hours since allowed as interval
// $A1   960909 WES  f304   Add support for selector in alarms
// $A2   960914 WES  f304  react to changes in IAFactSet and IAAtom
// $A3   961206 WES  f512   React to IAAtom and IAFactSet throws
// $A4 = 970225 WES f658   Additional mapping for alarms
// $A5 = 970312 WES f664   Persistent alarms
//
//******************************************************************************

#include <iapl/common.h>

static char _partid_[] = "@(#) 1.10.1.3 src/adapters/timer/alarmint.cpp, tka_timer, abe.l6";

#include <alarmint.h>
#include <iatk/iaerrors.h>

int AlarmInterval::persistentVersion=0;                                   //@A5
const int AlarmInterval::version=1;                                       //@A5
const AString AlarmInterval::classid="int";                               //@A5


// no other static variables need to be saved?                            //@A5
boolean AlarmInterval::versionAdded=                                      //@A5
             PersistentObject::addStaticVars("AlarmInterval",              //@A5
             &restoreStaticVars,                                          //@A5
             &saveStaticVars);                                            //@A5

PersistentObject * AlarmInterval::alarmDefault=                                      //@A5
                   PersistentObject::addObject(classid,                   //@A5
                                             new AlarmInterval());        //@A5

// only static variable from persistence is currently persistentVersion   //@A5
boolean AlarmInterval::restoreStaticVars(PersistentFile& objectFile)      //@A5
    {                                                                     //@A5
      if (!objectFile.read(persistentVersion)) return false;              //@A5
      return objectFile.readEnd();                                        //@A5
    }                                                                     //@A5

// only static variable from persistence is currently persistentVersion   //@A5
boolean AlarmInterval::saveStaticVars(PersistentFile& objectFile)         //@A5
    {                                                                     //@A5
      if (!objectFile.write(version)) return false;                       //@A5
      return objectFile.writeEnd();                                       //@A5
    }                                                                     //@A5

// Interval alarm - Default start times

AlarmInterval::AlarmInterval(const AString& alarmId,
               Interval thisInterval,const IntervalUnits& units, //@A1
                              const AString& selector):                   //@A1
       AlarmRecurring(thisInterval,units,                                 //@A1
           makeUid(alarmId,thisInterval,units,selector))                       //@A4
  {                                                                       //@A1
     // default identified alarms have uniform start times
     if (alarmId.numWords() == 0)
        adjustStart(thisInterval,units);                                  //@A1
  }                                                                       //@A1

// Interval alarm - Non-Default start times                                 @A1
// default for the startTime (used for comparing alarms) is time zero       @A1
AlarmInterval::AlarmInterval(const AString& alarmId,
               Interval thisInterval,const IntervalUnits& units,           //@A1
               const AString& thisStartDate, const AString& thisStartTime, //@A1
               const AString& selector):                                  //@A1
            AlarmRecurring(thisStartDate,thisStartTime,                   //@A1
            thisInterval,units,                                           //@A1
            makeUid(alarmId,
                thisInterval,units,thisStartDate,thisStartTime,selector)) //@A4
  {                                                                       //@A1
     // default identified alarms have uniform start times
     if (alarmId.numWords() == 0)
        adjustStart(thisInterval,units);                                  //@A1
  }                                                                       //@A1


// Interval Alarm ctor for subclasses                                       @A1
AlarmInterval::AlarmInterval(Interval thisInterval,const IntervalUnits& units, //@A1
               const AString& thisStartDate, const AString& thisStartTime, //@A1
               const AlarmUniqueId& alarmuid):                            //@A4
               AlarmRecurring(thisStartDate,thisStartTime,                //@A1
               thisInterval,units,alarmuid)                               //@A4
  {                                                                       //@A1
     // default identified alarms have uniform start times
     if (alarmuid.getType()=='n')
        adjustStart(thisInterval,units);                                  //@A1
  }                                                                       //@A1


// Create a copy of any type of alarm
Alarm* AlarmInterval::createCopy() const
  {
   return new AlarmInterval(*this);
  }

// Alarm Methods

IAFactSet AlarmInterval::getFactSet() const

    {
      IAError rc;                                                         //@A2
      IAFactSet IntervalFacts;
      IntervalFacts.add(getAlarmTimeFact());                              //@A3
      IntervalFacts.add(getWeekdayFact());                                //@A3
      IntervalFacts.add(getIntervalFact());                               //@A3
      IntervalFacts.add(getAlarmIdFact());                                //@A3
      IntervalFacts.add(getAlarmCategoryFact());                             //@A3
      IntervalFacts.add(getIntervalsMissedFact());                        //@A3
      return IntervalFacts;
    }

IAAtom AlarmInterval::getIntervalFact() const
    {
      IAAtom IntervalFact;                                                //@A2
      IntervalFact.setPredicate("Interval");                              //@A3
      IntervalFact.addTerm(getInterval());
      IntervalFact.addTerm(getIntervalUnits());
      return IntervalFact;
     }

// helper function adjust interval start time to proper boundary
void AlarmInterval::adjustStart(Interval thisInterval,IntervalUnits units)

     {
       // if its allowed for interval
          Interval addInt=0;
          IATime anAlarmTime=getAlarmTime();
       if (units=="days")
          {
            anAlarmTime.setTime("00:00:00");
            addInt=thisInterval-(anAlarmTime.yearday()%thisInterval);
          }
       else if (units=="minutes")
          {
            AString stringMin(anAlarmTime.minute()),Zero("0");
            if (stringMin.length()==1) stringMin=Zero+stringMin;
            AString stringHour(anAlarmTime.hour());
            if (stringHour.length()==1) stringHour=Zero+stringHour;
            anAlarmTime.setTime(stringHour+":"+stringMin+":00");
            addInt=thisInterval-(anAlarmTime.dayminute()%thisInterval);
          }
       else if (units=="hours")                                           //@D1
          {                                                               //@D1
            AString stringHour(anAlarmTime.hour()),Zero("0");             //@D1
            if (stringHour.length()==1) stringHour=Zero+stringHour;       //@D1
            anAlarmTime.setTime(stringHour+":00:00");                     //@D1
            addInt=thisInterval-(anAlarmTime.hour()%thisInterval);        //@D1
          }                                                               //@D1
       else if (units=="seconds")
          {
            addInt=thisInterval-(anAlarmTime.second()%thisInterval);
          }
       setAlarmTime(anAlarmTime.addInterval(addInt,units));
     }

