/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: alarmrec.cpp
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/alarmrec.cpp, tka_timer, abe.l6
// VERSION: 1.12.1.2 97/06/13 11:18:15
//
// PURPOSE:  Implementation of AlarmRecurring
//
// DESCRIPTION:
//    An abstract class for alarms that may occur more than once
//
// CLASSES IMPLEMENTED:
//    AlarmRecurring
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $D1   960826 WES  d280  cleanup
// $A1   960909 WES  f304   Add support for selector in alarms
// $A2   960914 WES  f304   react to changes in IAAtom
// $A3   961031 WES  F441   Special case months and years
// $A4   961206 WES  f512   React to IAAtom and IAFactSet throws
// $A5   970213 WES  f651   Fix loop when interval 0
// $A6 = 970226 WES  f658   Additional mapping for alarms
// $A7   970409 WES  f664   Persistent Alarms
//
//******************************************************************************


#include <iapl/common.h>

static char _partid_[] = "@(#) 1.12.1.2 src/adapters/timer/alarmrec.cpp, tka_timer, abe.l6";

#include <iatk/iaerrors.h>
#include <alarmrec.h>

int AlarmRecurring::persistentVersion=0;                                  //@A7
const int AlarmRecurring::version=1;                                      //@A7

// no other static variables need to be saved?                            //@A7
boolean AlarmRecurring::versionAdded=                                     //@A7
               PersistentObject::addStaticVars("Alarm",                   //@A7
               &restoreStaticVars,                                        //@A7
               &saveStaticVars);                                          //@A7



// only static variable from persistence is currently persistentVersion   //@A7
boolean AlarmRecurring::restoreStaticVars(PersistentFile& objectFile)     //@A7
    {                                                                     //@A7
      if (!objectFile.read(persistentVersion)) return false;              //@A7
      return objectFile.readEnd();                                        //@A7
    }                                                                     //@A7

// only static variable from persistence is currently persistentVersion   //@A7
boolean AlarmRecurring::saveStaticVars(PersistentFile& objectFile)        //@A7
    {                                                                     //@A7
      if (!objectFile.write(version)) return false;                       //@A7
      return objectFile.writeEnd();                                       //@A7
    }                                                                     //@A7

boolean AlarmRecurring::save(PersistentFile &objFile) const                        //@A7
          {                                                               //@A7
            Alarm::save(objFile);                                         //@A7
            if (!objFile.write(aInterval))                                //@A7
              return false;                                               //@A7
            if (!objFile.write(theIntervalUnits))                         //@A7
              return false;                                               //@A7
            // startDay may be blank
            return objFile.writeWholeString(startDay);
          }                                                               //@A7

boolean AlarmRecurring::restore(PersistentFile &objFile)                  //@A7
          {                                                               //@A7
            if (persistentVersion > version) {                            //@A7
               cout << "AlarmRecurring Versions not compatible" << endl;  //@A7
               return false;                                              //@A7
            }                                                             //@A7
            Alarm::restore(objFile);                                      //@A7
            if (!objFile.read(aInterval))                                //@A7
              return false;                                               //@A7
            if (!objFile.read(theIntervalUnits))                         //@A7
              return false;                                               //@A7
            // startDay may be blank
            return objFile.readWholeString(startDay);
          }                                                               //@A7

// Recurring alarm

AlarmRecurring::AlarmRecurring(Interval thisInterval,                      //@A1
                        const IntervalUnits& units,
                        const AlarmUniqueId& anId):                       //@A6
   Alarm(anId),aInterval(thisInterval),theIntervalUnits(units)   //@A6
 {                                                                        //@A1
    if (thisInterval<=0)                                                  //@A5
    {                                                                     //@A5
      AString badIntervalString("Interval= ");                            //@A5
      badIntervalString=badIntervalString+(AString)thisInterval;          //@A5
      ITHROW(IntervalValueError((char*)badIntervalString));               //@A5
    }                                                                     //@A5
    IATime::checkUnits(units);                                            //@A1
    setStartDay();                                                        //@A3
 }                                                                        //@A1

AlarmRecurring::AlarmRecurring(const AString& startDate,                  //@A1
                               const AString& startTime,                  //@A1
                               Interval thisInterval,                     //@A1
                               const IntervalUnits& units,                //@A1
                               const AlarmUniqueId& anId):                //@A6
   Alarm(startDate,startTime,anId),aInterval(thisInterval),               //@A6
         theIntervalUnits(units)                                          //@A1
 {                                                                        //@A1
    if (thisInterval<=0)                                                  //@A5
    {                                                                     //@A5
      AString badIntervalString("Interval= ");                            //@A5
      badIntervalString=badIntervalString+(AString)thisInterval;          //@A5
      ITHROW(IntervalValueError((char*)badIntervalString));               //@A5
    }                                                                     //@A5
    IATime::checkUnits(units);                                            //@A1
    setStartDay();                                                        //@A3
 }                                                                        //@A1

void AlarmRecurring::setStartDay()                                        //@A3
  {                                                                       //@A3
    // we haven't incremented to the next time yet for alarm time         //@A3
    // only set startDay for special cases                                //@A3
    if (theIntervalUnits=="months")                                       //@A3
      {                                                                   //@A3
        startDay=(AString)getAlarmTime().getDate().subString(8,2);        //@A3
        if (startDay.asInt()<= 28) startDay="";                           //@A3
      }                                                                   //@A3
    else if (theIntervalUnits=="years")                                   //@A3
      {                                                                   //@A3
       startDay=(AString)getAlarmTime().getDate().subString(5,5);         //@A3
       if (startDay!="02-29") startDay="";                                //@A3
      }                                                                   //@A3
    else startDay="";                                                     //@A3
  }                                                                       //@A3

boolean AlarmRecurring::isIntervalEqual(const AlarmRecurring* aAlarm) const
  {
    return aInterval==aAlarm->aInterval && theIntervalUnits==aAlarm->theIntervalUnits;
  }

// Alarm Methods

// Increment to next interval past current time and date
// Return number of increments skipped in string format
int AlarmRecurring::incrementAlarm(IATime thisTime)                       //@A5
  {
   IATime anAlarmTime=getAlarmTime();                                     //@A5
   int skipped=-1;
   if  (startDay!="" )  // we have a special case                         //@A3
    {                                                                     //@A3
      while ((anAlarmTime<=thisTime)) // take out skipped <= 10           //@A3
       {                                                                  //@A3
         anAlarmTime=anAlarmTime.addInterval(aInterval,theIntervalUnits); //@A3
      // now adjust day                                                   //@A3
      // 29th - 31st of month                                             //@A3
      if (theIntervalUnits=="months")                                     //@A3
       {                                                                  //@A3
        int monthDays=IATime::daysInMonth(                                //@A3
                 anAlarmTime.year(),anAlarmTime.month());                 //@A3
        int extraDays=(startDay.asInt() > monthDays) ?                    //@A3
                      monthDays-anAlarmTime.day() :                       //@A3
                      startDay.asInt()-anAlarmTime.day();                 //@A3
        anAlarmTime=anAlarmTime.addInterval(extraDays,"days");            //@A3
       }                                                                  //@A3
      // leap day                                                         //@A3
      else if (theIntervalUnits=="years")                                 //@A3
       {                                                                  //@A3
        // we already know this alarm is for Feb 29                       //@A3
        // The alarm will have been adjusted to Feb 28 for non-leap years //@A3
        if (IATime::isLeapYear(anAlarmTime.year()) &&                     //@A3
            anAlarmTime.day()==28)                                        //@A3
           anAlarmTime=anAlarmTime.addInterval(1,"days");                 //@A3
       }                                                                  //@A3
         ++skipped;                                                       //@A3
       }                                                                  //@A3
    }                                                                     //@A3
   else                                                                   //@A3
    {                                                                     //@A3
      while ((anAlarmTime<=thisTime)) // take out skipped <= 10
       {
         anAlarmTime=anAlarmTime.addInterval(aInterval,theIntervalUnits);
         ++skipped;
       }
     }
   // if (skipped==-1)                                                    //@D1
   //  cout << "incrementAlarm called when no increment needed" << endl;  //@D1
   // Set new alarm time
   setAlarmTime(anAlarmTime);
   return skipped;
  }

int AlarmRecurring::update(IATime thisTime)                       //@A5
  {
   IATime anAlarmTime=getAlarmTime();                                     //@A5
   int skipped=-1;
   long diff;
   int multiples,leftoverSeconds,increment;
   // years, months, days always use increment function
   if (theIntervalUnits=="years" || theIntervalUnits=="months" ||
       theIntervalUnits=="days") return incrementAlarm(thisTime);
   long intervalSeconds=anAlarmTime.convertInterval(aInterval,theIntervalUnits);
   if (anAlarmTime <= thisTime) {
      diff=thisTime.diffTime(anAlarmTime);
      multiples=diff/intervalSeconds;
      leftoverSeconds=diff%intervalSeconds;
      increment=diff-leftoverSeconds+intervalSeconds;
      anAlarmTime=anAlarmTime.addInterval(increment,"seconds");
      skipped=skipped+multiples+1;
      }
   // if (skipped==-1)                                                    //@D1
   //  cout << "incrementAlarm called when no increment needed" << endl;  //@D1
   // Set new alarm time
   setAlarmTime(anAlarmTime);
   return skipped;
  }


boolean AlarmRecurring::isRecurring(IATime currentTime) const             //@A5
// If there is no end time defined, this is always true
  { return true; }

// Get the interval missed fact
IAAtom AlarmRecurring::getIntervalsMissedFact() const
   {
      Alarm * tempAlarm=createCopy();
      IAAtom IntervalsMissedFact;                                         //@A2
      IntervalsMissedFact.setPredicate("IntervalsMissed");                //@A3
      int skipped=tempAlarm->incrementAlarm();
      if (skipped==-1) skipped=0;
      IntervalsMissedFact.addTerm(skipped);
      delete tempAlarm;
      return IntervalsMissedFact;
   }

void AlarmRecurring::print(ostream& os) const
   {
     Alarm::print(os);
     os << "Interval= "<< aInterval << "  " << (char *)theIntervalUnits <<endl;
   }

