/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: iatime.cpp
**
** Description: This class gets the current time and provides specific
** information requested.
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/

//******************************************************************************
//
// FILE:    src/adapters/timer/iatime.cpp, tka_timer, abe.l6
// VERSION: 1.13.1.2 97/06/13 11:19:05
//
// PURPOSE:  Definition of iatime
//
// DESCRIPTION:  This class contains time information and is used as an
//               Openclass collection class key
//
// CLASSES IMPLEMENTED:
//    IATime - class for time information
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// $A1   960918 WES f304   Cleanup
// $A2   961017 WES f441   Support for intervals of months and years
// $A3   970227 WES  f668   Add end bounded identified interval alarm
//                         and fix bug in daylight savings time processing
// $A4 = 970408 WES  f664   Persistent alarms
//******************************************************************************
#include <iapl/common.h>

static char _partid_[] = "@(#) 1.13.1.2 src/adapters/timer/iatime.cpp, tka_timer, abe.l6";
static unsigned leapDays[12]={31,29,31,30,31,30,31,31,30,31,30,31};       //@A2
static unsigned nonLeapDays[12]={31,28,31,30,31,30,31,31,30,31,30,31};    //@A2

#include <iostream.h>
#include <stdio.h>
#include <iatime.h>

// 0 indicates persistent Version hasn't been restored yet                //@A4
int IATime::persistentVersion=0;                                          //@A4
const int IATime::version=1;                                              //@A4
const AString IATime::classid="iat";                                      //@A4

boolean IATime::versionAdded=                                             //@A4
                     PersistentObject::addStaticVars("IATime",            //@A4
                     &restoreStaticVars,                                  //@A4
                     &saveStaticVars);                                    //@A4


// only static variable from persistence is currently persistentVersion   //@A4
boolean IATime::restoreStaticVars(PersistentFile& objectFile)             //@A4
    {                                                                     //@A4
      if (!objectFile.read(persistentVersion)) return false;              //@A4
      return objectFile.readEnd();                                       //@A4
    }                                                                     //@A4

// only static variable from persistence is currently persistentVersion   //@A4
boolean IATime::saveStaticVars(PersistentFile& objectFile)                //@A4
    {                                                                     //@A4
      if (!objectFile.write(version)) return false;                       //@A4
      return objectFile.writeEnd();                                      //@A4
    }                                                                     //@A4


IATime::IATime()

  { now();
  }

// Ctors

IATime::IATime(time_t newTime):aTime(newTime)
  {
   if (newTime < 0) ITHROW(TimeFormatError("Bad time_t = "+(AString)(long)newTime));
   thisTime=*localtime(&aTime);                                          //@A1
  }

IATime::IATime(const DateType& newdate, const TimeType& newtime)
  {
    checkDate(newdate);
    checkTime(newtime);
    thisTime.tm_year=newdate.subString(0,4).asInt()-1900;                 //@A1
    thisTime.tm_mon=newdate.subString(5,2).asInt()-1;                     //@A1
    thisTime.tm_mday=newdate.subString(8,2).asInt();                      //@A1
    thisTime.tm_hour=newtime.subString(0,2).asInt();                      //@A1
    thisTime.tm_min=newtime.subString(3,2).asInt();                       //@A1
    thisTime.tm_sec=newtime.subString(6,2).asInt();                       //@A1
    thisTime.tm_isdst=-1;                                                 //@A1
    aTime = mktime(&thisTime);                                            //@A1
    if (aTime==-1) cout << "Bad time set in constructor "                 //@A1
                      << newdate << " " << newtime << endl;
 }

// Accessors

// reset to current time
void IATime::now()
  {
    time(&aTime);                                                         //@A1
    thisTime=*localtime(&aTime);                                          //@A1
  }

// To Strings
DateType IATime::getDate() const
   {
     AString Syear(thisTime.tm_year+1900),Smonth(thisTime.tm_mon+1),      //@A1
             Smday(thisTime.tm_mday),Zero("0");                           //@A1
     if (Smonth.length()==1) Smonth=Zero+Smonth;
     if (Smday.length()==1) Smday=Zero+Smday;
     return Syear+dateDelimeter+Smonth+dateDelimeter+Smday;
   }

TimeType IATime::getTime() const
   {
     AString Shour(thisTime.tm_hour),Smin(thisTime.tm_min),               //@A1
             Ssec(thisTime.tm_sec),Zero("0");                             //@A1
     if (Shour.length()==1) Shour=Zero+Shour;
     if (Smin.length()==1) Smin=Zero+Smin;
     if (Ssec.length()==1) Ssec=Zero+Ssec;
     return Shour+":"+Smin+":"+Ssec;
   }

TimeStamp IATime::getDateTime() const
   {
     AString Syear(thisTime.tm_year+1900),Smonth(thisTime.tm_mon+1),      //@A1
             Smday(thisTime.tm_mday),Zero("0");                           //@A1
     AString Shour(thisTime.tm_hour),Smin(thisTime.tm_min),               //@A1
             Ssec(thisTime.tm_sec);                                       //@A1
     if (Smonth.length()==1) Smonth=Zero+Smonth;
     if (Smday.length()==1) Smday=Zero+Smday;
     if (Shour.length()==1) Shour=Zero+Shour;
     if (Smin.length()==1) Smin=Zero+Smin;
     if (Ssec.length()==1) Ssec=Zero+Ssec;
     return Syear+dateDelimeter+Smonth+dateDelimeter+Smday+" "+Shour+":"+Smin+":"+Ssec;
   }


// Setters

void IATime::settime_t(time_t newTime)
  {
   if (newTime < 0) ITHROW(TimeFormatError("Bad time_t = "+(AString)(long)newTime)); //@A4
   aTime=newTime;                                                        //@A4
   thisTime=*localtime(&aTime);                                          //@A4
  }

void IATime::setDate(const DateType& newdate)
  {
    checkDate(newdate);                                                   //@A1
    thisTime.tm_year=newdate.subString(0,4).asInt()-1900;                 //@A1
    thisTime.tm_mon=newdate.subString(5,2).asInt()-1;                     //@A1
    thisTime.tm_mday=newdate.subString(8,2).asInt();                      //@A1
    thisTime.tm_isdst=-1;                                                 //@A1
    aTime=mktime(&thisTime);                                              //@A1
    if (aTime==-1) cout << "Bad date set= " << newdate << endl;           //@A1
  }

void IATime::setTime(const TimeType& newtime)

  {
    checkTime(newtime);
    thisTime.tm_hour=newtime.subString(0,2).asInt();                      //@A1
    thisTime.tm_min=newtime.subString(3,2).asInt();                       //@A1
    thisTime.tm_sec=newtime.subString(6,2).asInt();                       //@A1
    thisTime.tm_isdst=-1;                                                 //@A1
    aTime = mktime(&thisTime);                                            //@A1
    if (aTime==-1) cout << "Bad time set=" << newtime << endl;            //@A1
  }

void IATime::setDateTime(const DateType& newdate, const TimeType& newtime)
  {
    checkDate(newdate);
    checkTime(newtime);
    thisTime.tm_year=newdate.subString(0,4).asInt()-1900;                 //@A1
    thisTime.tm_mon=newdate.subString(5,2).asInt()-1;                     //@A1
    thisTime.tm_mday=newdate.subString(8,2).asInt();                      //@A1
    thisTime.tm_hour=newtime.subString(0,2).asInt();                      //@A1
    thisTime.tm_min=newtime.subString(3,2).asInt();                       //@A1
    thisTime.tm_sec=newtime.subString(6,2).asInt();                       //@A1
    thisTime.tm_isdst=-1;                                                 //@A1
    aTime = mktime(&thisTime);                                            //@A1
    if (aTime==-1) cout << "Bad time set" <<                              //@A1
       newdate << "  " << newtime << endl;                                //@A1
  }


// other member functions

// get time 0 for defaults
IATime IATime::getZeroTime()
   {
    IATime Zero;
    Zero.thisTime.tm_year=70;                                             //@A1
    Zero.thisTime.tm_mon=0;                                               //@A1
    Zero.thisTime.tm_mday=01;                                             //@A1
    Zero.thisTime.tm_hour=0;                                              //@A1
    Zero.thisTime.tm_min=0;                                               //@A1
    Zero.thisTime.tm_sec=0;                                               //@A1
    Zero.thisTime.tm_isdst=-1;                                            //@A1
    Zero.aTime = mktime(&Zero.thisTime);                                  //@A1
    if (Zero.aTime==-1) cout << "Bad time set in Zero time" << endl;      //@A1
    return Zero;
   }

boolean IATime::checkUnits(const IntervalUnits& units)
        {
         if (units!="seconds" && units!="minutes" && units!="days"
             && units!="hours" && units!="months" && units!="years")      //@A2
            {
              AString badUnits;
              badUnits=badUnits+"Units= "+units;
              ITHROW(IntervalUnitsError(badUnits));
            }
         return true;
        }

      // converts to a value of seconds
long IATime::convertInterval(Interval aInterval,
                       const IntervalUnits& units) const                  //@A2
   {
    if (units=="minutes") return aInterval*60;
    else if (units=="seconds") return aInterval;
    else if (units=="hours") return aInterval*60*60;
    else if (units=="days") return aInterval*60*60*24;
    else if (units=="months")                                             //@A2
    // do last day of month if that day doesn't exist                     //@A2
     {                                                                    //@A2
       int days=0,thisMonth=month(),thisYear=year()+1900;                 //@A2
       int monthDaysNext,monthDaysThis,nextMonth,nextYear,incr,prevMonth; //@A2
       if (aInterval >= 0) incr=1;                                        //@A2
       else                                                               //@A2
        {                                                                 //@A2
         incr=-1;                                                         //@A2
         if (thisMonth > 0) thisMonth--;                                  //@A2
         else                                                             //@A2
          {                                                               //@A2
            thisMonth=11;                                                 //@A2
            thisYear--;                                                   //@A2
          }                                                               //@A2
        }                                                                 //@A2
       // monthDaysNext is actually previous if aInterval is negative       @A2
       for (int i=0;i!=aInterval;i=i+incr)                                //@A2
        {                                                                 //@A2
         // get new thisMonth and thisYear here                             @A2
         // get nextMonth and nextYear                                      @A2
         if (incr==1 && thisMonth==11)                                    //@A2
          {                                                               //@A2
           nextMonth=0;                                                   //@A2
           nextYear=thisYear+1;                                           //@A2
          }                                                               //@A2
         else if (incr==-1 && thisMonth==0)                               //@A2
          {                                                               //@A2
           nextMonth=11;                                                  //@A2
           nextYear=thisYear-1;                                           //@A2
          }                                                               //@A2
         else                                                             //@A2
          {                                                               //@A2
           nextMonth=thisMonth+incr;                                      //@A2
           nextYear=thisYear;                                             //@A2
          }                                                               //@A2
         days=days+daysInMonth(thisYear,thisMonth);                       //@A2
         prevMonth=thisMonth;                                             //@A2
         thisMonth=nextMonth;                                             //@A2
         thisYear=nextYear;                                               //@A2
        }                                                                 //@A2
        // special handling if day was > 28                               //@A2
        if (day() > 28)                                                   //@A2
           {                                                              //@A2
            if (incr > 0)                                                 //@A2
               days=days+(daysInMonth(thisYear,thisMonth)-day())*incr;    //@A2
            else                                                          //@A2
               days=days+(daysInMonth(thisYear,prevMonth)-day())*incr;    //@A2
           }                                                              //@A2
        return 60*60*24*days*incr;                                        //@A2
       }                                                                  //@A2
    else if (units=="years")                                              //@A2
      {                                                                   //@A2
       // Figure out how many days                                          @A2
       int i,days=0;                                                      //@A2
       int incr,currYear=year()+1900,thisDay=day(),prevYear;              //@A2
       boolean leapDay;                                                   //@A2
       if (month()==1 && thisDay==29)                                     //@A2
        {                                                                 //@A2
          leapDay=true;                                                   //@A2
        }                                                                 //@A2
       else leapDay=false;                                                //@A2
       if (aInterval >= 0) incr=1;                                        //@A2
       else incr=-1;                                                      //@A2
       for (i=0;i!=aInterval;i=i+incr)                                    //@A2
        {                                                                 //@A2
         days=((month()<=1 && incr > 0 && !leapDay) ||                    //@A2
               ((month() > 1 || leapDay) && incr < 0) ) ?                 //@A2         //@A2
              days+daysInYear(currYear) :                                 //@A2
              days+daysInYear(currYear+incr);                             //@A2
         currYear=currYear+incr;                                          //@A2
        }                                                                 //@A2
        return 60*60*24*days*incr;                                        //@A2
      }                                                                   //@A2
    else                                                                  //@A2
       {                                                                  //@A2
        AString badUnits;                                                 //@A2
        badUnits=badUnits+"Units= "+units;                                //@A2
        ITHROW(IntervalUnitsError(badUnits));                             //@A2
       }                                                                  //@A2
    return 0;
   }

IATime IATime::addInterval(Interval aInterval, const IntervalUnits& units) const
{
  IATime tempTime(*this);                                                 //@A1
  tempTime.aTime=tempTime.aTime+                                          //@A1
           convertInterval(aInterval,units);                              //@A1
  tempTime.thisTime=*localtime(&tempTime.aTime);                          //@A1
  // for intervals of minutes, seconds or 1 hour no daylight savings      //@A3
  // time special handling                                                //@A3
  if (units=="seconds" || units=="minutes" || (units=="hours" &&          //@A3
       aInterval==1)) return tempTime;                                    //@A3
  else if (tempTime.thisTime.tm_isdst != thisTime.tm_isdst)  //fast test  @A2@A3
     {                                                                    //@A2
       // new time is daylight, old wasn't - subtract an hour to adjust   //@A2
       if (tempTime.thisTime.tm_isdst > 0 && thisTime.tm_isdst==0)        //@A2
          {                                                               //@A2
            tempTime.aTime=tempTime.aTime-60*60;                          //@A2
            tempTime.thisTime=*localtime(&tempTime.aTime);                //@A2
          }                                                               //@A2
       // new time isn't daylight, old was - add an hour to adjust        //@A2
       else if (tempTime.thisTime.tm_isdst==0 && thisTime.tm_isdst > 0)   //@A2
          {                                                               //@A2
            tempTime.aTime=tempTime.aTime+60*60;                          //@A2
            tempTime.thisTime=*localtime(&tempTime.aTime);                //@A2
          }                                                               //@A2
       // no conversion                                                   //@A2
       else return tempTime;                                              //@A2
     }                                                                    //@A2
  return tempTime;                                                        //@A1
}                                                                         //@A1

int IATime::diffDate(const IATime& newtime) const
 {
   int year1=year()+1900;                                                 //@A2
   int year2=newtime.year()+1900;                                         //@A2
   int day1=yearday();
   int day2=newtime.yearday();
   if (year1==year2) return day1-day2;
   int totdays;
   totdays=day1 + (daysInYear(year2)-day2);
   for (year2=year2+1;year2<year1;year2++)
      totdays=totdays+daysInYear(year2);
   return totdays;
 }

boolean IATime::checkDate(const DateType& newDate) const
   {
    AString Digits="0123456789";
    AString badDate,thisDate(newDate);
    thisDate.strip();
    badDate=badDate+"Date= "+thisDate;
    AString yearString=thisDate.subString(0,4);
    AString monthString=thisDate.subString(5,2);
    AString dayString=thisDate.subString(8,2);
    if (thisDate.length() != 10 || thisDate.subString(4,1) != dateDelimeter ||
        thisDate.subString(7,1)!=dateDelimeter
        || !(yearString.indexOfAnyBut(Digits)==-1)
        || !(monthString.indexOfAnyBut(Digits)==-1)
        || !(dayString.indexOfAnyBut(Digits)==-1)
        || (yearString.asInt() < 1970)
        || (yearString.asInt() > 3000)
        || (monthString.asInt() < 01)
        || (monthString.asInt() > 12)
        || (dayString.asInt() < 01)
        || (isLeapYear(yearString.asInt()) ?
            dayString.asInt() > leapDays[monthString.asInt()-1] :
            dayString.asInt() > nonLeapDays[monthString.asInt()-1] ))
                              ITHROW(DateFormatError((char*)badDate));
    return true;
   }

boolean IATime::checkTime(const TimeType& newTime) const
   {
    AString badTime,thisTime(newTime);
    AString Digits="0123456789";
    thisTime.strip();
    badTime=badTime+"Time= "+thisTime;
    AString hourString=thisTime.subString(0,2);
    AString minuteString=thisTime.subString(3,2);
    AString secondString=thisTime.subString(6,2);
    if (thisTime.length() != 8 || thisTime.subString(2,1) != ':' ||
        thisTime.subString(5,1)!=':'
        || !(hourString.indexOfAnyBut(Digits)==-1)
        || !(minuteString.indexOfAnyBut(Digits)==-1)
        || !(secondString.indexOfAnyBut(Digits)==-1)
        || (hourString.asInt() > 23)
        || (minuteString.asInt() > 59)
        || (secondString.asInt() > 59))
                              ITHROW(TimeFormatError((char*)badTime));
    return true;
   }

int IATime::daysInMonth(int year, int month)                              //@A2
{                                                                         //@A2
 return isLeapYear(year) ? leapDays[month] :                              //@A2
                           nonLeapDays[month];                            //@A2
}                                                                         //@A2

boolean IATime::isLeapYear(int year)
{
 return (0==(year%4) && (0!=(year%100) || 0==(year%400)));
}

int IATime::daysInYear(int year)
{
 return (isLeapYear(year) ? 366 : 365);
}
