/*************************************************************************
** (C) Copyright IBM Corp, 1996
**
** File: PersistentFile.cpp
**
** DISCLAIMER OF WARRANTIES:
** -------------------------
** The following [enclosed] code is sample code created by IBM
** Corporation.  This sample code is not part of any standard IBM product
** and is provided to you solely for the purpose of assisting you in the
** development of your applications.  The code is provided "AS IS",
** without warranty of any kind.  IBM shall not be liable for any damages
** arising out of your use of the sample code, even if they have been
** advised of the possibility of such damages.
*************************************************************************/
//******************************************************************************
//
// FILE:    src/adapters/timer/persfile.cpp, tka_timer, abe.l6
// VERSION: 1.1 97/06/13 11:30:53
//
// PURPOSE:  PersistentFile class implementation
//
// DESCRIPTION:
//    Class defining persistence
//
// CLASSES DECLARED:
//    PersistentFile
//
//******************************************************************************

//** Version History ***********************************************************
//
// FLG   DATE   WHO REASON DESCRIPTION
// ===   ====== === ====== ===========
// new   970331 WES  f664   Persistent alarms
//******************************************************************************

#include <iapl/common.h>

static char _partid_[] = "@(#) 1.1 src/adapters/timer/persfile.cpp, tka_timer, abe.l6";

#include <iatk/iaerrors.h>
#include <string.h>
#include <persobj.h>
#include <persfile.h>
#include <iostream.h>


// !!!!!!!!   Need to verify mode in routines?
// !!!!!!!!


// buffer length must be length of buffer
const int PersistentFile::bufferLength=1024;

boolean PersistentFile::atEOF()
{
  if (mode==NO_PERSISTENT_FILE_MODE) return true;
  else return feof(objectFile);
}

boolean PersistentFile::init(int newMode)
  {
    if (filename=="")
      {
        cout << "Filename must be set before init" << endl;
        return false;
      }
    else if (mode != NO_PERSISTENT_FILE_MODE)
      {
        cout << "Persistent File object must be closed before init called" << endl;
        return false;
      }
    else if (newMode!=READ_PERSISTENT_FILE_MODE &&
             newMode!=WRITE_PERSISTENT_FILE_MODE &&
             newMode!=APPEND_PERSISTENT_FILE_MODE)
       {
         cout << "Invalid mode: " << mode << " for PersistentFile.init()." << endl;
         return false;
       }
    else mode=newMode;
    AString fileMode;
    if (mode==READ_PERSISTENT_FILE_MODE) fileMode="r";
    else if (mode==WRITE_PERSISTENT_FILE_MODE) fileMode="w";
    else if (mode==APPEND_PERSISTENT_FILE_MODE) fileMode="a";
    objectFile=fopen(filename,fileMode);
    if (objectFile==NULL)
       {
        return false;
       }
    return true;
  }

boolean PersistentFile::processStaticVars()
  {
    if (mode==READ_PERSISTENT_FILE_MODE)
       return PersistentObject::restoreAllStatics(*this);
    else if (mode==WRITE_PERSISTENT_FILE_MODE)
       return PersistentObject::saveAllStatics(*this);
    // no other option currently
    else return false;
  }

boolean PersistentFile::close()
 {
  if (mode!=READ_PERSISTENT_FILE_MODE &&
      mode!=WRITE_PERSISTENT_FILE_MODE &&
      mode!=APPEND_PERSISTENT_FILE_MODE)
    {
      cout << "Invalid mode: " << mode << " for PersistentFile.close()." << endl;
      return false;
    }
  mode=NO_PERSISTENT_FILE_MODE;
  if (0 != fclose(objectFile)) {
     cout << "Unable to close file \"" << filename << "\"" << endl;
     return false;
    }
  return true;
}

boolean PersistentFile::saveObject(PersistentObject * object)
{
  if (!write(object->classId())) return false;
  if (!object->save(*this)) return false;
  return writeEnd();
}

PersistentObject * PersistentFile::restoreNewObject(const AString& classid )
{
  PersistentObject * retVal=PersistentObject::getDefaultObject(classid);
  if (retVal==NULL) {
     readEnd();
     return retVal;
  }
  retVal=retVal->createCopy();
  if (!retVal->restore(*this)) {
     delete retVal;
     retVal=NULL;
  }
  readEnd();
  return retVal;
}

PersistentObject * PersistentFile::restoreNewObject() {
  AString classid;
  if (!read(classid)) return NULL;
  else return restoreNewObject(classid);
}

boolean PersistentFile::write(const AString& string)
{
  // take these checks out later??
  if (string.length() > 1023 || string.numWords() > 1 ) {
     cout << "Inappropriate use of write(). Use writeWholeString" << endl;
     return false;
     }
  if (fprintf(objectFile,"%s ",(const char *)string) < 0) return false;
  return true;
}


boolean PersistentFile::write(const int& number)
{
  if (fprintf(objectFile,"%d ",number) < 0) return false;
  return true;
}

boolean PersistentFile::write(const long& number)
{
  if (fprintf(objectFile,"%ld ",number) < 0) return false;
  return true;
}

boolean PersistentFile::write(const char& aChar)
{
  if (fprintf(objectFile,"%c ",aChar) < 0) return false;
  return true;
}

boolean PersistentFile::writeEnd()
{
  if (fprintf(objectFile,"\n") < 0) return false;
  return true;
}

boolean PersistentFile::read(AString& string)
{
  // take these checks out later??
  if (fscanf(objectFile,"%1023s",buffer) < 0) return false;
  // for performance only one copy of buffer
  string="";
  string+=buffer;
  return true;
}

boolean PersistentFile::read(int& number)
{
  if (fscanf(objectFile,"%d",&number) < 0) return false;
  return true;
}

boolean PersistentFile::read(long& number)
{
  if (fscanf(objectFile,"%ld",&number) < 0) return false;
  return true;
}

boolean PersistentFile::read(char& aChar)
{
  if (fscanf(objectFile,"%c",&aChar) < 0) return false;
  return true;
}

boolean PersistentFile::readEnd()
{
  if (fscanf(objectFile,"%*[^\n]\n") < 0) return false;
  return true;
}

boolean PersistentFile::writeWholeString(const AString& string)
{
  if (fprintf(objectFile,"%d ",string.length()) < 0) return false;
  // don't write null terminator
  if (fwrite((char *)string,sizeof(char),string.length(),objectFile) !=
      string.length()) {
    cout << "Error writing whole String" << endl;
    return false;
  }
  // to make file look nice
  fputc(' ',objectFile);
  return true;
}

boolean PersistentFile::readWholeString(AString& string)
{
  int stringLength;
  // don't read any white space
  if (fscanf(objectFile,"%d",&stringLength) < 0) return false;
  // now read next blank
      fgetc(objectFile);
  // now read string
  if (stringLength < 1023) {
      if (fread(buffer,sizeof(char),stringLength,objectFile) !=
         stringLength) {
       cout << "Error reading whole string into reusable buffer" << endl;
       return false;
      }
      buffer[stringLength]='\0';
      // for performance only one copy of buffer
      string="";
      string+=buffer;
  }
  else {
     char *tempbuffer=new char[stringLength+1];
      if (fread(tempbuffer,sizeof(char),stringLength,objectFile) !=
         string.length()) {
       cout << "Error reading whole string into temporary buffer" << endl;
       delete tempbuffer;
       return false;
      }
      tempbuffer[stringLength]='\0';
      // for performance only one copy of buffer
      string="";
      string+=tempbuffer;
      delete tempbuffer;
  }
  // get blank used to make file look nice
  fgetc(objectFile);
  return true;
}
