/* BackProp class

 Constructing Intelligent Agents with Java
   (C) Joseph P. Bigus and Jennifer Bigus 1997
    
 */

import java.awt.*;
import java.util.*;
import java.lang.Math;
import java.io.* ;

// standard backward propagation with momentum
public class BackProp extends Object {
    String name ;

  // data parameters
    static DataSet ds ;
    Vector data ;       // train/test data from file
    int recInx=0 ;      // current record index
    int numRecs=0 ;     // number of records in data
    int fieldsPerRec=0;

  // error measures
    float sumSquaredError ;  // total SSE for an epoch
    float aveRMSError ;      // average root-mean-square error
    int numPasses ;          // number of passes over the data set

  // network architecture parameters
    int numInputs ;
    int numHid1 ;
    int numOutputs ;
    int numUnits ;
    int numWeights ;

  // network control parameters
    int mode;
    float learnRate ;
    float momentum;
    float tolerance;

  // network data
    float activations[] ;
    float weights[] ;
    float wDerivs[] ;
    float thresholds[];
    float tDerivs[] ;
    float tDeltas[] ;
    float teach[];   // target output values
    float error[];
    float deltas[];  // the error deltas
    float wDeltas[] ;

    TextArea textArea1 ;

  BackProp(String Name) {
     name = Name ;
     data = new Vector() ;
  }


    public void show_array(String name, float[] arr) {
        textArea1.append("\n" + name  + "= ") ;
     for (int i=0 ; i < arr.length ; i++) {
        textArea1.append(arr[i] + "  ") ;
     }

    }

    public void display_network() {
//      show_array("weights",weights);
//      show_array("thresholds", thresholds);
	    show_array("activations",activations) ;
//	    show_array("teach",teach) ;

        String desired = ds.getClassFieldValue(recInx-1);
        String actual = ds.getClassFieldValue(activations, numInputs+numHid1);
        textArea1.append("\n Desired: " + desired + " Actual: " + actual ) ;
    }

    public float logistic(double sum) {
       return (float)(1.0 / (1 + Math.exp(-1.0 * sum)));
    }

    // move data from train/test set into network input units
    public void readInputs() {
     recInx = recInx % numRecs ; // keep index from 0 to n-1 records
     int inx=0 ;
     float[] tempRec = (float[])data.elementAt(recInx) ; // get record
     for (inx=0 ; inx < numInputs ; inx++ ) {
       activations[inx] = tempRec[inx] ;
     }
     for (int i=0 ; i < numOutputs; i++ ) {
       teach[i] = tempRec[inx++] ;
     }
     recInx++ ;
    }

    // do a single forward pass through the network
    public void computeOutputs() {
     int i, j ;
     int firstHid1 = numInputs ;
     int firstOut = numInputs + numHid1 ;

     // first layer
     int inx = 0 ;
     for(i = firstHid1 ; i < firstOut ; i++) {
       float sum = thresholds[i];
       for (j = 0 ; j < numInputs ; j++) { // compute net inputs
         sum += activations[j] * weights[inx++] ;
       }
       activations[i] = logistic(sum) ;  // compute activation
     }
     // second layer
     for(i = firstOut ; i < numUnits ; i++) {
        float sum = thresholds[i] ;
       for (j = firstHid1 ; j < firstOut ; j++) { // compute net inputs
         sum += activations[j] * weights[inx++] ;
       }
       activations[i] = logistic(sum) ;  // compute activation
     }
    }

    // compute the errors using backward error propagation
    // weight changes get placed in (or added to) wDerivs
    // threshold changes get placed in (or added to) tDerivs
    public void computeError() {
     int i, j ;
     int firstHid1 = numInputs ;
     int firstOut = numInputs + numHid1 ;

     // clear hidden unit errors
     for(i = numInputs ; i < numUnits ; i++) {
        error[i] = (float)0.0 ;
     }
     // compute output layer errors and deltas
     for(i = firstOut ; i < numUnits ; i++) {
        error[i] = teach[i-firstOut] - activations[i] ; // compute output errors
        sumSquaredError += error[i] * error[i] ;       // accumulate squared errors
        if (Math.abs(error[i]) < tolerance) error[i] = (float)0.0 ; // close enough
        deltas[i] = error[i] * activations[i] * (1 - activations[i]);
     }

     // compute hidden layer errors
     int winx = numInputs * numHid1 ;    // offset into weight array
     for(i = firstOut ; i < numUnits ; i++) {
        for (j = firstHid1; j < firstOut ; j++) {
            wDerivs[winx] += deltas[i] * activations[j] ;
            error[j] += weights[winx] * deltas[i] ;
            winx++ ;
        }
        tDerivs[i] += deltas[i] ;
     }

     // compute hidden layer deltas
     for (i = firstHid1 ; i < firstOut ; i++) {
       deltas[i] = error[i] * activations[i] * (1 - activations[i]);
     }

     // compute input layer errors
     winx = 0 ;               // offset into weight array
     for(i = firstHid1 ; i < firstOut ; i++) {
        for (j = 0; j < firstHid1 ; j++) {
            wDerivs[winx] += deltas[i] * activations[j] ;
            error[j] += weights[winx] * deltas[i] ;
            winx++ ;
        }
        tDerivs[i] += deltas[i] ;
     }

    }


    // apply the changes to the weights
    public void adjustWeights() {
     int i ;

     // first walk through the weights array
     for(i = 0; i < weights.length ; i++) {
        wDeltas[i] = (learnRate * wDerivs[i]) + (momentum * wDeltas[i]);
        weights[i] += wDeltas[i] ;   // modify the weight
        wDerivs[i] = (float)0.0 ;
     }

     // then walk through the threshold array
     for(i=numInputs ; i < numUnits ; i++) {
        tDeltas[i] = learnRate * tDerivs[i]  + (momentum * tDeltas[i]);
        thresholds[i] += tDeltas[i] ; // modify the threshold
        tDerivs[i] = (float)0.0 ;
     }
     // if at the end of an epoch, compute average RMS Error
     if (recInx == numRecs) {
        numPasses++ ;  // increment pass counter
        aveRMSError = (float)Math.sqrt( sumSquaredError / (numRecs * numOutputs));
        sumSquaredError = (float)0.0 ;  // clear the accumulator
     }
    }


    // for networks with single continuous outputs only
    public double getPrediction(float[] inputRec) {
     int firstOut = numInputs + numHid1 ;
     // set input unit activations
     for (int inx=0 ; inx < numInputs ; inx++ ) {
       activations[inx] = inputRec[inx] ;
     }

     computeOutputs() ;  // do forward pass through network
     return (double)activations[firstOut] ;
    }

    public void process() {

     readInputs() ;   // set input unit activations
     computeOutputs() ;  // do forward pass through network
     computeError() ;  // compute error and deltas

     // only adjust if in training mode
     if (mode == 0) adjustWeights() ;   // apply changes to weights

    }

    public void reset() {
        int i ;
        for (i=0 ; i < weights.length ; i++) {
          weights[i] = (float)1.0 - (float)Math.random();    // between -1 and +1
          wDeltas[i] = (float)0.0 ;
          wDerivs[i] = (float)0.0 ;
        }
        for (i=0 ; i < numUnits ; i++) {
          thresholds[i] = (float)1.0 - (float)Math.random();    // between -1 and +1
          tDeltas[i] = (float)0.0 ;
          tDerivs[i] = (float)0.0 ;
        }
    }

    // create a Back Prop network with specified architecture
    public void createNetwork(int NumIn, int NumHid1, int NumOut){

        // set the network architecture
        numInputs = NumIn ;
        numHid1 = NumHid1;
        numOutputs = NumOut;
        numUnits = numInputs + numHid1 + numOutputs ;
        numWeights = (numInputs*numHid1) + (numHid1*numOutputs);

        // initialize control parameters
        learnRate = (float)0.2;
        momentum = (float)0.7 ;
        tolerance = (float)0.1 ;
        mode = 0 ;             // 0 = train mode, 1 = run mode
        aveRMSError = (float)0.0 ;
        numPasses = 0;

        // create weight and error arrays
        activations = new float[numUnits]; // unit activations
        weights = new float[numWeights];
        wDerivs = new float[numWeights];   // accumulated wDeltas
        wDeltas = new float[numWeights];   // weight changes
        thresholds = new float[numUnits];
        tDerivs = new float[numUnits];     // accumulated tDeltas
        tDeltas = new float[numUnits] ;    // threshold changes
        teach = new float[numOutputs] ;    // desired outputs
        deltas = new float[numUnits];
        error = new float[numUnits] ;

        reset() ;  // reset and initialize the weight arrays

       return;
    }





}
