/* FilterAgent class

   Constructing Intelligent Agents with Java
   (C) Joseph P. Bigus and Jennifer Bigus 1997

   This agent can score news articles using a keyword list
   maintain a user profile, and use it to filter articles

   Changes 7/21/97 - fixed trace statements in run ("/n" + )

*/
// package sunw.demo.ciagent ;
 import java.awt.* ;
 import java.io.* ;
 import java.util.*;

 public class FilterAgent extends CIAgent {

  NewsFilter newsFilter ;
  protected KMapNet clusterNet ;
  protected BackProp scoreNet ;
  protected boolean buildClusterNet = false;
  protected boolean buildScoreNet = false ;

  public FilterAgent() { name = "FilterAgent"; }
  public FilterAgent(String Name) { super(Name); } ;



 public void process() {
    stopped = false ;
    // start a thread running and send an update message every interval secs
    trace("Starting " + name + "\n") ;
    runnit = new Thread(this) ;
    runnit.start() ;
 }

 public void stop() {
    stopped = true;
    trace(name + " stopped \n") ;
    runnit.stop() ;
}

 // method of Runnable (Thread) interface
 public void run() {

      while(stopped == false){     // interval timer
        try {
           runnit.sleep((long) 5 * 1000) ;  // in milliseconds
           if (buildClusterNet) {
             trace(name + ": Starting to build Cluster network\n");
             trainClusterNet() ;  // train the network
             buildClusterNet = false ;
             newsFilter.clusterNetBuilt() ; // signal application
             trace("\n" + name + ": Completed build of Cluster network\n");
           } else if (buildScoreNet) {
             trace(name + ": Starting to build Score network \n");
             trainScoreNet() ;  // train the network
             buildScoreNet = false ;
             newsFilter.scoreNetBuilt();
             trace("\n" + name + ": Completed build of Score network\n");
           }
        }
        catch (InterruptedException e)
        {
          // interrupted
        }

      }

 }

  public  void ciaEventFired(CIAgentEvent e) {
      trace(name + ":  CIAgentEvent received by " + name +
           " from " + e.getSource() + " with arg " + e.getArgObject()) ;

      // set flag for autonomous thread to see
      String arg = (String)e.getArgObject() ;
      if (arg.equals("buildClusterNet")) {
          buildClusterNet = true ;
      } else if (arg.equals("buildScoreNet"))
          buildScoreNet = true ;

  }

  // used to trigger autonomous build of Kohonen Map
  public void buildClusterNet() {
     buildClusterNet = true ;
  }

  // used to trigger autonomous build of BackProp network
  public void buildScoreNet() {
     buildScoreNet = true ;
  }

  // score a single article
  void score(NewsArticle article, String[] keywords, int filterType) {

      article.counts = countWordMultiKeys(keywords, article.body);
      int size = article.counts.length ;
      int sum = 0 ;
      for (int i = 0 ; i < size ; i++) sum += article.counts[i];
      article.sum = sum ;  // for convenience

      // based on the type of filter the user wants
      // fill in the score slot of each article
      switch(filterType) {
       case 0:   // keyword
         article.score = (double)sum ;
         break;
       case 1:   // cluster
         if (clusterNet != null) {
            //  pass through network --- get clusterID
           float[] inputRec = new float[size+2] ;
           for (int i=0 ; i < size ; i++) inputRec[i] = (float)article.counts[i];
           inputRec[size] =   (float)article.sum ;
           inputRec[size+1] = (float)article.feedback ;
           article.score = clusterNet.getCluster(inputRec) ;
         }
         break ;
       case 2:   // feedback
         if (scoreNet != null) {
           //  pass through network --- get score
           float[] inputRec = new float[size+2] ;
           for (int i=0 ; i < size ; i++) inputRec[i] = (float)article.counts[i];
           inputRec[size] =   (float)article.sum ;
           inputRec[size+1] = (float)article.feedback ;
           article.score = scoreNet.getPrediction(inputRec) ;
         }
         break ;
      }
      // OK, now do an automatic feedback pass
      // so user doesn't have to do it for each article
      // User can override via Feedback menu option
      if (sum == 0) {
        article.feedback = 0.0 ; // negative feedback
      } else if (sum < 2) {
        article.feedback = 0.25 ;
      } else if (sum < 4) {
        article.feedback = 0.50 ;
      } else if (sum < 6) {
        article.feedback = 0.75 ;
      } else article.feedback = 1.0 ;
  }

  // score all loaded articles using the designated filter
  void score(Vector articles, String[] keywords, int filterType) {

      try {
          String id ;

          Enumeration enum = articles.elements() ;
          while (enum.hasMoreElements()) {
            trace("") ;
            NewsArticle article = (NewsArticle)enum.nextElement() ;
            score(article, keywords, filterType) ;
          }
       }
        catch (Exception e) {
            trace("Exception:" + e) ;
        }
      if (filterType == 1) computeClusterAverages(articles) ;
  }

// compute the average score for each cluster
// and set the score of each article in each cluster
// to that average value
void computeClusterAverages(Vector articles) {

   int numClusters = 4 ;          // we are using 4 for now
   int sum[] = new int[numClusters] ;
   int numArticles[] = new int[numClusters] ;
   double avgs[] = new double[numClusters] ;

   Enumeration enum = articles.elements() ;
   while (enum.hasMoreElements()) {
        NewsArticle article = (NewsArticle)enum.nextElement() ;
        int cluster = article.clusterId ;
        sum[cluster] += article.sum; // sum of counts
        numArticles[cluster]++ ; // bump counter
   }

   // now compute the average score for each cluster
   for (int i=0 ; i < numClusters ; i++) {
      if (numArticles[i] > 0) {
       avgs[i] = (double)sum[i] / (double)numArticles[i];
      } else {
        avgs[i] = 0.0 ;
      }
      trace(" cluster " + i + " avg = " + avgs[i] + "\n") ;
   }

   enum = articles.elements() ;
   while (enum.hasMoreElements()) {
      NewsArticle article = (NewsArticle)enum.nextElement() ;
      article.score = avgs[article.clusterId] ;
   }


}


// count the number of occurances of the specified keys
// in the text
int[] countWordMultiKeys(String[] keys, String text) {
     StringTokenizer tok = new StringTokenizer(text) ;
     int keyLen = 1 ;  // for now
     int counts[] = new int[keys.length] ;
     Vector table[] = new Vector[50] ; // up to length 50
     Hashtable keyHash = new Hashtable() ;

     trace("Searching for keywords ... \n") ;

     for (int i=0 ; i < keys.length ; i++) {
        int len = keys[i].length() ;
        if (table[len] == null) table[len] = new Vector() ;
        table[len].addElement(keys[i]) ;
        keyHash.put(keys[i], new Integer(i)) ;
        counts[i] = 0 ;
     }

      while (tok.hasMoreTokens()) {
        String token = tok.nextToken() ;
        int len = token.length();
        if ((len < 50) && (table[len] != null)) {
         Vector searchList = table[len] ;
         Enumeration enum = searchList.elements() ;
         while (enum.hasMoreElements()) {
           String key = (String)enum.nextElement() ;
           if (token.equalsIgnoreCase(key)) {
               Integer index = (Integer)keyHash.get(key);
               counts[index.intValue()]++ ; // found another one
               continue ;
           }
         }
        }
       }

      for (int i=0 ; i < keys.length ; i++) {
         trace("key = " + keys[i] +
               " count = " + counts[i] + "\n") ;
      }

      return counts ;
    }

 // used by neural networks to read newsfilter.dat
 void writeProfileDataDefinition(String[] keywords) {

   try {
     FileWriter writer = new FileWriter("newsfilter.dfn") ;
     BufferedWriter out = new BufferedWriter(writer) ;

     for (int i=0 ; i < keywords.length ; i++) {
      out.write("continuous ") ;
      out.write(keywords[i]) ;
      out.newLine() ;
     }
     out.write("continuous ClassField") ;  // feedback
     out.newLine() ;
     out.write("continuous score") ;
     out.newLine() ;
     out.flush() ;
     out.close() ;
   }
   catch (IOException e) {
     System.out.println("Error: couldn't create 'newsfilter.dfn' \n");
   }
  }



  void addArticleToProfile(NewsArticle currentArt) {

   try {
     FileWriter writer = new FileWriter("newsfilter.dat", true) ;
     BufferedWriter out = new BufferedWriter(writer) ;

     out.write(currentArt.getProfileString()) ;
     out.newLine() ;
     out.flush() ;
     out.close() ;
   }
   catch (IOException e) {
     System.out.println("Error: couldn't append article to profile \n");
   }
  }

  void addAllArticlesToProfile(Vector articles) {
  try {
     FileWriter writer = new FileWriter("newsfilter.dat", true) ;
     BufferedWriter out = new BufferedWriter(writer) ;

     Enumeration enum = articles.elements() ;
     while (enum.hasMoreElements()) {
       NewsArticle art = (NewsArticle)enum.nextElement() ;
       out.write(art.getProfileString()) ;
       out.newLine() ;

     }
       out.flush() ;
       out.close() ;
   }
   catch (IOException e) {
     System.out.println("Error: couldn't append article to profile \n");
   }


  }

  void trainClusterNet() {

      DataSet dataSet = new DataSet("ProfileData", "newsfilter") ;
      dataSet.setDisplay(textArea) ;
      dataSet.loadDataFile() ; // load the data set

      // create a KMap neural network
      // specify the newsfilter.dat as the training data
      // cluster it into 4 clusters

      clusterNet = new KMapNet("NewsFilter Cluster Profile");
      clusterNet.textArea1 = textArea ;
      clusterNet.ds = dataSet ;
      clusterNet.numRecs = dataSet.numRecords ;
      clusterNet.fieldsPerRec = dataSet.fieldsPerRec ;
      clusterNet.data = dataSet.normalizedData ; // get vector of data

      // create network, all fields are inputs
      clusterNet.createNetwork(clusterNet.fieldsPerRec, 2, 2) ;
      int maxNumPasses = 20 ;
	  int numRecs = clusterNet.numRecs ;

      // train the network
      for (int i= 0 ; i < maxNumPasses ; i++) {
	    for (int j=0 ; j < numRecs; j++) {
    	      clusterNet.cluster() ;   // train
	    }
	    runnit.yield() ;  // after each pass
      }

	  clusterNet.mode = 1 ;  // lock the network weights
	  for (int i=0 ; i < clusterNet.numRecs ; i++) {
        clusterNet.cluster() ;   // test
 //       clusterNet.display_network() ;
      }


  }


  void trainScoreNet() {
      DataSet dataSet = new DataSet("ProfileData", "newsfilter") ;
      dataSet.setDisplay(textArea) ;
      dataSet.loadDataFile() ; // load the data set

      scoreNet = new BackProp("NewsFilter Score Model");
	  scoreNet.textArea1 = textArea ;
	  scoreNet.ds = dataSet ;
      scoreNet.numRecs = dataSet.numRecords ;
      scoreNet.fieldsPerRec = dataSet.normFieldsPerRec ;
      scoreNet.data = dataSet.normalizedData ;    // get vector of data
      int numOutputs = dataSet.getClassFieldSize() ;
	  int numInputs = scoreNet.fieldsPerRec - numOutputs;

	  scoreNet.createNetwork(numInputs,2 * numInputs,numOutputs) ;
	  int maxNumPasses = 2500 ;  // default -- could be on applet
	  int numRecs = scoreNet.numRecs ;
	  for (int i = 0 ; i < maxNumPasses ; i++) {
     	 for (int j=0 ; j < numRecs; j++) {
	       scoreNet.process() ;   // train
	     }
 	     runnit.yield() ;  // after each pass
	  }
	  scoreNet.textArea1.append("\n Passes Completed: " + maxNumPasses +
                            "  RMS Error = " + scoreNet.aveRMSError + " \n") ;

      scoreNet.mode = 1 ;  // lock the network
	  // do a final pass and display the results
	  for (int i=0 ; i < scoreNet.numRecs ; i++) {
	      scoreNet.process() ;   // test
//	      scoreNet.display_network() ;
	  }


  }


 };

